// dummy implementation of BLAS interface 


//#include <cblas.h>
#include <cblas_RSF.h>

//#include <cblas.h>
double cblas_dsdot(int n, const float *x, int sx, const float *y, int sy)
/*< x'y float -> double >*/
{
  int i, ix, iy;
  double dot;

  dot = 0.;
  for (i=0; i < n; i++) {
	  ix = i*sx; iy = i*sy;
    dot += (double) x[ix] * y[iy];
  }
  return dot;
}

void cblas_saxpy(int n, float a, const float *x, int sx, float *y, int sy)
/*< y += a*x >*/
{
  int i, ix, iy;
  
  for (i=0; i < n; i++) {
	  ix = i*sx;
	  iy = i*sy;
	  y[iy] += a * x[ix];
  }
}

void cblas_sswap(int n, float *x, int sx, float* y, int sy) 
/*< swap x and y >*/
{
  int i, ix, iy;
  float t;

  for (i=0; i < n; i++) {
	  ix = i*sx; iy = i*sy;
	  t = x[ix];
  	x[ix] = y[iy];y[iy] = t;
  }
  return;
}

float cblas_sdot(int n, const float *x, int sx, const float *y, int sy)
// x'y float
{
  int i, ix, iy;
  double dot;
  dot = 0.;

  for (i=0; i < n; i++) {
	  ix = i*sx; iy = i*sy;
	  dot += x[ix] * y[iy];
  }
  return (float)dot;
}

float cblas_snrm2 (int n, const float* x, int sx) 
/*< sum x_i^2 >*/
{
  int i, ix;
  double xn = 0.0;

  for (i=0; i < n; i++) {
	  ix = i*sx;
	  xn += x[ix]*x[ix];
  }
  return (float)xn;
}

float cblas_scnrm2 (int n, const void* x, int sx) 
/*< sum |x_i|^2 >*/
{
  int i, ix;
  double xn;
  xn = 0.0;
  const float* fx = (const float*) x;
  for (i=0; i < n; i++) {
    ix = i*sx*2;
    xn += (fx[ix]*fx[ix] + fx[ix+1]*fx[ix+1]);
  }
  return (float)(xn);
}

void cblas_sscal(int n, float alpha, float *x, int sx)
/*< x = alpha*x >*/
{
  int i, ix;
  for (i=0; i < n; i++) {
    ix = i*sx;
    x[ix] *= alpha;
  }
}

void cblas_csscal(int n, float alpha, void *x, int sx)
/*< x = alpha*x >*/
{
  int i, ix;
  float* fx = (float*)x;
  for (i=0; i < n; i++) {
    ix = i*sx*2;
    fx[ix]   *= alpha;
    fx[ix+1] *= alpha;
  }
}

void cblas_cdotc_sub(int n, 
		     const void *x, int sx,
		     const void *y, int sy, void *dot)
/*< complex hermitian dot product, conj(A)*B >*/
{
  float *cx, *cy;
  double prodr,prodi, dotr,doti;
  int i, ix, iy;
  
  cx = (float*)x; cy = (float*)y;

  for (i=0; i < n; i++) {
    ix = i*sx*2;
    iy = i*sy*2;
    prodr = cx[ix]*cy[iy]   + cx[ix+1]*cy[iy+1];
    prodi = cx[ix]*cy[iy+1] - cx[ix+1]*cy[iy];
    dotr += prodr;
    doti += prodi;
  }
  
  float* fdot = (float*) dot;
  fdot[0] = dotr;
  fdot[1] = doti;
  return;
}

