#ifndef CUBE_ALGO2_H
#define CUBE_ALGO2_H

#include "cubeData.hpp"
#include "decl_fwd.h"
using std::pair;

namespace Yang {

///Invert a 2X2 matrix [ in[0]  in[1] ]  ==> (out[0]  out[1] )
///[ in[2]  in[3] ]  ==> (out[2]	 out[3] )
///Function returns the conditional number
inline float invert2X2(float *out, const float *in, float& determ) {
   determ = in[0]*in[3] - in[1]*in[2];
   float tmp = 1.0f/determ;
   out[0] = tmp * in[3];
   out[1] = tmp * -in[1];
   out[2] = tmp * -in[2];
   out[3] = tmp * in[0];
   //Calculate the conditional number, which is an indication of the linear dependency of the matrix's columns.
   tmp = in[0] + in[3];
   float cond, tmp2;
   tmp2 = tmp*tmp - 4 * determ;
   if (tmp2<= 0.0f) {
	  cond = 1.0f; //eigenvalues are complex conjugate compair.
   }
   else {
	  cond = (tmp+sqrtf(tmp2)) / (tmp-sqrtf(tmp2));
	  if (fabsf(cond) < 1.0f) cond = 1.0f / cond;
	  cond = fabsf(cond);
   }
   return cond;
}

  void sqrtCube(cubeData<float>& out, const cubeData<float>& input);
  
  float clipCubeVal(cubeData<float>& out, float fmax, float fmin,
                       bool bClipMax=true, bool bClipMin=true);
  //Elem by Elem divide, but be more careful about zeros.
  void buildThresholdMask(cubeData<float>& Mask, const cubeData<float>& input, float vThres , bool bUseAbsVal = true, bool bMarkHighValue=true);
  void divideWithCaveat(cubeData<float>& out, const cubeData<float>& in, cubeData<float>& divident,
                      float qt_mask, float qt_eps, bool bUseMask, cubeData<float>* pMask, bool bMaskInput = false);
  
  void duplicateAlongAxis(cubeData<float>& out, const cubeData<float>& in, int iAxis);
  
  void deriv1stAtIndex(cubeData<float>& resultCube, const cubeData<float>& inCube, int iAxis, int index, bool bScaleBy_d);
  void deriv2ndAtIndex(cubeData<float>& resultCube, const cubeData<float>& inCube, int iAxis, int index, bool bScaleBy_d);
  
  void getGaussfunc(cubeDataF& gaussfunc,const axis& ax,float fhalfMaxWidth, float b=0.0f, int order=0);
  //maxRatio=0.5 mean half maximum width, 0.1 means 10th maximum width
  void getGaussfuncWithMaxWidth(const PosFloat& maxRatio, cubeDataF& gaussfunc,const axis& ax,float fMaxWidth, float b=0.0f, int order=0);
  void getOrmsbyFunc(cubeDataF& orbFunc,const axis& ax_t,const float freqs[4]);
    
  void dotProdAlongAxis(cubeDataF& resultCube, const cubeDataF& inCube, const cubeDataF& dot2, int iAxis);
  
  
  ///Input cube and outcube must be <=5 dimension. Must has the same grid spacing. But the grid.n grid.o might be different.
  ///This function will first calculate the overlap of domain's of Input and output, and carve only that poriton to outcube.
  ///Note that The uncovered part in outcube will be kept original if bClearOutputFirst=false;
  template<typename T>
  void cropAlign(cubeData<T>& outCube, const cubeData<T>& inCube, bool bAdd=false, bool bClearOutputFirst=false) {
    FUNC_TRACE_BEG(cropAlign);
    if (bClearOutputFirst) {
      assert(!bAdd); //logically should not happen
      outCube.zero();
    }
    
    const hypercube& shape1 = outCube.domain();
    const hypercube& shape2 = inCube.domain();
    int ndim = shape1.get_ndim();
    assert(ndim == shape2.get_ndim());
    assert(ndim <= 5);
    vecAxis axes1,axes2;
    axes1 = shape1.get_axes(); axes2 = shape2.get_axes();
    //pad the axes
    axis ax_dummy(1,0.f,1.0f); int i;
    for (i=ndim+1; i<=5; i++) {axes1.push_back(ax_dummy); axes2.push_back(ax_dummy);}
    
    int n1,n2,n3,n4,n5; n2=n3=n4=n5=1;
    float d1,d2,d3,d4,d5;
    float o1beg,o2beg,o3beg,o4beg,o5beg;
    int o1beg1n,o2beg1n,o3beg1n,o4beg1n,o5beg1n;
    int o1beg2n,o2beg2n,o3beg2n,o4beg2n,o5beg2n; //The start location for each cube
    o2beg1n = o2beg2n = 0; o3beg1n = o3beg2n = 0;
    o4beg1n = o4beg2n = 0; o5beg1n = o5beg2n = 0;
    
    int axisMergeTo = 1;
    long long int n1_lump=1; //find if we can accelerate the memcpy if the first several axes are exactly same, therefore we can lump the first few axes.
    for (i=0; i<ndim-1; i++) { //don't lump the last axis, saftey guard
      if (axes1[i].same_axis(axes2[i])) {
        axisMergeTo=i+1; n1_lump *= axes1[i].n;
        if (n1_lump > 1073741824) { //don't put it too big, i.e. ax.n < 2^30, if exceed roll back
          VERIFY(i>0);
          axisMergeTo=i; n1_lump /= axes1[i].n; break;
        }
      } else {
        break; //otherwise just break the loop
      }
    }
    axis ax_tmp(axes1[0]); ax_tmp.n=n1_lump;
    //lump from the 1st to the "axisMergeTo"_th axis
    if (axisMergeTo>=2) {
      axes1.erase(axes1.begin(),axes1.begin()+axisMergeTo);
      axes1.insert(axes1.begin(),ax_tmp);
      axes2.erase(axes2.begin(),axes2.begin()+axisMergeTo);
      axes2.insert(axes2.begin(),ax_tmp);      
      ndim = ndim - axisMergeTo + 1;
    }
    
    //Calculate the range of overlap for axis 1
    alignAxis(o1beg,n1,d1,axes1[0],axes2[0]); 
    o1beg1n = lroundf((o1beg-axes1[0].o)/d1); o1beg2n = lroundf((o1beg-axes2[0].o)/d1);
    if (ndim>1) { //2nd axis
      alignAxis(o2beg,n2,d2,axes1[1],axes2[1]);
      o2beg1n = lroundf((o2beg-axes1[1].o)/d2); o2beg2n = lroundf((o2beg-axes2[1].o)/d2);
    }
    if (ndim>2) { //3rd axis
      alignAxis(o3beg,n3,d3,axes1[2],axes2[2]);
      o3beg1n = lroundf((o3beg-axes1[2].o)/d3); o3beg2n = lroundf((o3beg-axes2[2].o)/d3);
    }
    if (ndim>3) { //4th axis
      alignAxis(o4beg,n4,d4,axes1[3],axes2[3]);
      o4beg1n = lroundf((o4beg-axes1[3].o)/d4); o4beg2n = lroundf((o4beg-axes2[3].o)/d4);
    }
    if (ndim>4) { //5th axis
      alignAxis(o5beg,n5,d5,axes1[4],axes2[4]);
      o5beg1n = lroundf((o5beg-axes1[4].o)/d5); o5beg2n = lroundf((o5beg-axes2[4].o)/d5);
    }

    //PRV(o1beg1n);PRV(o1beg2n);PRV(o1beg);PRV(n1);PRV(d1);PRV(o2beg1n);PRV(o2beg2n);PRV(o2beg);PRV(n2);PRV(d2);
    //PRV(o3beg1n);PRV(o3beg2n);PRV(o3beg);PRV(n3);PRV(d3);
     
    //Copy the values
    T * __restrict p1; const T* __restrict p2;
    T *pOff1; const T* pOff2;
    p1 = outCube.dataPtr(); p2 = inCube.dataPtr();
    unsigned int  off4_1,off4_2, off5_1,off5_2;
    long long     off2_1,off2_2,off3_1,off3_2;
    for (int i5=0; i5<n5; i5++) {
      off5_1 = o5beg1n+i5; off5_2 = o5beg2n+i5;
      for (int i4=0; i4<n4; i4++) {
        off4_1 = off5_1*axes1[3].n+o4beg1n+i4; 
        off4_2 = off5_2*axes2[3].n+o4beg2n+i4;
        for (int i3=0; i3<n3; i3++) {
          off3_1 = off4_1*axes1[2].n+o3beg1n+i3;
          off3_2 = off4_2*axes2[2].n+o3beg2n+i3;
          for (int i2=0; i2<n2; i2++) {
            off2_1  = off3_1*axes1[1].n + o2beg1n+i2;
            off2_2  = off3_2*axes2[1].n + o2beg2n+i2;
            pOff1   = p1+off2_1*axes1[0].n+o1beg1n;
            pOff2   = p2+off2_2*axes2[0].n+o1beg2n;
            if (!bAdd) {
              memcpy(pOff1,pOff2,sizeof(T)*n1);
            } else {
              for (int i1=0; i1<n1; i1++) {
                *(pOff1+i1) += *(pOff2+i1);
              }
            }
          }
        }
      }
    }
    FUNC_TRACE_END()
  }
  
  int rectSmooth (cubeData<float>& cubIn, cubeData<float>& cubOut, int rect1, int rect2, int rect3, int nrepeat=1, bool bAbsval=false);
  void RtoC (      cubeData<cmplxf>& cubC, const cubeData<float>& cubR, const cubeData<float>& cubI);
  void CtoR (const cubeData<cmplxf>& cubC, cubeData<float>& cubR,       cubeData<float>& cubI);
  void cubeAbs (const cubeData<float>& cubi, cubeData<float>& o); //Take the absolute value

  //Like Transpose, but more flexible.
  template <typename T>
  void shuffleAxis(cubeData<T>& out, const cubeData<T>& in,
                   const stdVecInt& reshape, const stdVecInt& axMapping,bool bStuffDummyAxis=false ,bool bOutMemAllocInside=true,bool bVerbose=false) {
    FUNC_TRACE_BEG(Yang::shuffleAxis);
    int nAxes = reshape.size();
    hypercube inShape = in.domain();
    assert(nAxes = int(axMapping.size()));
    stdVecInt axtt(axMapping);
    assert((unique(axtt.begin(),axtt.end())) == axtt.end());
    
    int ndim = reshape[nAxes-1]; int ndimIn = inShape.get_ndim();
    //PRV(ndim);PRV(ndimIn);
    
    if (!bStuffDummyAxis) {
  	 assert(ndim == ndimIn);
    } else {
  	 assert(ndim >=ndimIn);
  	 //Stuff dummy axisT
  	 vecAxis inAxes = inShape.get_axes();
  	 for (int idim=ndimIn+1; idim<=ndim; idim++) {
  		inAxes.push_back(axis(1,0.0f,23.75f));
  	 }
  	 inShape.set_axes(inAxes);
    }
    stdVecInt ns(nAxes) ; //The number of points on those super axis.
    
    //record the shuffle result
    std::vector<std::pair<int,int> > dimSegVec(nAxes);
    
    int iax,iaxMap, idimBeg=1,idimEnd, idim;
    for (iax=1;iax<=nAxes;iax++) {
  	  ns[iax-1] = 1;
  	 idimEnd = reshape[iax-1];
  	 assert(idimEnd >= idimBeg);
  	 for (idim=idimBeg; idim<=idimEnd; idim++) {
  		ns[iax-1] *= inShape.get_n(idim);
  	 }
  	 //Need to put the axes to the correct location at outAxes accroding to the axMapping
  	 iaxMap = axMapping[iax-1];
  	 assert(iaxMap>0 && iaxMap<= nAxes);
  	 dimSegVec[iaxMap-1] = pair<int,int>(idimBeg,idimEnd);
  	 
  	 //update idimBeg
  	 idimBeg = idimEnd + 1;
    }
    //PRVEC(ns,nAxes); 
    
    //Figure out the out shape
    hypercube outShape;
    vecAxis outAxes;
    vector<long long> n123Out(nAxes+1); //The accumulated n123 for the output shape
    n123Out[0] = 1;
    n123Out[1] = 1;
    for (iax=1; iax<=nAxes; iax++) {
  	 
  	 idimBeg = dimSegVec[iax-1].first;
  	 idimEnd = dimSegVec[iax-1].second;
  	 for (idim=idimBeg; idim<=idimEnd; idim++) {
  		outAxes.push_back(inShape.get_axis(idim));
  	 }
  	 n123Out[iax] *= inShape.get_n23(idimBeg,idimEnd);
  	 if (iax!=nAxes) {
  		n123Out[iax+1] = n123Out[iax];
  	 }
    }
    
    assert(int(outAxes.size())==ndim);
    outShape.set_axes(outAxes);
    
    //outShape.info("outShape");
    
    if (bOutMemAllocInside) {
  	 out.alloc(outShape);
    } else {
  	 VERIFY(out.domain().sameSize(outShape));
    }
    
    //Transp the memory
    const T* pIn = in.dataPtr();
    T* pOut = out.dataPtr();
  
    vector<int> axMapT = axMapping;
  #define MAX_ALLOW_AXES_NUM 7
    //tranverse this input nAxes dimension array, and put the input into correct output location.
    VERIFY(nAxes<=MAX_ALLOW_AXES_NUM); //Only handles dimension smaller than 7
    for (int i=nAxes; i<MAX_ALLOW_AXES_NUM; i++) {
  	 ns.push_back(1);
  	 axMapT.push_back(i+1);
    }
    for (int i=nAxes+1; i<MAX_ALLOW_AXES_NUM+1; i++) {
  	 n123Out.push_back(n123Out[nAxes]);
    }
  
    long long off7,off6,off5,off4,off3,off2,off1, o7,o6,o5,o3,o4,o2,o1;
    long long nt;
    off7 = 0;
    
    if (!bVerbose) {
      for (int i7=0; i7<ns[6]; i7++) {
        o7 = i7*n123Out[axMapT[6]-1];
        off6 = (off7+i7)*ns[5];
        for (int i6=0; i6<ns[5]; i6++) {
          o6 = o7 + i6*n123Out[axMapT[5]-1];
          off5 = (off6+i6)*ns[4];
          for (int i5=0; i5<ns[4]; i5++) {
            o5 = o6 + i5*n123Out[axMapT[4]-1];
            off4 = (off5+i5)*ns[3];
            for (int i4=0; i4<ns[3]; i4++){
              o4 = o5 + i4*n123Out[axMapT[3]-1];
              off3 = (off4+i4)*ns[2];
              for (int i3=0; i3<ns[2]; i3++){
                o3 = o4 + i3*n123Out[axMapT[2]-1];
                off2 = (off3+i3)*ns[1];
                for (int i2=0; i2<ns[1]; i2++){
                  o2 = o3 + i2*n123Out[axMapT[1]-1];
                  off1 = (off2+i2)*ns[0];
                  for (int i1=0; i1<ns[0]; i1++){
                    nt = n123Out[axMapT[0]-1];
                    o1 = o2 + i1*nt;
                    pOut[o1] = pIn[off1+i1];
                  }
                }
              }
            }
          }
        }
      }
    } else { //bVerbose, report progress
      prt("\n ");
      char ch_rn = '\r';
      for (int i7=0; i7<ns[6]; i7++) {
        if (nAxes==7 && ns[6]>1) { prt("%c ShuffleAxis, starting on i7=%d",ch_rn,i7);}
        o7 = i7*n123Out[axMapT[6]-1];
        off6 = (off7+i7)*ns[5];
        for (int i6=0; i6<ns[5]; i6++) {
          if (nAxes==6 && ns[5]>1) { prt("%c ShuffleAxis, starting on i6=%d ",ch_rn,i6);}
          o6 = o7 + i6*n123Out[axMapT[5]-1];
          off5 = (off6+i6)*ns[4];
          for (int i5=0; i5<ns[4]; i5++) {
            if (nAxes==5 && ns[4]>1) { prt("%c ShuffleAxis, starting on i5=%d ",ch_rn,i5);}
            o5 = o6 + i5*n123Out[axMapT[4]-1];
            off4 = (off5+i5)*ns[3];
            for (int i4=0; i4<ns[3]; i4++){
              if (nAxes==4 && ns[3]>1) { prt("%c ShuffleAxis, starting on i4=%d ",ch_rn,i4);}
              o4 = o5 + i4*n123Out[axMapT[3]-1];
              off3 = (off4+i4)*ns[2];
              for (int i3=0; i3<ns[2]; i3++){
                if (nAxes==3 && ns[2]>1) { prt("%c ShuffleAxis, starting on i3=%d ",ch_rn,i3);}
                o3 = o4 + i3*n123Out[axMapT[2]-1];
                off2 = (off3+i3)*ns[1];
                for (int i2=0; i2<ns[1]; i2++){
                  o2 = o3 + i2*n123Out[axMapT[1]-1];
                  off1 = (off2+i2)*ns[0];
                  for (int i1=0; i1<ns[0]; i1++){
                    nt = n123Out[axMapT[0]-1];
                    o1 = o2 + i1*nt;
                    pOut[o1] = pIn[off1+i1];
                  }
                }
              }
            }
          }
        }
      }
      prt("\n ");
    } //end if bVerbose
    FUNC_TRACE_END();
  } //end func shuffleAxis
} //end namespace Yang

#endif

