#include "stdafx.h"
#include "dampingFunc.h"

using namespace Yang;
using namespace std;

void Yang::dampingFunc::getCoef(stdVecFloat &coefVec, int length, float extent) {
  int fullLength = int(length / extent + 0.5);
  
  //calc the coef based on full length
  coefVec.resize(length);
  float fpow = this->pars_->power_;
  taperFunction func_type = pars_->func_type_;
  float frac;
  
  //PRV2(length,extent);  
  if (func_type == powerfunc) {
    for (int i = 0; i < length; i++) {
      frac = float(i) / float(fullLength);
      coefVec[i] = (1.0f - powf(frac,fpow));
    }
  } else if (func_type == cosfunc) {
    for (int i = 0; i < length; i++) {
      frac = float(i) / float(fullLength);
      coefVec[i] = powf(cos(frac * PI / 2),fpow);
    }
  } else if (func_type == cosPlus1) {
    float sgn = 1.0f;
    for(int i=0; i<length; i++) {
      frac = float(i)/float(fullLength);
      sgn = cosf(frac*PI)<0.0f? -1.0f:1.0f;
      coefVec[i] = 0.5f*(sgn*powf(fabs(cosf(frac*PI)),fpow)+1.0f);
    } 
  } else {
    VERIFY(0);
  }
  
  //scale is 0.0f if frac >1.0f
  for (int i = 0; i < length; i++) {
    frac = float(i) / float(fullLength);
    if (frac >=1.0f) {
      coefVec[i] = 0.0f;
    } 
  }
}

//the formula cos(0,pi/2) 
void cosDampingFunc::getCoef(stdVecFloat& coefVec, int length, float extent)
{
   int fullLength = int(length/extent+0.5);
   
   //calc the coef based on full length
   coefVec.resize(length);
   float frac;
   for(int i=0; i<length; i++)
   {
	  frac = float(i+1)/float(fullLength);
	  coefVec[i] = cos(frac*PI/2);
   }
}

//the cos(0,pi/2) 
void lineDampingFunc::getCoef(stdVecFloat& coefVec, int length, float extent)
{
   int fullLength = int(length/extent+0.5);
   
   //calc the coef based on full length
   coefVec.resize(length);
   float frac;
   for(int i=0; i<length; i++)
   {
	  frac = float(i+1)/float(fullLength);
	  coefVec[i] = 1 - frac;
   }
}

