#ifndef _DAMPINGFUNC_H_
#define _DAMPINGFUNC_H_

#include "stdafx.h"
#include "decl_fwd.h"
#include "param_func.h"


namespace Yang {

using std::make_pair;
using std::pair; 

//damping func, exponential/cosine/hamming window all kinds of smooth decreasing function
class dampingFunc {
public:
  typedef enum __taperFunction {
    powerfunc,
    cosfunc,
    cosPlus1
  } taperFunction; //taper function type, all these assume change for 1.0 to 0.0f.

  struct params_df {
  public:
    taperFunction func_type_;
    fwk::NnegFloat power_; //the power of functions.
  public:
    void readParam(param_func* par, const string& prefix) {
      string strPar(prefix);
      strPar += "funcType";

      pair<int, taperFunction> TF_match_table[] = {make_pair(0, powerfunc), make_pair(1, cosfunc), make_pair(2, cosPlus1)};
      this->func_type_ = findEnumItemFromPar(par, TF_match_table, NELEM(TF_match_table), strPar, 0);
      
      //The sharpness is determined by the power of the power function, pow=1 is linear shape and pow2 is parabolic shape.
      strPar = prefix; strPar += "pow";
      power_ = fwk::NnegFloat(par->get_float(strPar,1.0f));
      return;
    }
  };
  
  
  //constructor
  dampingFunc(const shared_ptr<params_df>& __par): pars_(__par) {} 
  virtual ~dampingFunc() {}
   
  //extent=1.0, means damping to 0(full window length). otherwise damping ends at a fraction of the
  //full window length
  void getCoefEx(stdVecFloat &coefVec, int length, float extent) {
    if (!equalf(extent, 0.0f)) {
      this->getCoef(coefVec, length, extent);
    }
    //since extent=0.0 equals to disable the dampping. Set everything to 1.0
    else {
      coefVec.assign(length, 1.0f);
    }
  }
  
protected:
  shared_ptr<params_df> pars_;
    
  virtual void getCoef(stdVecFloat &coefVec, int length, float extent);
  
};

class cosDampingFunc : public dampingFunc {
public:
  virtual void getCoef(stdVecFloat &coefVec, int length, float extent);
};

class lineDampingFunc : public dampingFunc {
public:
  virtual void getCoef(stdVecFloat &coefVec, int length, float extent);
};

} //end namespace
#endif
