#include "stdafx.h"
#include "mpi_util.h"
#include "dmHeaderReader.h"
// TODO: your code here
static string strCmd;
static char szCmd[4096];

#include <sstream>
using namespace Yang;
using namespace std;

string Yang::get_basename_from_path(const string& fullpath) {
  size_t pos = fullpath.find_last_of("/");
  string base;
  if (pos != std::string::npos) {
    base.assign(fullpath.begin()+pos+1,fullpath.end());
  } else {
    base = fullpath;
  }
  return base;
}

Yang::branchFileName::branchFileName(const std::string &origFileName, const string &suf) {
  origFn_ = origFileName; suf_ = suf;
  //find the extension . position
  extPos_ = origFn_.rfind('.');
  VERIFY(extPos_ != string::npos); VERIFY(origFn_.at(extPos_ + 1) != '/');
  return;
}

string branchFileName::fn_at(int irank) const {
  stringstream &ss = const_cast<stringstream&>(ss_);
  ss.str(""); ss << "-" << suf_ << "-np" << irank;
  string new_fn(origFn_); new_fn.insert(extPos_, ss.str());
  return new_fn;
}

string branchFileName::stackCmd(int nrank) const {
  string cmd;
  cmd = "Add ";
  for (int ir = 1; ir < nrank; ir++) {
    cmd += this->fn_at(ir); cmd.append(" ");
  }
  cmd.append(" > "); cmd += origFn_;
  return cmd;
}

string branchFileName::catCmd(int nrank, int iAxis) const {
  std::stringstream ss;
  ss << "Cat ";
  for (int ir = 1; ir < nrank; ir++) {
    ss << " " << this->fn_at(ir) << " ";
  }
  ss << " axis=" << iAxis << " ";
  ss << " > " << origFn_;
  return ss.str();
}

Yang::workScheduler1d::workScheduler1d(int njobs, int np, scheduleType sType) : njobs_(njobs), np_(np) {
  ns_.resize(np); os_.resize(np); ds_.resize(np);
  int ip = 0;
  int n, o, d;
  VERIFY(njobs >= 1); VERIFY(np >= 1);
  if (sType == hopping) {
    for (ip = 0; ip < np; ip++) {
      //the number of shots for this process, it handles shots ir,ir+nr,ir+2*nr,..., till shots nsx
      n = int((njobs - (ip + 1)) / np) + 1;
      o = ip;
      d = np;
      ns_[ip] = n; os_[ip] = o; ds_[ip] = d;
    }
  } else if (sType == chunk) {
    int nj_left = njobs;
    for (ip = 0; ip < np; ip++) {
      //the number of shots for this process, it handles shots ir,ir+1,ir+nr-1, nr might vary.
      d = 1;
      n = nj_left / (np - ip) + (nj_left % (np - ip) == 0 ? 0 : 1);
      o = njobs - nj_left;
      nj_left -= n;
      ns_[ip] = n; os_[ip] = o; ds_[ip] = d;
    }
    VERIFY(nj_left == 0);
  } else {
    VERIFY(0);
  }
}

int Yang::split_file_MPI(const string& fnIn, const string& sufOut, int iAxis, const workScheduler1d& ws, string datapath) {
  //The outfile will be in the same folder as in File
  int np    = ws.np(); //np is number of slaves.
  VERIFY(np>=1);

  //Read in the header info.
  sepHeaderReader hReader(fnIn, dm::IN, "", true);
  hypercube shp;
  hReader.getShape(shp);

  shp.info("shp");
  //const axis& ax = shp.get_axis(iAxis);
  
  branchFileName bf(fnIn,sufOut);
  int npp,fpp;
  char szCmd[4096];

  for (int ir=1; ir<=np; ir++) {
    int n,o,d;
    ws.get_nod(ir-1,&n,&o,&d);
    npp = n;  fpp = o; 
    sprintf(szCmd,"Window3d <%s f%d=%d j%d=%d n%d=%d squeeze=n >%s", fnIn.c_str(), iAxis,fpp, iAxis,d, iAxis,npp, bf.fn_at(ir).c_str()); 
    if (!datapath.empty()) {
      char sz[1024];
      sprintf(sz," datapath=%s ", datapath.c_str());
      strcat(szCmd,sz);
    }
    PRV(szCmd);
    VERIFY(system(szCmd)==0);
  }
  return 0;
}

//simply duplicate file for the purpose of MPI
int Yang::dup_file_MPI(const string& fnIn, const string& sufOut,const workScheduler1d& ws, string datapath) {
  //The outfile will be in the same folder as in File
  int np    = ws.np(); //np is number of slaves.
  VERIFY(np>=1);

  branchFileName bf(fnIn,sufOut);
  for (int ir=1; ir<=np; ir++) {
    if (!datapath.empty()) {
      sprintf(szCmd,"Cp %s %s datapath=%s ", fnIn.c_str(), bf.fn_at(ir).c_str(),datapath.c_str());  //copy binary
    } else { //not copying binary
      sprintf(szCmd,"cp %s %s ", fnIn.c_str(), bf.fn_at(ir).c_str());
    }
    PRV(szCmd);
    VERIFY(system(szCmd)==0);
  }
  return 0;
}

