#include "stdafx.h"
#include "decl_fwd.h"
#include "waveOpMPI.h"
#include "my_mpi.h"
#include "cubeData.hpp"
#include "my_timer.h"

#define MPI_BIN_MSG_SIZE 65536 //The magic size for sending binary blocks through mpi, in bytes

extern "C" {
  #include "mpi_sep.h"
}

#define DEBUG
#include "myDebug.h"
#include "mpi_util.h"

using std::string;
using std::stringstream;
using namespace mpiY;
using namespace Yang;

mpiY::environment::environment(int* p_argc, char*** p_argv) {
  MPI_Init(p_argc,p_argv);
}

int mpiY::slaveModeMPI::on() {
  string strCmd;
  mpiY::communicator world;

  Yang::cubeDataF data;

  while(1) {
    const MPI_Status& stat = world.recv(0,MPI_ANY_TAG,strCmd);
    if (stat.MPI_TAG == EXITAG) {
      break; //exit the loop
    } else if (stat.MPI_TAG == WRKTAG) {
      //receive a cmd line and do the wrk
      if (world.rank()==2) {
        //LOG("From slaveModeMPI::on");PRV(strCmd);
      }
      VERIFY(system(strCmd.c_str())==0); usleep(int(1e5)); //puase for a while.
      //if (world.rank()==2) LOG("after executing cmd in slaveMode::on");
    } else if (stat.MPI_TAG == IDLETAG) { //for the master to test that if slaves are Idle.
      if (world.rank()==2) {
        LOG("Recv IDLETAG");
      } //do nothing.
    } else if (stat.MPI_TAG == REDUCETAG) {
      //Load the file name, assume its size is not too big, and do an MPI_Reduce
      string fn = strCmd;
      //BroadCast the size of the file, number of elements per blk and nblk, MPI cannot handle very large msg size.
      // N = per_blk*nblk+nleft;
      unsigned per_blk, nblk, nleft; int ne[3];
      exeMPI(MPI_Bcast(ne,3,MPI_UNSIGNED,0,MPI_COMM_WORLD));
      //Load the file
      data.load(fn);
      per_blk = ne[0]; nblk = ne[1]; nleft = ne[2];
      long long count = per_blk*nblk+nleft;
      if (count!=data.domain().get_n123()) {
        SEEV2(count,data.domain().get_n123());
        VERIFY(0);
      }
      float* pSend = const_cast<float*>(data.dataPtr());
      for (unsigned int iblk=0; iblk<nblk; iblk++) {
        exeMPI(MPI_Reduce(pSend,NULL,per_blk,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
        pSend += per_blk;
      }
      //the last remaining block
      exeMPI(MPI_Reduce(pSend,NULL,nleft,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
    } else {
      VERIFY(0);
    }
  }
  return 0;
}

//3D case, use dynamic scheduling
int mpiY::slaveModeMPI3d::on() {
  string strCmd;
  mpiY::communicator world;

  Yang::cubeDataF data;
  unsigned int per_blk, nblk, nleft; unsigned int ne[3];
  int irank = world.rank();
  stringstream ss;
  char hostname[1024];
  gethostname(hostname,1024);
  
  ss << "wrker No." << irank << ", host:" << hostname;
  Yang::timer clk_wrker("0",ss.str());
  string str_clkmsg;
  while(1) {
    //always listen to the server for cmd
    //cerr << "M1, wrker "<< irank << " about to recv cmd";
    const MPI_Status& stat = world.recv(0,MPI_ANY_TAG,strCmd);
    if (stat.MPI_TAG == EXITAG) {
      break; //exit the loop
    } else if (stat.MPI_TAG == WRKTAG) {
      //receive a cmd line and do the wrk
      if (world.rank()==2) {
        //LOG("From slaveModeMPI::on");PRV(strCmd);
      }
      clk_wrker.reset("startwrk",&str_clkmsg); clk_wrker.go();
      //PRV(str_clkmsg);
      //cerr << endl; PRV(strCmd); cerr << endl; cerr.flush();
      if (system(strCmd.c_str())!=0) {
        prt("irank=%d, cmd execution failed!! \n", strCmd.c_str());
        VERIFY(0);
      }
      usleep(int(1e5)); //puase for a while.
      clk_wrker.pause(); clk_wrker.check("finsih wrk", &str_clkmsg);
      //PRV(str_clkmsg);
      world.send(0,FINISHTAG,str_clkmsg); //tell the root that I am done.

      //if (world.rank()==2) LOG("after executing cmd in slaveMode::on");
    } else if (stat.MPI_TAG == IDLETAG) { //for the master to test that if slaves are Idle.
      if (world.rank()==2) {
        //LOG("Recv IDLETAG");
      } //do nothing.
    } else if (stat.MPI_TAG == COLLECT_BINARY_TAG) { //collect binary from one of the workers.
      //Load the file name, assume its size is not too big, and send it to master node
      string fn = strCmd;
      //Load the file
      data.load(fn);
      
      ///for debug, make a copy of the file
      //string fn_base = get_basename_from_path(fn); char szt[1024];
      //sprintf(szt,"Scp <%s >/data/sep/zyang03/proj/RMO/tmp2/%s",fn.c_str(),fn_base.c_str());
      //PRV(szt); VERIFY(0==system(szt));

      long long count = data.domain().get_n123(); 
      // N = per_blk*nblk+nleft;//Send the binary to root
      float* pSend = const_cast<float*>(data.dataPtr());
      clk_wrker.reset("send binary start", &str_clkmsg); clk_wrker.go();
      world.collect_binary(0,irank,pSend,MPI_FLOAT,count);
      //clk_wrker.pause(); clk_wrker.check("send binary finish",&str_clkmsg);
    } else if (stat.MPI_TAG == REDUCE3DTAG) {
      VERIFY(0); //In 3D I don't use this now.
      //Load the file name
      string fn = strCmd;
      data.load(fn);
      //send the size info of the binary to the master node , number of elements per blk and nblk, MPI cannot handle very large msg size.
      //N = per_blk*nblk+nleft;
      //Load the file
      per_blk = ne[0]; nblk = ne[1]; nleft = ne[2];
      exeMPI(MPI_Bcast(ne,3,MPI_UNSIGNED,0,MPI_COMM_WORLD));

      long long count = per_blk*nblk+nleft;
      if (count!=data.domain().get_n123()) {
        SEEV2(count,data.domain().get_n123());
        VERIFY(0);
      }
      float* pSend = const_cast<float*>(data.dataPtr());
      for (unsigned int iblk=0; iblk<nblk; iblk++) {
        exeMPI(MPI_Reduce(pSend,NULL,per_blk,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
        pSend += per_blk;
      }
      //the last remaining block
      exeMPI(MPI_Reduce(pSend,NULL,nleft,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
    } else {
      VERIFY(0);
    }
  }

  //do the clean up, remove all /tmp/ files.
  string str_cleanup = "find /tmp/ -maxdepth 1 -type f -user zyang03 -exec rm {} \\;" ;
  int stat = 0;
  stat = system(str_cleanup.c_str());
  if (stat!=0) {
    PRV(str_cleanup);
  }
  return 0;
}

//for root process, due to the constrain of MPI_Reduce, din and dout cannot be the same.
int mpiY::ReduceCube_MPI(cubeDataF& dout, PosInt iAxis_per_blk, const branchFileName& bf_in) {
  mpiY::communicator world;
  int nrank = world.size();
  string strCmd;
  //BroadCast the size of the file, number of elements
  long long count;
  const hypercube& shp = dout.domain();
  
  count = dout.domain().get_n123();
  for (int ir=1;ir<nrank;ir++) {
    strCmd = bf_in.fn_at(ir);
    world.send(ir,REDUCETAG,strCmd);
  }
  LOG("Begin MPI_BCast in Reduce");
  unsigned int ne[3];
  int ndim = shp.get_ndim(); VERIFY(iAxis_per_blk<=ndim);

  if (0) { //we can choose the blk based on the model shape
    ne[0] = shp.get_n23(1,iAxis_per_blk);
    ne[1] = 1; if (iAxis_per_blk<ndim) {ne[1] = shp.get_n23(iAxis_per_blk+1,ndim);}
    ne[2] = 0;
    assert(count == ne[0]*ne[1]);
  } else { //or just use 1M blocks as kind of optimum
    ne[0] = MPI_BIN_MSG_SIZE/sizeof(float); //64KB is the magic number for cees-cluster
    ne[1] = count/ne[0];
    ne[2] = count % ne[0];
  }
  VERIFY( (ne[0]*ne[1]+ne[2]) == count);
  if (ne[0]<4e3 || ne[0]>65e3) { //if the msg size is not between 4k~64k, then it might hurt the performance 
    prt("blk size in ReduceCube_MPI might not be optimal, size is %d Bytes\n", ne[0]*4);
  }

  exeMPI(MPI_Bcast(ne,3,MPI_UNSIGNED,0,MPI_COMM_WORLD));
  LOG("Finish MPI_BCast in Reduce");
  Yang::timer clk;
  
  float* pRecv = dout.dataPtr(); dout.zero();
  std::valarray<float> arrSend(ne[0]); arrSend = 0.f; //dummy buffer

  for (unsigned int iblk=0; iblk<ne[1]; iblk++) {
    exeMPI(MPI_Reduce(va_ptr(arrSend),pRecv,ne[0],MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
    pRecv += ne[0];
  }
  exeMPI(MPI_Reduce(va_ptr(arrSend),pRecv,ne[2],MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));

  clk.check("reduce finish");
  return 0;
}

communicator::communicator() {
  exeMPI(MPI_Comm_rank(MPI_COMM_WORLD,&irank_));
  exeMPI(MPI_Comm_size(MPI_COMM_WORLD,&nrank_));
  
  int membersize = 0;
  //pack the string size
  exeMPI(MPI_Pack_size(1,MPI_INT,MPI_COMM_WORLD,&membersize));
  bufSize_ = membersize;
  exeMPI(MPI_Pack_size(sz_size,MPI_CHAR,MPI_COMM_WORLD,&membersize));
  bufSize_ += membersize;
  buffer_ = new char[bufSize_];
}

int communicator::send(int dst,int tag,const string& val) {
  //send using string pack
  szPack_.length_ = val.length()+1; //'\0' ended c string
  VERIFY(szPack_.length_<=sz_size);
  strcpy(szPack_.sz_,val.c_str());
  
  //int pos = 0;
  //exeMPI(MPI_Pack(&szPack_.length_,1,MPI_INT,buffer_,bufSize_,&pos,MPI_COMM_WORLD));
  //exeMPI(MPI_Pack(&szPack_.sz_,1024,MPI_CHAR,buffer_,bufSize_,&pos,MPI_COMM_WORLD));
  //exeMPI(MPI_Send(buffer_,pos,MPI_PACKED,dst,tag, MPI_COMM_WORLD));
  if (this->rank()==0) {
    prt("send cmd_str from %d to %d: %s\n ", rank(),dst,szPack_.sz_);
  }
  exeMPI(MPI_Send(szPack_.sz_,szPack_.length_,MPI_CHAR,dst,tag, MPI_COMM_WORLD));
  return 0;
}

const MPI_Status& communicator::recv(int src,int tag, string& val) {
  //int pos, msgsize;
  //exeMPI(MPI_Recv(buffer_,bufSize_,MPI_PACKED,src,tag,MPI_COMM_WORLD,&status));
  //exeMPI(MPI_Get_count(&status,MPI_PACKED,&msgsize));
  //exeMPI(MPI_Unpack(buffer_,msgsize,&pos,&szPack_.length_,1,MPI_INT,MPI_COMM_WORLD));
  //exeMPI(MPI_Unpack(buffer_,msgsize,&pos,szPack_.sz_,1024,MPI_CHAR,MPI_COMM_WORLD));
  exeMPI(MPI_Recv(buffer_,bufSize_,MPI_CHAR,src,tag,MPI_COMM_WORLD,&MPI_stat_));
  //Now put it back to string.
  strcpy(szPack_.sz_,buffer_);
  val = szPack_.sz_;
  if (this->rank()==0) {
    prt("Recv cmd_str from %d to %d: %s\n ", MPI_stat_.MPI_SOURCE,this->rank(),val.c_str());
  }
  return MPI_stat_;
}

int mpiY::communicator::collect_binary(int dst, int src, void* array, MPI_Datatype dataType, long long cnt) {
  bool bSender = false;
  int irank = rank();
  if (irank==src) {bSender=true;}
  else if (irank==dst) {bSender=false;}
  else {VERIFY(0);}
  
  int esize; long long n123;
  n123 = 0;
  exeMPI(MPI_Type_size(dataType, &esize));
  //then check the size and recv the buffer
  int ne[3]; unsigned int per_blk,nblk,nleft;
  if (!bSender) { //receiver side
    MPI_Status status;
    exeMPI(MPI_Recv(ne,3,MPI_UNSIGNED,src,BINARY_SIZE_TAG,MPI_COMM_WORLD,&status));
    n123 = ne[0]*ne[1]; n123+=ne[2];
    if(cnt != n123) {
      PRV4(n123,ne[0],ne[1],ne[2]);
      PRV(cnt);
      VERIFY(0);
    }
    per_blk = ne[0]; nblk=ne[1]; nleft=ne[2];
    //Do the actual receive
    byte* pRecv = (byte*)array;
    for (unsigned int iblk=0; iblk<nblk; iblk++) {
      exeMPI(MPI_Recv(pRecv,per_blk,dataType,src,BINARY_CONTENT_TAG,MPI_COMM_WORLD,&status));
      pRecv += per_blk*esize;
    }
    //the last remaining block, then finish
    if (nleft>0) {
      exeMPI(MPI_Recv(pRecv,nleft,dataType,src,BINARY_CONTENT_TAG,MPI_COMM_WORLD,&status));
    }
  } else { //being a sender
    per_blk = MPI_BIN_MSG_SIZE/esize;
    nblk    = cnt/per_blk;
    nleft   = cnt%per_blk;
    ne[0] = per_blk; ne[1] = nblk;  ne[2] = nleft;
    //send the transfer size information to the root
    exeMPI(MPI_Send(ne,3,MPI_UNSIGNED,dst,BINARY_SIZE_TAG,MPI_COMM_WORLD));
    //Do the actually send
    byte* pSend = (byte*)array;
    for (unsigned int iblk=0; iblk<nblk; iblk++) {
      exeMPI(MPI_Send(pSend,per_blk,dataType,dst,BINARY_CONTENT_TAG,MPI_COMM_WORLD));
      pSend += per_blk*esize;
    }
    //the last remaining block, then finish
    if (nleft>0) {
      exeMPI(MPI_Send(pSend,nleft,dataType,dst,BINARY_CONTENT_TAG,MPI_COMM_WORLD));
    }
  }
  return 0;
}

