#ifndef __MY_MPI_H_INCL__
#define __MY_MPI_H_INCL__

#ifndef USE_MPI
#error
#else
//Tag for MPI master/slave model
#define WRKTAG  120
#define EXITAG  121
#define IDLETAG 122
#define FINISHTAG 123
#define REDUCETAG 124
#define REDUCE3DTAG 125
#define COLLECT_BINARY_TAG 126 //Collect a binary file from the workers.
#define BINARY_SIZE_TAG 127 //Collect a binary file from the workers.
#define BINARY_CONTENT_TAG 128 //Collect a binary file from the workers.

#include <mpi.h>
extern "C" {
  #include "mpi_sep.h"
}
#include "decl_fwd.h"


// TODO: your code here
#define exeMPI(x) (VERIFY(MPI_SUCCESS == (x)));

namespace mpiY{
  class environment {
  public:
    environment(int* p_argc, char*** p_argv);
    virtual ~environment() {
      MPI_Finalize();
    }
  };
  
  class slaveModeMPI {
  public:
    int on();
  };
  
  class slaveModeMPI3d {
  public:
    int on();
  };

  struct string_pack{
    unsigned int length_;
    char sz_[1024];
  };
  
  class communicator {
  public:
    communicator();
    virtual ~communicator() {
      delete[] buffer_;
    }
    
    int rank()  {return irank_;}
    int size()  {return nrank_;}
    int barrier() {exeMPI(MPI_Barrier(MPI_COMM_WORLD)); return 0;}
    
    int send(int dst,int tag,const string& val); //send string
    const MPI_Status& recv(int src,int tag, string& val); //send string
    
    //transfer a binary array from src to dst.
    int collect_binary(int dst, int src, void* array, MPI_Datatype dataType, long long cnt);
  private:
    int irank_,nrank_;
    string_pack szPack_;
    enum {sz_size=1024} ;
    char* buffer_; unsigned int bufSize_;
    MPI_Status MPI_stat_;
  };

  int ReduceCube_MPI(Yang::cubeDataF& dout, fwk::PosInt iAxis_per_blk, const Yang::branchFileName& bf_in);
} //end namespace

#endif

#endif // __MY_MPI_H_INCL__

