#ifndef YANG_TIMER_H
#define YANG_TIMER_H

#include "stdio.h"
#include "stdlib.h"
#include <sstream>
#include <sys/time.h>

namespace Yang {
  //return time difference in seconds.
  inline double time_diff_secs(const struct timeval& t_cur,const struct timeval& t_prev) {
    double sec = t_cur.tv_sec - t_prev.tv_sec;
    double useconds = t_cur.tv_usec - t_prev.tv_usec;
    double mtime = sec + useconds*1e-6;
    return mtime;
  }

  class timer{
  public:
    //timer () {gettimeofday(&t_last_,NULL); sz_chkpt_ = "Clk created"; sz_name_ = "";}
    timer(const string chkpt="Clk created", const string& timer_name=string("") ) {
      //may need date as well if program runs longer than 24 hrs.
      gettimeofday(&t_last_,NULL); //get the current time stamp
      sz_chkpt_ = chkpt; sz_name_ = timer_name;
      accu_time_sec_ = 0;
      bCounting_ = false;
    }
    
    void start() {
      VERIFY(!bCounting_);
      this->reset("clk started");
      this->go();
    }

    int pause() {
      int ret = -1;
      if (bCounting_) {
        struct timeval t_cur;
        gettimeofday(&t_cur,NULL);
        accu_time_sec_ += time_diff_secs(t_cur,t_last_);
        t_last_ = t_cur;
        bCounting_ = false;\
        ret = 0;
      } //otherwise do nothing
      return ret;
    }
    
    int go() {
      int ret = -1;
      if (!bCounting_) {
        bCounting_ = true;
        gettimeofday(&t_last_,NULL);
        ret = 0;
      } //otherwise do nothing
      return ret;
    }

    string chime(double mtime, const string& chkpt) {
      sprintf(sz_tmp_,"Timer:%s, reads %.3f secs from chk pt %s, to pt %s  ",sz_name_.c_str(),mtime,sz_chkpt_.c_str(),chkpt.c_str());
      return string(sz_tmp_);
    }

    double reset(const string& chkpt, string* p_str_msg=NULL) {
      struct timeval t_cur;
      gettimeofday(&t_cur,NULL);
      double mtime = accu_time_sec_;
      if (bCounting_) {
        mtime += time_diff_secs(t_cur,t_last_);
      }
      string str_chime = this->chime(mtime,chkpt);
      if (p_str_msg == NULL) {
        fprintf(stderr,"%s, timer reset.\n",str_chime.c_str());
      } else {
        *p_str_msg = str_chime + ",timer reset.\n";
      }
      //reset the counter and other states
      sz_chkpt_ = chkpt; t_last_ = t_cur;
      accu_time_sec_ = 0; bCounting_ = false;
      return mtime;
    }

    double check (const string& chkpt, string* p_str_msg=NULL) {
      double mtime = accu_time_sec_;
      if (bCounting_) {
        struct timeval t_cur;
        gettimeofday(&t_cur,NULL);
        mtime += time_diff_secs(t_cur,t_last_);
      }
      string str_chime = this->chime(mtime,chkpt);
      if (p_str_msg == NULL) {
        fprintf(stderr,"%s. \n",str_chime.c_str());
      } else {
        *p_str_msg = str_chime;
      }
      return mtime;
    }

    ~timer() {
      double mtime = accu_time_sec_;
      if (bCounting_) {
        struct timeval t_cur;
        gettimeofday(&t_cur,NULL);
        mtime += time_diff_secs(t_cur,t_last_);
      }
      cerr << this->chime(mtime,"timer's end \n");
    }

  private:
    struct timeval t_last_;
    double accu_time_sec_; //accumulated time in seconds;
    string sz_chkpt_,sz_name_;
    bool bCounting_;
    char sz_tmp_[1024];
  };
} //end namespace Yang

#endif
