#include "stdafx.h"
#include "dmRsf.hpp"
#include "param_func.h"
#include "param_clib.h"
#include "off2ang3d.param.h"

using namespace Yang;

//inshape could be (hx,hy,kmx,kmy,kz) or (gamma,azim,kmx,kmy,kz), shapezxy is a 3D cube that has z,x,y as axes
void Yang::read_param_off2ang3d_hxy_kxyz(param_off2ang3d* par, param_func* pf, const hypercube& inShape_kxyz, const hypercube& shapexyz, bool bForward) {
  //reorder the axis because of legacy reason.
  hypercube inShape_kzxy, shape_zxy;
  vecAxis axes_in   = inShape_kxyz.get_axes();
  vecAxis axes_zxy  = shapexyz.get_axes();
  axes_in[2] = inShape_kxyz.get_axis(5);
  axes_in[3] = inShape_kxyz.get_axis(3);
  axes_in[4] = inShape_kxyz.get_axis(4);

  if (axes_zxy.size()==3) { //In some scenario we can put in a dummy parameter for this one.
    axes_zxy[0] = shapexyz.get_axis(3);
    axes_zxy[1] = shapexyz.get_axis(1);
    axes_zxy[2] = shapexyz.get_axis(2);
  }
  inShape_kzxy.set_axes(axes_in);
  shape_zxy   .set_axes(axes_zxy);
  read_param_off2ang3d_hxy_kzxy(par, pf, inShape_kzxy, shape_zxy, bForward);
  return;
}

//Different axis orders, inshape could be (hx,hy,kz,kmx,kmy) or (gamma,azim,kz,kmx,kmy), shapezxy is a 3D cube that has z,x,y as axes
void Yang::read_param_off2ang3d_hxy_kzxy(param_off2ang3d* par, param_func* pf, const hypercube& inShape, const hypercube& shapezxy, bool bForward) {
  par->boff2ang               = bForward;
  bool  bOff2ang              = par->boff2ang;
  
  par->bVerb                  = pf->get_bool("verb", true);
  par->create_coplanar_stack  = pf->get_bool("create_coplanar_stack",false);
  par->create_coplanar_ang    = pf->get_bool("create_coplanar_ang",false);
  par->create_stack           = pf->get_bool("create_stack",false);
  par->max_memory             = pf->get_int("max_memory",1000);

  par->apply_jac_gamma        = pf->get_bool("apply_jac_gamma",false);
  par->apply_jac_azim         = pf->get_bool("apply_jac_azim",false);

  par->number_aliases         = pf->get_int("number_aliases",0);
  par->bSign_khy              = pf->get_bool("sign_khy",true);
  par->debug1                 = pf->get_bool("debug1",true);  
  par->debug2                 = pf->get_bool("debug2",true);
  
  par->shape_zxy = shapezxy; //the original z,x,y axis info.
  //Check the dimension of the input cube
  if (bOff2ang) { //input is off image
    par->shapeOff = inShape;
  } else {
    par->shapeAng = inShape;
    par->ax_azim  = par->shapeAng.get_axis(2);
    par->ax_gamma = par->shapeAng.get_axis(1);
    //deterimine the gamma,azim->khx,khy
    int a2omappingmethod = pf->get_int("a2h_mapping_method",2);
    PRV(a2omappingmethod);
    par->bDoFTsmoothing = false;
    switch (a2omappingmethod) {
    case 0:
      par->m_mapping_method = gkm_NN;
      break;
    case 1:
      par->m_mapping_method = gkm_ap_NN;
      par->bDoFTsmoothing = true;
      break;
    case 2:
      par->m_mapping_method = gkm_polyfill;
      par->bDoFTsmoothing = true;
      break;
    case 3:
      par->m_mapping_method = gkm_triTexMapping;
      break;
    default:
      errPrt("a2h_mapping_method parameter invalid!!\n");
      break;
    }
  }
  VERIFY(inShape.get_ndim()==5);

  //The axis order (hx,hy,kz,kmx,kmy) or (gamma,azim,kz,kmx,kmy)
  PRT_LOC; inShape.info("input shape in read_param_off2ang3d_hxy_kzxy");
  
  stdVecInt ns = inShape.get_ns();
  int nhx,nhy; nhx = ns[0]; nhy = ns[1];
  
  par->n_hx_pad = pf->get_int("n_hx_pad",nhx);
  par->n_hy_pad = pf->get_int("n_hy_pad",nhy);
  
  int n; float o,d;
  vecAxis axes;
  float dx = shapezxy.get_d(2), dy = shapezxy.get_d(3);
  VERIFY(dx>0.1*dy && dx<10*dy); //(dx, dy should be close)

  if (bOff2ang) {
    float dhx = inShape.get_d(1), dhy = inShape.get_d(2);
    VERIFY(dhx>0.1*dhy && dhx<10*dhy); //(dhx, dhy should be close)
    //read in the output axis azim nod.
    n = pf->get_int("azim_n",1); d = pf->get_float("azim_d",1.); o = pf->get_float("azim_o",0.f);
    par->ax_azim.set(n,o,d,"azim","deg");
    // read in axis gamma info
    n = pf->get_int("gamma_n",1); d = pf->get_float("gamma_d",1.f); o = pf->get_float("gamma_o",0.f);
    par->ax_gamma.set(n,o,d,"gamma","deg");

    axes = inShape.get_axes();
    axes[0] = par->ax_gamma; axes[1] = par->ax_azim; //replace the first 2 axis
    par->shapeAng.set_axes(axes); //out shape
  } else {
    //should read the nhx and nhy axis
    axis ax_hx,ax_hy;
    n = pf->get_int(csz_img_nhx); d = pf->get_int(csz_img_dhx)*dx; o = pf->get_int(csz_img_ohx)*d;
    ax_hx.set(n,o,d,"hx","m");
    n = pf->get_int(csz_img_nhy); d = pf->get_int(csz_img_dhy)*dy; o = pf->get_int(csz_img_ohy)*d;
    ax_hy.set(n,o,d,"hy","m");
    axes = inShape.get_axes();
    axes[0] = ax_hx; axes[1] = ax_hy; //replace the first 2 axis with the orignal hx,hy
    par->shapeOff.set_axes(axes);
    float azim_d = par->shapeAng.get_d(2);
    VERIFY(azim_d > 0.5 && azim_d < 50.0f); //The azimuth sampling ratio should be in a reasonable range.
  }
  
  if (bOff2ang) {
    if (par->create_coplanar_stack || par->create_stack)  //this means stacking over reflection angle, resulting 4D cube by (azim,kz,kmx,kmy);
    {
      par->tag_coplanar_stack = pf->get_string("tag_coplanar_stack","coplanar_stack");
      
      axes = inShape.return_axes(2,5);
      axes[0] = par->ax_azim;
      par->shapeAng.set_axes(axes);
    }
  
    if (par->create_coplanar_ang)
    {
      par->tag_coplanar_ang = pf->get_string("tag_coplanar_ang","coplanar_ang");
      par->ang_2d           = pf->get_bool("ang_2d",false);
    }
    par->shapeAng.info("angle img shape in read_param_off2ang3d_hxy_kzxy");
  } else {
    par->shapeOff.info("off img shape in read_param_off2ang3d_hxy_kzxy");
  }

  return;
}

