#include "stdafx.h"
#include <sstream>
#include <fstream>
#include "decl_fwd.h"
#include <ctype.h>

#include "waveOpMPI.h"
#include "misc.h"

using namespace Yang;
using namespace std;

shotFilesInfo::shotFilesInfo():m_nShots(0),m_nSuperShots_(0),bMarineGeom_(true) {
  array_assign(bColFlags_,false,6);
}

void shotFilesInfo::getImagingDomainFlags(bool& xmin, bool& xmax, bool& ymin, bool& ymax, bool& zmin, bool& zmax) const {
  xmin = bColFlags_[0]; xmax = bColFlags_[1];
  ymin = bColFlags_[2]; ymax = bColFlags_[3];
  zmin = bColFlags_[4]; zmax = bColFlags_[5];
}

void shotFilesInfo::getDataAndSrcFile(string& fn_data, string& fn_src, int iShot) const {
  fn_data = this->fnames_shots_[iShot];
  if (this->bSameSrcForAllShots_) {fn_src = fnames_src_[0];}
  else {fn_src = fnames_src_[iShot];}
}

void shotFilesInfo::getImagingDomain(float& xmin, float& xmax, float& ymin, float& ymax, float& zmin, float& zmax, int iShot) const {
  if (bColFlags_[0]) xmin=xminv_[iShot];
  if (bColFlags_[1]) xmax=xmaxv_[iShot];
  if (bColFlags_[2]) ymin=yminv_[iShot];
  if (bColFlags_[3]) ymax=ymaxv_[iShot];
  if (bColFlags_[4]) zmin=zminv_[iShot];
  if (bColFlags_[5]) zmax=zmaxv_[iShot];
}

//Get the domain that covers the domain of all shots
void shotFilesInfo::getGlobalImgDomain(float& xmin, float& xmax, float& ymin, float& ymax, float& zmin, float& zmax) const {
  if (bColFlags_[0]) xmin=*std::min_element(xminv_.begin(),xminv_.end());
  if (bColFlags_[1]) xmax=*std::max_element(xmaxv_.begin(),xmaxv_.end());
  if (bColFlags_[2]) ymin=*std::min_element(yminv_.begin(),yminv_.end());
  if (bColFlags_[3]) ymax=*std::max_element(ymaxv_.begin(),ymaxv_.end());
  if (bColFlags_[4]) zmin=*std::min_element(zminv_.begin(),zminv_.end());
  if (bColFlags_[5]) zmax=*std::max_element(zmaxv_.begin(),zmaxv_.end());
}

std::string trim(const std::string& str, const std::string& whitespace=" ") {
  std::size_t strBegin = str.find_first_not_of(whitespace);
  if (strBegin == std::string::npos) {
    return "";
  }
  std::size_t strEnd = str.find_last_not_of(whitespace);
  std::size_t strRange = strEnd - strBegin + 1;
  return str.substr(strBegin,strRange);
}

int shotFilesInfo::parsefile(const string& fname, const boost::shared_ptr<hypercube>& vmodShape, int ishot_beg, int ishot_end) {
  vmodShape_2way_ = vmodShape;
  fname_desc_ = fname;
  fstrm_desc_.reset(new std::ifstream(fname_desc_.c_str(),std::ifstream::in));
  VERIFY(fstrm_desc_->is_open());

  //then parse the data_desc_file
  string str_ln,str_tmp,str_val,str_key; stringstream ss;
  string fn_shot;
  int iline = 0; int nTotalShots,nSshots, ncol;
  std::vector<std::string> str_vec;
  stdVecFloat* pField_vec[6]; int ip=0, ifiles=0;
  bSameSrcForAllShots_ = true;
  while (std::getline(*fstrm_desc_,str_ln)) { //if read not end
    //skip empty lines, remove the trailing spaces
    str_ln = trim(str_ln," ");
    if (str_ln.empty() || str_ln[0]=='\n') {
      continue;
    }
    iline++;
    //parse this line
    ss.clear(); //clear the status for stringstream, this is important, otherwise ss only can parse one single line
    ss.str(str_ln);
    //first line should tell the total number of files in this run, and then whether we use same(single) src file for All shots
    if (iline==1) {
      ss >> nTotalShots;
      //determine how many shots we need to extract based on the ishot_beg and ishot_end info
      ishot_beg_ = ishot_beg; 
      if (ishot_end==-1) {
        ishot_end_ = nTotalShots;
      } else {
        ishot_end_ = ishot_end;
      }
      m_nSuperShots_ = ishot_end_ - ishot_beg_;
      //PRV2(m_nSuperShots_,nTotalShots);
      VERIFY(m_nSuperShots_>=0 && m_nSuperShots_<=nTotalShots);
      nSshots = m_nSuperShots_;
      //allocate memory
      fnames_shots_.resize(nSshots);  fnames_src_.resize(nSshots);
      ss >> str_tmp;
      VERIFY(parseCmdlineOption(str_tmp, str_key,str_val)==0);
      if (str_key.compare("bMarineGeom")!=0) {
        PRV(nSshots); PRV2(str_tmp,str_val);
        VERIFY(0);
      }
      //bMarineGeom option is either y or n
      if (str_val[0]=='y') {
        bMarineGeom_ = true;
      } else if (str_val[0]=='n') {
        bMarineGeom_ = false;
      } else {VERIFY(0);}
      
      ss >> str_tmp;
      VERIFY(parseCmdlineOption(str_tmp, str_key,str_val)==0);
      if (str_key.compare("singleSrcFileForAllShots")!=0) {
        PRV(nSshots); PRV2(str_tmp,str_val);
        VERIFY(0);
      }
      //This option is either y or n
      if (str_val[0]=='y') {
        bSameSrcForAllShots_ = true;
        //read in the file name again
        ss >> str_val;
        fnames_src_.resize(1); fnames_src_[0] = str_val;
      } else if (str_val[0]=='n') {
        bSameSrcForAllShots_ = false;
        VERIFY(0); //for comsou, not impl yet.
      } else {VERIFY(0);}
    } else if (iline==2) { //the 2nd column defines the columns meaning, currently just xmin,xmax,ymin,ymax,zmin,zmax
      //split the ss into strings
      std::copy(istream_iterator<string>(ss),istream_iterator<string>(),back_inserter<vector<string> >(str_vec));
      ncol = str_vec.size();
      PRVEC(str_vec,ncol);
      ip = 0;
      for (int i=0; i<ncol; i++) { //parse each column
        str_tmp = str_vec[i];
        if (i==0) {
          VERIFY(str_tmp.compare("fname_data")==0); //first col has to be filename
          continue;
        }
        if (str_tmp.compare("fname_src")==0) {
          VERIFY(bSameSrcForAllShots_ == false); VERIFY(i==1); //has to be in the 2nd column
        } else if (str_tmp.compare("xmin")==0) {
          bColFlags_[0]=true; pField_vec[ip] = &xminv_; ip++;
        } else if (str_tmp.compare("xmax")==0) {
          bColFlags_[1]=true; pField_vec[ip] = &xmaxv_; ip++;
        } else if (str_tmp.compare("ymin")==0) {
          bColFlags_[2]=true; pField_vec[ip] = &yminv_; ip++;
        } else if (str_tmp.compare("ymax")==0) {
          bColFlags_[3]=true; pField_vec[ip] = &ymaxv_; ip++;
        } else if (str_tmp.compare("zmin")==0) {
          bColFlags_[4]=true; pField_vec[ip] = &zminv_; ip++;
        } else if (str_tmp.compare("zmax")==0) {
          bColFlags_[5]=true; pField_vec[ip] = &zmaxv_; ip++;
        } else {
          PRV(str_tmp);
          VERIFY(0);
        }
      }
      PRVEC(bColFlags_,6); PRV(nSshots);
      //after that allocate memories
      xminv_.resize(nSshots),xmaxv_.resize(nSshots);
      yminv_.resize(nSshots),ymaxv_.resize(nSshots);
      zminv_.resize(nSshots),zmaxv_.resize(nSshots);
    } else { //parse the remaining lines
      //first should be filename, then the xmin, xmax, and then ymin, and ymax
      if (ifiles-ishot_beg_ == nSshots) {
        break; //read enough files, discard the rest.
      }
      if (ifiles >= ishot_beg_) {
        ss >> str_tmp; fnames_shots_[ifiles-ishot_beg_] = str_tmp; 
        //PRV(str_tmp);
        if (!bSameSrcForAllShots_) {
          ss >> str_tmp; fnames_src_[ifiles-ishot_beg_] = str_tmp;
        }
        //scan in the target domain dimensions.
        for (int ii=0; ii<ip; ii++) {
          ss >> (*pField_vec[ii])[ifiles-ishot_beg_];
        }
      }
      ifiles++;
      //ss.rdbuf()->in_avail() == 0; //check if buffer is empty now.
    }
  } //finished parsing
  //PRVEC(xminv_,nSshots);PRVEC(yminv_,nSshots);PRVEC(zminv_,nSshots); PRVEC(xmaxv_,nSshots);PRVEC(ymaxv_,nSshots);PRVEC(zmaxv_,nSshots);
  if (ifiles-ishot_beg_ < nSshots) { //the num of lines of file is not correct
    PRV2(ifiles,nSshots); 
    VERIFY(0);
  }
  return 0;
}

