#include "stdafx.h"
#include "cubeData.hpp"
#include "point3d.hpp"
#include "sep_params.h"
#include "cubeAlgo.h"
#include "cubeAlgo2.h"

using namespace Yang;
using namespace std;

   
//Apply Jon's Marine data gain t^2/t0 to the data
int main (int argc, char *argv[])
{
  sep_params* pars = new sep_params(argc,argv);
  cubeData<float> inc,outc,t0c,t0c_after; //after interpolation
  int nt,ntrc;
  
  //read the input
  inc.load("in");
  hypercube shpIn = inc.domain();
  nt    = shpIn.get_n(1);
  axis ax_t = shpIn.get_axis(1);

  hypercube shp_t0(shpIn.return_compatible(2,shpIn.get_ndim()));
  
  
  //Might need to do multiple panels equalization
  int ntrc23 = shpIn.get_n23(2,shpIn.get_ndim()) / shpIn.get_n(2);
  ntrc  = shpIn.get_n23(); 
  int ntrcPerPanel = shpIn.get_n(2);

  t0c.alloc(shp_t0); t0c_after.alloc(shp_t0); t0c_after = 1.0f; //by default, needs to be positive!!! Otherwise ax_t.get_offset will create error
  bool verb = pars->get_bool("verb",true);
  float scThres = pars->get_float("thres_sc",10.0f);
  if (verb) {PRV2(nt,ntrc);}
  
  //First find all t0, calculate each trace's RMS, search for the first peak, and then take it as the t0
  int ir=0, nCompWindow = 10; //find the maximum that among the window of (2*10+1) samples
  hypercube shpnt(nt);
  cubeData<float> trcc(shpnt);
  
  float rmsnz,thres; int it;
  stdVecFloat inCoord,outCoord, inVec, outVec;
  rmsnz = inc.RMS_NZ(1e-5);
  thres = rmsnz * scThres;
  PRV(thres);
  VERIFY(ntrc23*ntrcPerPanel == ntrc);
  int ind = 0, trcind=0;
  for (int iR=0; iR<ntrc23; iR++) {
    inCoord.clear(); outCoord.clear();
    inVec.clear(); outVec.clear();
    for (ir=0; ir<ntrcPerPanel; ir++) {
      inc.seq_grab(&trcc);
      
      //search the maximum as t increases
      float fmax; int curmax;
      curmax = 0; fmax = 0.0f;
      bool bReachThres = false;
      for (it=0; it<nt; it++) {
        if (fmax < fabs(trcc(it))) {
          fmax = fabs(trcc(it));
          curmax = it;
          if (fmax > thres) {
            bReachThres = true;
            break;
          }
        }
      }
      trcind = iR*ntrcPerPanel+ir;
      if (bReachThres) {
        t0c(trcind) = ax_t.get_coord(curmax); t0c_after(trcind) = t0c(trcind);
        inCoord.push_back(float(ir)); inVec.push_back(t0c(trcind));
      } else {
        t0c(trcind) = -1.0f; //means could not find any meaningful value
        outCoord.push_back(float(ir));  outVec.push_back(0.0f);
      }
    } //end ir
    //VERIFY(inCoord.size() > ntrc/2);
    //Now do some interpolation of t0 arrays for places where it cannot find, here just use std input
    if (inVec.size()>0) {
      linInterp1d(&outVec[0],outCoord,&inVec[0],inCoord,true);
      for (ind=0; ind<outCoord.size(); ind++) {
        t0c_after[iR*ntrcPerPanel+(outCoord[ind]+0.5f)] = outVec[ind];
      }
    } else {} //this is an empty trace, then use default value
  } //end for iR=ntrc23
  t0c.dump("t0_pre"); t0c_after.dump("t0");
  //Now apply the gain
  outc.alloc(shpIn);
  VERIFY(t0c_after.maxval()<9.0f);
  VERIFY(t0c_after.minval()>0.99f);

  float t0,t; int it0; float *pIn,*pOut;
  for (ir=0; ir<ntrc; ir++) {
    t0 = t0c_after[ir];
    pIn  = inc.pval+ir*nt;
    pOut = outc.pval+ir*nt;
    it0 = ax_t.get_offset(t0);
    //if smaller than t0, then apply the gain t
    for (it=0; it<std::min(nt,it0+1); it++) {
      pOut[it] = ax_t.get_coord(it) * pIn[it];
    }
    //apply the gain t^2/t0 
    for (int it=it0;it<nt; it++) {
      t = ax_t.get_coord(it);
      pOut[it] = t*t/t0*pIn[it];
    }
  }
  outc.dump("out");
  return 0;
}

