#undef NDEBUG
#include <assert.h>

#include "stdafx.h"
#include "sep_params.h"
#include "decl_fwd.h"
#include "cubeData.hpp"

using namespace Yang;
using namespace fwk;

///Generate waterbottom traveltime two-way, based on the velocity profile.
int main(int argc, char** argv) {
  boost::shared_ptr<sep_params> pars(new sep_params(argc,argv));

  //in is the input velocity model.
  cubeDataF vel;
  vel.load("in",3);
  const hypercube& shpIn = vel.domain();
  //int ndim = shpIn.get_ndim();

  float v_water   = pars->get_float("v_water",1500.0f);
  float v_delta   = pars->get_float("v_delta",50.0f);
  shpIn.info("shpIn");
  const axis& ax_z = shpIn.get_axis(1);
  const axis& ax_x = shpIn.get_axis(2);
  const axis& ax_y = shpIn.get_axis(3);
  
  bool b2D = (ax_y.n==1);

  vecAxis axes(2);  axes[0] = ax_x; axes[1] = ax_y; hypercube shpOut(axes);
  cubeDataF WB_tt(shpOut);
  
  int iz;
  for (int iy=0; iy<ax_y.n; iy++) {
    for (int ix=0; ix<ax_x.n; ix++) {
      float tsum = 0.0f;
      //To detect the water bottom depth, detect the first velocity jump from 1500m to a bigger value.
      for (iz=1;iz<ax_z.n;iz++) {
        tsum += ax_z.d/vel(iy,ix,iz-1);
        if (vel(iy,ix,iz)-vel(iy,ix,iz-1)>v_delta) {
          break;
        }
      }
      if (iz==ax_z.n) {
        errPrt("Could not found water bottom at iy=%d, ix=%d !!", ix,iy);
      }
      WB_tt(iy,ix) = tsum*2;
    }
  }
  WB_tt.dump("out");
  return 0;
}

