#undef NDEBUG
#include <assert.h>

#include "stdafx.h"
#include "sep_params.h"
#include "decl_fwd.h"
#include "cubeData.hpp"

using namespace Yang;
using namespace fwk;

///Mute program intially from SEPLIB
///Mute common-shot gather, assume vmute is constant (water velocity), but tmute might be differnt because the waterbottom is not always flat
void mute_hyper2d(float tmute0, float vmute, float shotLoc, const axis& ax_recv, const axis& ax_t, float* data2D, const cubeDataF* pWB_traveltime_2way = NULL, bool bApplyMGain=false) {
  float t,t0,t_mute,th,hx;
  float tramp=0.25;

  bool bVariable_tmute0 = (pWB_traveltime_2way != NULL);
  
  axis ax_WB;
  if (bVariable_tmute0) {ax_WB = pWB_traveltime_2way->domain().get_axis(1);}

  int nTrc; float o_r,d_r,mx;
  ax_recv.get_nod(nTrc,o_r,d_r);
  VERIFY(nTrc > 0);

  int   nt = ax_t.n; mx = 0.0f;
  for (int iTrc=0;iTrc<nTrc; iTrc++) {
    hx = ax_recv.get_coord(iTrc) - shotLoc;
    //cout <<"rx = " << rx << endl;
    th=hx/vmute;
    if (!bVariable_tmute0) {
      t_mute=sqrtf(tmute0*tmute0+th*th);
    } else { //variable tmute0, need to be decided from WaterBottomTravelTime
      //Find mid-pt location
      mx = shotLoc+hx*0.5f;
      int iwbx = ax_WB.get_offset(mx); VERIFY(iwbx>=0 && iwbx<ax_WB.n);
      t0 = (*pWB_traveltime_2way)(iwbx)+tmute0;
      t_mute = sqrtf(t0*t0+th*th);
    }

    //cerr <<"hx="<<hx <<sep<< "th="<<th<<" "<<endl;  cerr <<" t_mute = " << t_mute << endl;
    //after we get the tmute value, mute the trace with a short time window taper.
    for (int it=0;it<nt;it++) {
      t = ax_t.get_coord(it);
      if (t<t_mute){
        data2D[iTrc*nt+it]*=0.0f;
      }
      else if ( t-t_mute<=tramp) {
        data2D[iTrc*nt+it] *= (t-t_mute)/(tramp+1.e-10f); 
      }
      else {
        //data2D[iTrc*nt+it]*= 1.0f; //leave as is
      }
      if (bApplyMGain) { //Apply MGain, assume t_mute is the WB travel time.
        for (it=std::max(0,ax_t.get_offset(t_mute)-1);it<nt;it++) {
          t = ax_t.get_coord(it);
            data2D[iTrc*nt+it] *= (t*t/t_mute);
        }
      }
    }
  }
}

void mute_hyper3d(float tmute0, float vmute, float shotLocX, float shotLocY, const axis& ax_recvx, const axis& ax_recvy, const axis& ax_t, float* data3D, const cubeDataF* pWB_traveltime_2way = NULL, bool bApplyMGain=false) {
  float t,t_mute,thx,thy,hx,hy;
  float tramp=0.25;
  
  bool bVariable_tmute0 = (pWB_traveltime_2way != NULL);
  axis ax_WBx,ax_WBy;
  if (bVariable_tmute0) {
    ax_WBx = pWB_traveltime_2way->domain().get_axis(1);
    ax_WBy = pWB_traveltime_2way->domain().get_axis(2);
  }

  int   nt = ax_t.n, base = 0; float t0, mx,my;
  for (int iTrcY=0;iTrcY<ax_recvy.n; iTrcY++) {
    hy = ax_recvy.get_coord(iTrcY) - shotLocY;
    thy = hy/vmute;
    for (int iTrcX=0;iTrcX<ax_recvx.n; iTrcX++) {
      hx = ax_recvx.get_coord(iTrcX) - shotLocX;
      //cout <<"rx = " << rx << endl;
      thx = hx/vmute;
      
      if (!bVariable_tmute0) {
        t_mute=sqrtf(tmute0*tmute0+thx*thx+thy*thy);
      } else { //variable tmute0, need to be decided from WaterBottomTravelTime
        //Find mid-pt location
        mx = shotLocX+hx*0.5f; my = shotLocY+hy*0.5f;
        int iwbx = ax_WBx.get_offset(mx); VERIFY(iwbx>=0 && iwbx<ax_WBx.n);
        int iwby = ax_WBy.get_offset(my); VERIFY(iwby>=0 && iwby<ax_WBy.n);
        t0 = (*pWB_traveltime_2way)(iwby,iwbx)+tmute0;
        t_mute = sqrtf(t0*t0+thx*thx+thy*thy);
      }

      base = iTrcY*ax_recvx.n+iTrcX;
      //cerr <<"hx="<<hx <<sep<< "th="<<th<<" "<<endl;  cerr <<" t_mute = " << t_mute << endl;
      for (int it=0;it<nt;it++) {
        t = ax_t.get_coord(it);
        if (t<t_mute){ data3D[base*nt+it]*=0.0f;}
        else if ( t<=t_mute+tramp) { data3D[base*nt+it] *= (t-t_mute)/(tramp+1.e-10f); }
        else {//data3D[base*nt+it]*= 1.0f; //leave as is
        }
      }
      if (bApplyMGain) { //Apply MGain, assume t_mute is the WB travel time.
        for (int it=std::max(0,ax_t.get_offset(t_mute)-1);it<nt;it++) {
          t = ax_t.get_coord(it);
            data3D[base*nt+it] *= (t*t/t_mute);
        }
      }
    }
  }
  return;
}

int main(int argc, char** argv) {
  boost::shared_ptr<sep_params> pars(new sep_params(argc,argv));

  //in is the input data, should be common-shot gather.
  cubeDataF input;
  bool b3D  = pars->get_bool("b3D",false);
  if (!b3D) { //2D case, only 3 axes maximum
    input.load("in",3);
  } else {
    input.load("in",5);
  }
  
  const hypercube& shpIn = input.domain();
  //int ndim = shpIn.get_ndim();
  bool bOffset  = pars->get_bool("offset",true);
  bool bDoMGain = pars->get_bool("bDoMGain",false);
  float vmute   = pars->get_float("vmute");
  float tmute   = pars->get_float("tmute",0.0f);
  
  PRV(bDoMGain);
  const axis& ax_2 = shpIn.get_axis(2);
  const axis& ax_t    = shpIn.get_axis(1);
  float* pdata = input.dataPtr();
  
  ax_t.print_info("Time axis info");  ax_2.print_info("Receiver_x axis info");
  axis ax_shotx,ax_recvx,ax_shoty,ax_recvy;
  ax_recvx = ax_2;
  if (!b3D) {
    ax_shotx = shpIn.get_axis(3);
  } else {
    ax_recvy = shpIn.get_axis(3);
    ax_shotx = shpIn.get_axis(4);
    ax_shoty = shpIn.get_axis(5);
  }
  
  bool bUseWaterBotTravelTime = false;
  //check if we can extract a WB-TravelTime file
  cubeDataF WB_tt,*pWB_tt = NULL;
  bUseWaterBotTravelTime = tagExist("WB_tt");
  if (bUseWaterBotTravelTime) {
    WB_tt.load("WB_tt");
    pWB_tt = &WB_tt;
  } 
  
  if (!b3D) { //2D case
    for (int ish=0; ish<ax_shotx.n; ish++) {
      float* data2D = pdata + ish*ax_t.n*ax_2.n;
      float shotLoc = ax_shotx.get_coord(ish);
      if (bOffset) { //the axes are (t,g-s,s), marine geometry
        //convert the ax_h to ax_recv;
        ax_recvx.o = shotLoc + ax_2.o;
      } //otherwise, no need to change, the axes are (t,g,s), land geometry
      mute_hyper2d(tmute,vmute,shotLoc, ax_recvx, ax_t, data2D,pWB_tt,bDoMGain);
    }
  } else{ //3D case
    float *data3D = pdata;
    axis ax_rx,ax_ry;
    ax_rx = ax_recvx; ax_ry = ax_recvy;
    for (int ishy=0; ishy<ax_shoty.n; ishy++) {
      float shotLocY = ax_shoty.get_coord(ishy);
      if (bOffset) {ax_ry.o = ax_recvy.o + shotLocY;}
      for (int ishx=0; ishx<ax_shotx.n; ishx++) {
        float shotLocX = ax_shotx.get_coord(ishx);
        if (bOffset) { //the axes are (t,g-s,s), marine geometry
          //convert the ax_h to ax_recv;
          ax_rx.o = ax_recvx.o + shotLocX;
        } //otherwise, no need to change, the axes are (t,g,s), land geometry
        mute_hyper3d(tmute,vmute,shotLocX,shotLocY,ax_rx,ax_ry, ax_t,data3D,pWB_tt,bDoMGain);
        data3D  += ax_recvx.n*ax_recvy.n*ax_t.n;
      }
    }
  }
  input.dump("out");
  return 0;
}
