#include "stdafx.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "off2ang3d.h"
#include "sep_params.h"
#include "off2ang3d.param.h"

/**
 * Main application entry point
 * 
 * Off2ang 3d tool, algorithm based on Biondi and Tisserant's SEP report, SEP113. etc.
 * Format:
 * Off2ang3dB <in.H >out.H jac_gamma=y/n jac_beta=y/n [params] 
 * input is of format (hx,hy,kmx,kmy,kz), output is of 
 * (gamma,azim,kmx,kmy,kz) Other params pls see the 
 * read_param_off2ang3d function. 
 */

using namespace Yang;
int main (int argc, char *argv[])
{
  // TODO: Add application code here
  initCallStackTrace();
  init_omp(0); //only use one thread.
  
  //have to init par
	sep_params* pars = new sep_params(argc,argv);
  string szin = pars->get_string("input","");
  string szout = pars->get_string("output","");
  cerr << "szin:szout; " << szin << " : " << szout << endl;

  //Read in/out tag
  bool bNotDummy = true;;
  dmRsf_oc<cmplxf> dmIn("in",Yang::dm::IN,"input",bNotDummy);  dmRsf_oc<cmplxf> dmOut("out",Yang::dm::OUT,"output",bNotDummy);
  dmRsf<cmplxf>& ddmIn = dmIn; dmRsf<cmplxf>& ddmOut=dmOut;
  dmIn.init(NULL);
  bool boff2ang = pars->get_bool("bforward", true);

  hypercube inShape = dmIn.domain();
  //read in params
  param_off2ang3d par_op;
  read_param_off2ang3d_hxy_kxyz(&par_op, pars, inShape, hypercube(1,1,1),boff2ang); //the 4th argument is dummy in this case (only in this case).
  //set the output shape
  vecAxis axes_shpAng_xyz = par_op.shapeAng.get_axes();
  axes_shpAng_xyz[2] = par_op.shapeAng.get_axis(4);
  axes_shpAng_xyz[3] = par_op.shapeAng.get_axis(5);
  axes_shpAng_xyz[4] = par_op.shapeAng.get_axis(3);
  hypercube shpAng_xyz(axes_shpAng_xyz);
  dmOut.init(&shpAng_xyz);

  //Apply the transform
  if (boff2ang) {
    off2ang3d_kxyz(par_op,dmIn,dmOut);
  } else {
    //ang2off3d_kxyz(par_op,dmIn,dmOut);
  }
  
  //finish
  delete pars;
  return 0;
}

