#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"
#include "stackTraceC.h"
#include "commonOps.h"
#include "imaging.h"

using namespace Yang;
using namespace std;

//Do semblance scan.
int main( int argc, char** argv){
    initCallStackTrace();
   //Read params
   auto_ptr<sep_params> pars(new sep_params(argc,argv));
   hypercube modShape,datShape;
   
   MOtype moveOutT = Yang::Quadratic;
   int iMoveout = pars->get_int("moveoutType",2);
   switch(iMoveout) {
   case 1:
     moveOutT = Yang::Linear; break;
   case 2:
     moveOutT = Yang::Quadratic; break;
   case 3:
     moveOutT = Yang::Hyperbolic; break;
   case 4:
     moveOutT = Yang::tangentSquare;
     break;
   default:
     VERIFY(0);
   }
   
   bool bAdj = pars->get_bool("bAdj",true); //Adjoint is semblance scan.
   PRV(iMoveout);
   //Just input two cube for the geometry
   cubeDataF *pMod,*pDat;
   

   int   n2 = pars->get_int   ("n2",1);
   float o2 = pars->get_float ("o2",0.f);
   float d2 = pars->get_float ("d2",1.f);
   axis ax(n2,o2,d2);
   axis ax1; vecAxis axes;
   if (bAdj) {
     pDat = new cubeDataF("dat",dm::IN,"dat shape");
     pMod = new cubeDataF("mod",dm::OUT,"mod shape");
     
     pDat->init(NULL,true);
     //Forc
     datShape = pDat->domain();
     ax1 = datShape.get_axis(1);
     axes.push_back(ax1);
     ax.label = "mu"; axes.push_back(ax);
     axes.push_back(datShape.get_axis(3));
     
     modShape.set_axes(axes);
     pMod->init(&modShape);
     if (moveOutT == Yang::tangentSquare) {//assume deg input, convert to radian.
       degRadianConvert(*pDat,2,true);
       pDat->domain().info("pDat in radian");
     }
   } else {
     pDat = new cubeDataF();
     pMod = new cubeDataF("mod",dm::IN,"mod shape");
     pMod->init(NULL,true);
     modShape = pMod->domain();
     ax1 = modShape.get_axis(1);
     ax.label = "hx";
     axis ax_new = ax;
     if (moveOutT == Yang::tangentSquare) {//assume deg output, convert from deg to radian for calcualation
       deg2radianAxis(ax_new,ax,true);
     }
     axes.push_back(ax1);
     axes.push_back(ax_new);
     axes.push_back(modShape.get_axis(3));
     datShape.set_axes(axes);
     pDat->init(&datShape);
   }
   
   float faa = pars->get_float("f_AA",1.0f);
   PRV(faa);
   
   calcMoveoutSemblance(*pMod, *pDat, moveOutT, faa, bAdj);   
   pDat->info("data"); pMod->info("Model");
   if (!bAdj) {
     if (moveOutT == Yang::tangentSquare) {//assume deg output, convert to deg.
       degRadianConvert(*pDat,2,false);
     }
     pDat->dump("dat");
   }
   
   pMod->unInit();
   pDat->unInit();
}


