#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"
#include "dmHeaderReader.h"
#include "cubeAlgo2.h"
#include "stackTraceC.h"
using namespace Yang;
using namespace std;

///Shuffle Axis, manipulation, but is not optimized and only incore. Therefore does not suit very large data.
/// Usage:
/// *.x < in.H output=out.H reshape=1,3,4 mapping=2,1,3 verb=y/n

int main( int argc, char** argv){
   initCallStackTrace();
   //Read params
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   
   //in is the model
   //out is the gradiant
   sepHeaderReader inHeader("in",dm::IN,"",true);
   hypercube inShp;

   //Take the time derivative
   int nmax=20;
   int n1,n2;
   stdVecInt ndef(nmax,-1);

   int* reshape = autoPars->get_ints("reshape",nmax,&ndef[0],&n1);
   int* mapping = autoPars->get_ints("mapping",nmax,&n2);
   bool verb = autoPars->get_bool("verb",true);

   if (n1==0) { //no need to reshape, use default
     n1 = n2;
     reshape = new int[n2];
     for (int i=1;i<=n2;i++) {
       reshape[i-1] = i;
     }
   }
   for (int i=1;i<=n2;i++) {
     VERIFY(mapping[i-1]<=n2 && mapping[i-1]>0);
   }
   
   VERIFY(n1==n2);
   VERIFY(tagExist("output"));
   stdVecInt v1(reshape,reshape+n1),v2(mapping,mapping+n1);
   
   int ndim = reshape[n1-1];

   inHeader.getShape(inShp,ndim);
   int esize = inHeader.getElemSize();
   if (verb) {
     PRVEC(v1,v1.size()); PRV2(esize,ndim); 
   }
   
   if (esize==4) {
     cubeData<float> input("in",dm::IN,"model"); input.init(NULL,ndim);
     cubeData<float> out;
     Yang::shuffleAxis(out,input,v1,v2,true,true,verb);
     if (verb) {
       out.info("out cube"); input.info("input cube");
     }
     out.dump("output");
   } else if (esize==1) {
     cubeData<byte> input("in",dm::IN,"",true); input.init(NULL,ndim);
     cubeData<byte> out;
     Yang::shuffleAxis(out,input,v1,v2,true,true,verb);
     if (verb) {
       out.info("out cube"); input.info("input cube");
     }
     out.dump("output");
   } else if (esize==8) {
     cubeData<cmplxf> input("in",dm::IN,"",true); input.init(NULL,ndim);
     cubeData<cmplxf> out;
     Yang::shuffleAxis(out,input,v1,v2,true,true,verb);
     if (verb) {
       //out.info("out cube"); input.info("input cube"); min() and max() function for in/out cube have not been implemented yet
     }
     out.dump("output");
   } else {
     VERIFY(0);
   }
}

