#include "stdafx.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "sep_params.h"
#include "misc.h"
#include <fftw3.h>
#include "myFFT.h"
#include <complex>

#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>


/**
 * Main application entry point
 * 
 * Fourier FT 3d tool, a wrap up for SEP3d.
 * Format:
   Off2ang3dB <in.H >out.H sign{1/2/3}={-1/0/1}
   params follows exactly the Ft3d SEPLib program.
 */
      
using namespace Yang;
using namespace std;

int main (int argc, char *argv[])
{
  // TODO: Add application code here
  initCallStackTrace();
  init_omp(0);
  
  //have to init par
	sep_params* pars = new sep_params(argc,argv);
  
  int nth_par = pars->get_int("nth",-1);
  PRV(nth_par);
  if (nth_par != -1) { //we read something from nth
    VERIFY(nth_par > 0);
    change_omp(nth_par);
  }

  FFT_IOtype e_iotype = fft_c2c; //default
  string str_iotype = pars->get_string("IOtype","c2c");
  if (str_iotype.compare("c2c")==0) {
    e_iotype = fft_c2r;
  } else if (str_iotype.compare("r2c")==0) {
    e_iotype = fft_r2c;
  } else if (str_iotype.compare("c2r")==0) {
    e_iotype = fft_c2r;
  } else {
    prt("IOtype needs to be one of 'c2c,r2c,c2r' \n");
    VERIFY(0);
  }
  
  int nth = g_NUM_THREADS; PRV(g_NUM_THREADS);
  //Read in/out tag
  bool bNotDummy = true;;
  dmRsf_oc<cmplxf> dmIn_c("in",Yang::dm::IN,"input",bNotDummy);  dmRsf_oc<cmplxf> dmOut_c("out",Yang::dm::OUT,"output",bNotDummy);
  dmRsf_oc<float> dmIn_f("in",Yang::dm::IN,"input",bNotDummy);  dmRsf_oc<float> dmOut_f("out",Yang::dm::OUT,"output",bNotDummy);

  hypercube inShape;
  if (e_iotype==fft_c2c || e_iotype==fft_c2r) {
    dmIn_c.init(NULL);
    inShape = dmIn_c.domain();
  } else {
    dmIn_f.init(NULL);
    inShape = dmIn_f.domain();
  }

  
  int ndim = inShape.get_ndim();
  int sign[3]; 
  sign[0] = pars->get_int("sign1",0); VERIFY(sign[0]==-1 || sign[0]==1 || sign[0]==0);
  sign[1] = pars->get_int("sign2",0); VERIFY(sign[1]==-1 || sign[1]==1 || sign[1]==0);
  sign[2] = pars->get_int("sign3",0); VERIFY(sign[2]==-1 || sign[2]==1 || sign[2]==0);
  
  int n[3]; n[1]=n[2]=n[0]=1;
  n[0] = inShape.get_n(1); n[1] = inShape.get_n(2);
  if (ndim > 1) {n[1] = inShape.get_n(2);}
  if (ndim > 2) {n[2] = inShape.get_n(3);}
  
  //The output dimensions info
  int no[3]; myMemcpy(no,n,3);
  if (sign[0]!=0) {no[0] = pars->get_int("n1",n[0]); VERIFY(no[0]>=n[0])}; 
  if (sign[1]!=0) {no[1] = pars->get_int("n2",n[1]); VERIFY(no[1]>=n[1])}; 
  if (sign[2]!=0) {no[2] = pars->get_int("n3",n[2]); VERIFY(no[2]>=n[2])}; 
  
  vecAxis axes = inShape.get_axes();
  axis ax;
  for (int i=0; i<min(3,ndim); i++) {
    if (sign[i]!=0) { //change axis 1
      get_axis_inv(axes[i], no[i], axes[i]);
    }
  }

  //Do the fft
  hypercube shapeOut; shapeOut.set_axes(axes);
  cubeData<float> cb_f_read,cb_f_write;
  hypercube shpOutcube = hypercube(no[0],no[1],no[2]), shpIncube(n[0],n[1],n[2]);
  bool bInput_r = false, bOutput_r = false;

  if (e_iotype==fft_c2c) {
    dmOut_c.init(&shapeOut);
  } else if (e_iotype==fft_c2r) {
    dmOut_f.init(&shapeOut);
    cb_f_write.alloc(shpOutcube);
    bOutput_r = true;
  } else { //r2c
    dmOut_c.init(&shapeOut);
    cb_f_read.alloc(shpIncube);
    bInput_r = true;
  }
  
  cubeData<cmplxf> inCube(shpIncube);
  cubeData<cmplxf> *poutCube[nth]; 
  for (int i=0; i<nth; i++) {
    poutCube[i] = new cubeData<cmplxf>(shpOutcube);
  }
  
  int nRead = 1; if(ndim>3) nRead = inShape.get_n23(4,ndim);
  int n123 = no[0]*no[1]*no[2];
  //create the fftw plan
  fftwf_complex *fftwin2d[nth], *fftwin[nth];
  fftwf_plan FFTplan[3][nth];
  int ith;
  for (ith=0;ith<nth;ith++) {
    fftwin[ith]  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n123);         memset(fftwin[ith],0,n123*sizeof(fftwf_complex));
    fftwin2d[ith]= (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*no[0]*no[1]);  memset(fftwin2d[ith],0,no[0]*no[1]*sizeof(fftwf_complex));
  }
  
  int rank=1; int dir;
  float scalar=1.0f;
  
  //along axis 1
  if (sign[0]!=0) {
    dir = (sign[0]==-1)? FFTW_FORWARD:FFTW_BACKWARD;
   for (ith=0;ith<nth;ith++) {
     FFTplan[0][ith] = fftwf_plan_many_dft(rank,&no[0],n123/no[0], fftwin[ith], &no[0],1          ,no[0], fftwin[ith],&no[0],1,no[0], dir, FFTW_MEASURE);
     VERIFY( FFTplan[0][ith]!=NULL);
   }
    scalar *= 1.f/sqrtf(float(no[0]));
  }
  
  //along axis 3
  if (sign[2]!=0) {
    dir = (sign[2]==-1)? FFTW_FORWARD:FFTW_BACKWARD;
    for (ith=0;ith<nth;ith++) {
      FFTplan[2][ith] = fftwf_plan_many_dft(rank,&no[2],n123/no[2], fftwin[ith], &no[2],no[0]*no[1],1,     fftwin[ith],&no[2],no[0]*no[1],1, dir, FFTW_MEASURE);
      VERIFY( FFTplan[2][ith]!=NULL);
    }
    scalar *= 1.f/sqrtf(float(no[2]));
  }
  
  //along axis 2 cannot be done at once. need to do each 3D slice by slice.
  if (sign[1]!=0) {
    dir = (sign[1]==-1)? FFTW_FORWARD:FFTW_BACKWARD;
    for (ith=0;ith<nth;ith++) {
      FFTplan[1][ith] = fftwf_plan_many_dft(rank,&no[1],no[0]     , fftwin2d[ith], &no[1],no[0]      ,1,     fftwin2d[ith],&no[1],no[0]      ,1, dir, FFTW_MEASURE);
      VERIFY( FFTplan[1][ith]!=NULL);
    }
    scalar *= 1.f/sqrtf(float(no[1]));
  }
  
  windParam wnd(3,windowAxis());
  wnd[0].set(n[0],0,1); wnd[1].set(n[1],0,1); wnd[2].set(n[2],0,1);
  long long n123_in,n123_out;
  n123_in   = shpIncube.get_n123(); n123_out  = shpOutcube.get_n123();

  //Enter the main compuation part
  int iRead = 0;
  #pragma omp parallel default(shared) private(iRead,ith) 
  {
  ith = omp_get_thread_num();
  for (iRead=0; iRead<nRead; iRead+=nth) {
    int mRead = std::min(nth,nRead-iRead);
    //Only the thread 0 does the IO
    if (ith==0) {
      for (int i = 0; i<mRead; i++) {
        if (bInput_r) {
          dmIn_f.seq_grab(&cb_f_read);
          //then put float into cmplx arrays
          cmplxf *pout=inCube.dataPtr(); float *pin=cb_f_read.dataPtr();
          for (long long jind=0; jind<n123_in;jind++) {
            pout[jind] = cmplxf(pin[jind],0.f);
          }
        } else {
          dmIn_c.seq_grab(&inCube);
        }
        poutCube[i]->zero();
        poutCube[i]->updateWindow(&inCube,wnd);
        //Then copy to fftwin array
        memcpy(fftwin[i],poutCube[i]->pval,sizeof(cmplxf)*n123);
      }
    }
    #pragma omp barrier //set a barrier
    
    //Every thread that gets the data does the work
    if (ith < mRead) {
      //Do first axis
      if (sign[0]!=0) {
        fftwf_execute(FFTplan[0][ith]);
      }
      //axis 3
      if (sign[2]!=0) {
        fftwf_execute(FFTplan[2][ith]);
      }
      //axis 2
      if (sign[1]!=0) {
        int n12 = no[0]*no[1];
        for (int i3=0; i3<no[2];i3++) {
          myMemcpy(fftwin2d[ith],fftwin[ith]+i3*n12,n12);
          fftwf_execute(FFTplan[1][ith]);
          myMemcpy(fftwin[ith]+i3*n12,fftwin2d[ith],n12);
        }
      }
      memcpy(poutCube[ith]->pval,fftwin[ith],sizeof(cmplxf)*n123);
      poutCube[ith]->scale(cmplxf(scalar,0.f));
    }
    
    //let all thread finish the computation
    #pragma omp barrier 
    
    //do the write out
    if (ith == 0) {
      for (int i = 0; i<mRead; i++) {
        if (!bOutput_r) {
          dmOut_c.seq_update(poutCube[i]);
        } else {
          float *pout=cb_f_write.dataPtr(); const cmplxf *pin=poutCube[i]->dataPtr();
          for (long long jind=0; jind<n123_out;jind++) {
            pout[jind] = pin[jind].real();
          }
          dmOut_f.seq_update(&cb_f_write);
        }
      }
    }
  } //end for iRead
  } //end omp parallel
  
  delete pars;
  return 0;
}


