#ifndef VELFD2D_H
#define VELFD2D_H 1

#include "cubeData.hpp"

namespace Yang {

   //Might not be proper to inherit from dmCube, cause vel might be a out core as well.
   //Or might be as output.
   //The region within the aperture each time is considered always-incore
   class vel: public cubeData<float>
   {
   protected:
	  cubeData<float>* m_pVelSubCube;
   public:
      vel(string itag, dm::typeIO _tIO, const string& szInfo): cubeData<float>(itag, _tIO, szInfo),bInCore_(false) {
			m_pVelSubCube = new cubeData<float>("",dm::OUT,"velSubCube");
		}
		~vel() {
			SAFE_DELETE(m_pVelSubCube);
		}
	  //set the velocity to be this cube
	  void setVelocity(const cubeData<float>& velCube) {
      if (this==&velCube) { //if is the pointer itself, then do nothing.
        return; 
      } else {
       VERIFY(this->sameSize(velCube));
       velCube.copyTo(*this);
       this->copyTo(*m_pVelSubCube);
      }
	  }
	  
    void loadVel() {
			bool bRemoveDummyAxis = true;
		 this->init(NULL, bRemoveDummyAxis);
		 //Copy the vel into the subCube
		 this->copyTo(*m_pVelSubCube);
		 bInCore_ = true; vShp_ = this->domain();
	  }

    void loadVel(int n_axis_desired) {
     this->init(NULL, n_axis_desired);
     //Copy the vel into the subCube
     this->copyTo(*m_pVelSubCube);
		 bInCore_ = true; vShp_ = this->domain();
    }
	  
		void peekVelShape(int n_axis_desired); //just extract the header information.
	  const cubeData<float>& get2dVelPtr() {
		 VERIFY(m_pVelSubCube->domain().get_ndim()>=2);
		 //return m_pVelSubCube->get2dPtr();
     return *m_pVelSubCube;
	  }
		const hypercube& vmodShp() const {return vShp_;}
	  const cubeData<float>& get3dVelPtr() {
		  VERIFY(m_pVelSubCube->domain().get_ndim()>=3);
      return *m_pVelSubCube;
		  //return m_pVelSubCube->get3dPtr();
	  }
	  
	  float* get1dVelPtr() {
		 return m_pVelSubCube->dataPtr();
	  }
	  
	  const cubeData<float>& getVelCube() {
		 return *m_pVelSubCube;
	  }

	 private:
		//There needs to be an out-of-core component, that only get the header info, but not the binary content.
		hypercube vShp_;
		bool bInCore_;
   };

class vel2d : public vel
   {
   public:
      vel2d(string itag, dm::typeIO _tIO, const string& szInfo):vel(itag, _tIO, szInfo) {}
   };
}
#endif

