#include "stdafx.h"
#include <sstream>
#include "decl_fwd.h"

#include "waveOpOneway.h"
#include "dmHeaderReader.h"
#include "cubeAlgo2.h"
#include "velAna.param.h"
#include "misc.h"

#include "param_func.h"
#include "my_mpi.h"
#include "waveOpMPI.h"
#include "mpi_util.h"

//#include <boost/filesystem/convenience.hpp>
//namespace fs = boost::filesystem; needs lib files during linking

using namespace Yang;
using namespace std;

static string strCmd;
static char szCmd[4096];

#ifndef USE_MPI
#error "not-using-mpi"
#else
#include "my_mpi.h"

//Do 3D mpi wave-equation operators with Yaxun's oneway code, everything now needs to be disk files rather than RAM blocks, and needs dynamic scheduling
waveOpOnewayMPI3d::waveOpOnewayMPI3d(const shared_ptr<waveOpParams>& waveOpPar, const shared_ptr<AcqPar>& acqPars, const shared_ptr<ImgPar>& imgPars, 
const vel& vmod, param_func* pars, const boost::shared_ptr<shotFilesInfo>& shotFilesInfo_in): waveOpOneway(waveOpPar,acqPars,imgPars,vmod,pars) { 
  strTmpDir_ = param_func::s_tmpPath;
  shotFilesInfo_ = shotFilesInfo_in;
}

//Different ways of setting oneway pararm, because we use shotsInfo, and the velocity model is stored as a file
void waveOpOnewayMPI3d::setOneWayParam( OneWayParamTang* par) {
  FUNC_TRACE_BEG("OneWayMPI3d::setOneWayParam");
  
  const hypercube& vmodShp = *this->vmodShp();
  axis ax_x = vmodShp.get_axis(1); //vmodshp is (x,y,z) in oneway setup.
  axis ax_y = vmodShp.get_axis(2); 
  
  char szApp[256];
  //set the offset image migration parameters
  const ImgPar* p = this->imgPars().get();
  VERIFY(NULL!=p);
  const ImgPar& imgPar = *p;
  bool bDoOffset = imgPar.bGenOffsetGather();
  par->nhx = NumSubOffset(1); par->ohx = 0.0f; par->dhx = 0.0f; float d=0.f;
  bool b3D = ax_y.n>1; VERIFY(b3D);
  const offsetImagingParams& offImgPar = imgPar.offImgPars();
  if (bDoOffset) {
    d = ax_x.d; //dx
    par->ohx = offImgPar.ohx * d;
    par->dhx = offImgPar.dhx * d;
    par->nhx = offImgPar.nhx;
    sprintf(par->szOffImgPar, " ohx=%f dhx=%f nhx=%d ", par->ohx, par->dhx, par->nhx.value());
    if (b3D) {
      d = ax_y.d; //dy
      par->ohy = offImgPar.ohy * d;
      par->dhy = offImgPar.dhy * d;
      par->nhy = offImgPar.nhy;
      sprintf(szApp, " ohy=%f dhy=%f nhy=%d ", par->ohy, par->dhy, par->nhy.value());
      strcat(par->szOffImgPar, szApp);
    }
  } else { //not do offset
  }
  
  const OnewayOpPar* wavePar = this->waveOpPars()->onewayPar();
  sprintf(par->szMigPar," %s ", wavePar->sz_additional_mig_par.c_str());
    
  this->genTempFileNamesOneway(par);
  //after this, set the velocity files name to be the single velocity model files name.
  par->g_bvelFileName = this->vmodFileName(); PRV(par->g_bvelFileName);
  VERIFY(!(par->g_bvelFileName.empty()));
  FUNC_TRACE_END();
}

//Used for MPI case, write to szGeomPar for each process, this is for modeling data
void waveOpOnewayMPI3d::setGeomPar(char* szGeom, const OneWayParamTang* par, int nrank, int irank) {
  FUNC_TRACE_BEG("OneWay::setGeomParam");
  
  const hypercube& vmodShp = this->vmod().domain();
  //axis ax_z = vmodShp.get_axis(1);
  axis ax_x = vmodShp.get_axis(1);
  axis ax_y = vmodShp.get_axis(2);
  
  VERIFY(0); //not finished yet!
  boost::shared_ptr<const AcqPar> acqPars = this->acqPars();
  
  //int nproc = nrank-1;
  int nsx; nsx=1;
  float osx,dsx;
  //See how many processes are in total.
  if (nrank==1) { //emulating the no MPI case
    sprintf(szGeom," ngx=%d nsx=%d ogx=%f dgx=%f osx=%f dsx=%f offset=%c ",
          par->ngx,par->nsx,par->ogx,par->dgx,par->osx,par->dsx, par->offset);
  } else { //if MPI mode
    VERIFY(irank>0);
    int n=1,o=0.f,d=1.0f;
    //ws_->get_nod(irank-1,&n,&o,&d);
    nsx = n; //the number of shots for this process, it handles shots ir,ir+nr,ir+2*nr,..., till shots nsx
    osx = par->osx + par->dsx*o;
    dsx = par->dsx*d;
    sprintf(szGeom,"ngx=%d nsx=%d ogx=%f dgx=%f osx=%f dsx=%f offset=%c ",
          par->ngx,nsx,par->ogx,par->dgx,osx,dsx, par->offset);
  }
  
  //just break along x, not along Y;
  char szApp[256];
  bool b3D = acqPars->b3D();
  if (b3D) {
    sprintf(szApp," ngy=%d nsy=%d ogy=%f dgy=%f osy=%f dsy=%f ",
            par->ngy, par->nsy, par->ogy, par->dgy, par->osy, par->dsy);
    strcat(szGeom, szApp);
  }
  FUNC_TRACE_END();
}

//All the cubes are in 1way axis arrangement
void waveOpOnewayMPI3d::gen_mig3d_cmd(string& str_cmd, hypercube& outImgShp, const OneWayParamTang& param,int iJob, int iwrk, const Yang::shotFilesInfo& sfInfo, const hypercube& global_imgShp, const branchFileName& bf_imag, const branchFileName& bf_log, const branchFileName& bf_vel_local) {
  FUNC_TRACE_BEG("waveOpOnewayMPI3d::gen_mig3d_cmd");
  string f_csou, f_crec;
  //Get the filenames for csou and crec
  sfInfo.getDataAndSrcFile(f_crec,f_csou,iJob);

  PRV2(f_crec,iJob);
  //gen img dimension string
  bool bxmin,bxmax,bymin,bymax,bzmin,bzmax;
  float xmin,xmax,ymin,ymax,zmin,zmax;
  xmin=xmax=ymin=ymax=zmin=zmax=0.f;
  sfInfo.getImagingDomainFlags(bxmin,bxmax,bymin,bymax,bzmin,bzmax);
  sfInfo.getImagingDomain(xmin,xmax,ymin,ymax,zmin,zmax,iJob);
  
  axis ax_x,ax_y,ax_z;
  const hypercube& vshp = *(this->vmodShp());
  ax_x = vshp.get_axis(1); ax_y = vshp.get_axis(2); ax_z = vshp.get_axis(3);
  
  axis axg_x,axg_y,axg_z;
  axg_x = global_imgShp.get_axis(3);axg_y = global_imgShp.get_axis(4);axg_z = global_imgShp.get_axis(5);

  //if the global imageShp is broader than single shot img aperture, then just use the shot img aperture as the imaging domain. If vice verse, then reduce the shot imaging domain using the global imageShp
  if (!bxmin) { //compare velmodel and global image shp
    if (axg_x.o > ax_x.o) { bxmin = true; xmin = axg_x.o;}
  } else {  //compare global imageShp and single shot aperture
    if (axg_x.o > xmin) {xmin=axg_x.o;}
  }
  if (!bymin) { 
    if (axg_y.o > ax_y.o) { bymin = true; ymin = axg_y.o;}
  } else {  
    if (axg_y.o > ymin) {ymin = axg_y.o;}
  }
  if (!bzmin) { 
    if (axg_z.o > ax_z.o) { bzmin = true; zmin = axg_z.o;}
  } else {  
    if (axg_z.o > zmin) {zmin=axg_z.o;}
  }

  float gxmax,gymax,gzmax, vxmax,vymax,vzmax;
  axg_x.getRange(NULL,&gxmax);  axg_y.getRange(NULL,&gymax);  axg_z.getRange(NULL,&gzmax);
  ax_x.getRange(NULL,&vxmax);   ax_y.getRange(NULL,&vymax);   ax_z.getRange(NULL,&vzmax);

  if (!bxmax) { //compare velmodel and global image shp
    xmax = vxmax;
    if (gxmax < vxmax) { bxmax = true; xmax=gxmax;}
  } else {  //compare global imageShp and single shot aperture
    if (gxmax < xmax) {xmax=gxmax;}
  }
  if (!bymax) { 
    ymax = vymax;
    if (gymax < vymax) { bymax = true; ymax = gymax;}
  } else {  
    if (gymax < ymax) {ymax = gymax;}
  }
  if (!bzmax) { 
    zmax=vzmax;
    if (gzmax < vzmax) { bzmax = true; zmax = gzmax;}
  } else {  
    if (gzmax < zmax) {zmax=gzmax;}
  }

  std::stringstream ss(""); 
  if (bxmin) {ss << " image_xmin=" << xmin; ax_x.o = xmin;}
  if (bxmax) {ss << " image_xmax=" << xmax; }
  if (bxmin | bxmax) {ax_x.n = ax_x.get_offset(xmax)+1;}
  if (bymin) {ss << " image_ymin=" << ymin; ax_y.o = ymin;}
  if (bymax) {ss << " image_ymax=" << ymax; }
  if (bymin | bymax) {ax_y.n = ax_y.get_offset(ymax)+1;}
  if (bzmin) {ss << " image_zmin=" << zmin; ax_z.o = zmin;}
  if (bzmax) {ss << " image_zmax=" << zmax; }
  if (bzmin | bzmax) {ax_z.n = ax_z.get_offset(zmax)+1;}
  
  //The first two axes not touched, therefore
  outImgShp.set_axis(3,ax_x);outImgShp.set_axis(4,ax_y);outImgShp.set_axis(5,ax_z);
  
  char ch_offset = 'y';
  bool bMarine = sfInfo.bMarineGeometry();
  if (!bMarine) {ch_offset='n';}
  //first copy the data-record file
  string crec_base = get_basename_from_path(f_crec);
  string csou_base = get_basename_from_path(f_csou);
  PRV2(crec_base,csou_base);

  string str_ltp = "/tmp/"; //local_tmp_path
  crec_base = str_ltp+crec_base; csou_base = str_ltp + csou_base;

  str_cmd = "";
  const OnewayOpPar *par_1way = this->waveOpPars()->onewayPar();
  int iw_beg=par_1way->iw_beg_, nw=par_1way->nws_; //Only do a portion of the frequencies
  char sz_wnd[128];
  if (nw == 0) {sprintf(sz_wnd," f3=%d ", iw_beg);}
  else {    sprintf(sz_wnd," f3=%d n3=%d ", iw_beg,nw);}

  sprintf(szCmd,"Window3d squeeze=n %s <%s >%s datapath=/tmp/ ;", sz_wnd,f_crec.c_str(), crec_base.c_str()); str_cmd += szCmd;
  sprintf(szCmd,"Window3d squeeze=n %s <%s >%s datapath=/tmp/ ;", sz_wnd,f_csou.c_str(), csou_base.c_str()); str_cmd += szCmd;
  //sprintf(szCmd,"Cp <%s >%s datapath=/tmp/ ;", f_crec.c_str(), crec_base.c_str()); str_cmd += szCmd;
  //sprintf(szCmd,"Cp <%s >%s datapath=/tmp/ ;", f_csou.c_str(), csou_base.c_str()); str_cmd += szCmd;

  sprintf(szCmd,"%s offset=%c csou=%s crec=%s bvel=%s bimg=%s mode=imgadj memchk=n report=y %s %s %s >&%s datapath=/tmp/ ", //redirect the output
          szWemig,ch_offset,
          //f_csou.c_str(),f_crec.c_str(),param.bvelFileName.c_str(), //if do not utilize local file system.
          csou_base.c_str(),crec_base.c_str(),bf_vel_local.fn_at(iwrk).c_str(),
          bf_imag.fn_at(iwrk).c_str(),
          param.szMigPar,param.szOffImgPar, ss.str().c_str(), bf_log.fn_at(iwrk).c_str());
  str_cmd += szCmd;
  //LOG("finish invoking the Mig wem3d program in MPI version"); PRV(str_cmd);
  FUNC_TRACE_END();
}

//Cp velocity model to local tmp directory
void waveOpOnewayMPI3d::deployVelModel(OneWayParamTang& param, const branchFileName& bf_vel_local ) {
  string str_cmd;
  int nWrkers = world_.size()-1;
  //copy the files to local directory
  for (int iwrk=1; iwrk<=nWrkers; iwrk++) {
    sprintf(szCmd,"Scp <%s >%s datapath=/tmp/ ", param.g_bvelFileName.c_str(),bf_vel_local.fn_at(iwrk).c_str()); //Program will hang (in MPI mode) if I do not supply stdout
    str_cmd = szCmd;
    world_.send(iwrk,WRKTAG,str_cmd);
  }
  
  int nWrkersLeft = nWrkers, iwrk;
  string str_msg;
  vector<bool> idle(nWrkers,false);
  //Then need to wait for them to finish
  while (nWrkersLeft>0) {//stop only if all Wrkers has finished, a barriere method might be easier.
    //receiving idle signals from the workers.
    //if collect ideltag, then send a task msg for it to run
    LOG("root about to recv in deployVelModel");
    const MPI_Status& status = world_.recv(MPI_ANY_SOURCE,MPI_ANY_TAG,str_msg);
    iwrk = status.MPI_SOURCE;
    if (status.MPI_TAG == FINISHTAG) {
      VERIFY(idle[iwrk-1]==false);
      nWrkersLeft--;
      idle[iwrk-1] = true; //now reset the idle state.
    } else {
      PRV(status.MPI_TAG); PRV(iwrk);
      VERIFY(0);
    }
  }
}

void waveOpOnewayMPI3d::migration(OneWayParamTang& param, cubeData<float>* pImageCube) {
  FUNC_TRACE_BEG("waveOpOnewayMPI3d::migration");
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ    = acqPar->shotDepth();
  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  LOGC("Now the inputs are ready for waveOpOnewayMPI::migration\n");
  //string vmod_fname = param.bvelFileName;

  string bf_suf = this->waveOpPars()->suf_MPI_;
  branchFileName bf_vel_local(param.bvelFileName,bf_suf);
  //now in 3d, we have pre-defined the small pieces in files, we need to use a geometry file that contains all geometry information for each shot, also, we need to do dynamic scheduling
  branchFileName bf_imag(param.imagFileName,bf_suf), bf_log(param.logfilename,bf_suf); //The form the cmd for the actual migration
  
  //first send out the velocity input to local disk
  this->deployVelModel(param,bf_vel_local);

  //dynamic scheduling. Look at the geometry, take one shot, calclulate the corresponding imaging domain, send the cmd to one of the workers, and then wait for response.
  int nWrkers = world_.size()-1;
  int nJobLeft,nJobFinished,iJob;
  int nJob = shotFilesInfo_->m_nSuperShots();
  std::valarray<bool> idle(nWrkers);  idle = true; //at start, all nodes are idle.
  nJobLeft = nJob; nJobFinished = 0;
  string str_msg,str_cmd,str_imgName;
  int iwrk = 1; iJob = 0; //start job number from zero
  long long n123;
  const hypercube& offImgShpGlobal = this->imgPars()->finalSubOffImageShp1way();
  VERIFY(pImageCube->domain().sameShape(offImgShpGlobal));
  
  vector<hypercube> vec_domains(nWrkers);
  //first send the first batch of work to the workers, wrker id starts from 1
  for (iwrk=1;iwrk<=nWrkers;iwrk++) {
    //get the exact cmd using the oneway par and combined with the geometry object
    vec_domains[iwrk-1] = offImgShpGlobal;  //record the imaging domain of the wrkers
    gen_mig3d_cmd(str_cmd,vec_domains[iwrk-1],param,iJob,iwrk,*shotFilesInfo_,offImgShpGlobal,bf_imag,bf_log,bf_vel_local);
    vec_domains[iwrk-1].info("vec_domains");
    //now the domain is recorded in this vector
    world_.send(iwrk,WRKTAG,str_cmd);
    idle[iwrk-1] = false; //set that idle to busy
    nJobLeft--; iJob++;
    //check if nJobLeft is already zero
    if (nJobLeft == 0) {break;}
  }
  cubeDataF partial_img;
  while (nJobLeft>0 || nJobFinished < nJob) {//stop only if all job has been distributed and all job has been finished
    //receiving idle signals from the workers.
    //if collect ideltag, then send a task msg for it to run
    LOG("root about to recv");
    const MPI_Status& status = world_.recv(MPI_ANY_SOURCE,MPI_ANY_TAG,str_msg);
    iwrk = status.MPI_SOURCE;
    if (status.MPI_TAG == FINISHTAG) {
      nJobFinished++;
      VERIFY(idle[iwrk-1]==false);
      //then try to collect the result
      str_imgName = bf_imag.fn_at(iwrk);
      LOG("root about to send");
      world_.send(iwrk,COLLECT_BINARY_TAG,str_imgName);
      
      //then check the size and recv the buffer
      const hypercube& shp_part = vec_domains[iwrk-1];
      partial_img.alloc(shp_part); 
      n123 = shp_part.get_n123();
      PRV(iwrk); //shp_part.info("shp_part");
      world_.collect_binary(0,iwrk, partial_img.dataPtr(), MPI_FLOAT, n123);
      //partial_img.info("partial_img"); 
      LOG("d1"); //pImageCube->info("pImageCube");
      //then we need to add this part to the global offset img array
      Yang::cropAlign(*pImageCube, partial_img, true, false);
      //LOG("d2");
      //now it is finished, 
      idle[iwrk-1] = true; //now reset the idle state.
    } else if (status.MPI_TAG == IDLETAG) {
      idle[iwrk-1] = true;
      VERIFY(0); 
    }
    bool bCanDispatchWrk = idle[iwrk-1];
    PRV2(nJobFinished,nJobLeft);
    if (bCanDispatchWrk && nJobLeft>0) {
      //Send the wrk to the corresponding nodes.
      gen_mig3d_cmd(str_cmd,vec_domains[iwrk-1],param,iJob,iwrk,*shotFilesInfo_,offImgShpGlobal,bf_imag,bf_log,bf_vel_local);
      LOG("root about to send");
      world_.send(iwrk,WRKTAG,str_cmd);
      idle[iwrk-1] = false; //set that idle to busy
      nJobLeft--; iJob++; 
    }
  } //finish
  LOG("finish invoking the Mig wem3d program in MPI3d version");
  FUNC_TRACE_END();
}

//Do born modeling
void waveOpOnewayMPI3d::bornModeling(OneWayParamTang& param, const cubeDataF* pRefl) {
  //const AcqPar* acqPar  = this->acqPars().get();
  //int   shotOffZ        = acqPar->shotDepth();
  LOG("DD1");
  VERIFY(0); //not finished yet.
  cubeData<float> vmodCube1,vmodCube2;
  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  cubeDataF& vel = const_cast< cubeDataF & > (this->vmod());
  this->convertVelModel(vmodCube1,vel,true);
  pad2D_Tvmod(vmodCube2,vmodCube1);
  //Dump the vmod to file
  vmodCube2.dump(param.bvelFileName,true);  usleep(int(1e5)); //for IO safety
  PRV(param.bvelFileName);
  
  //Convert the input image to Tang's format, and pad the images.
  cubeDataF Image2;
  this->convertImage(Image2,const_cast<cubeDataF&>(*pRefl),true);
  string tmpFn = param.imagFileName+"-a.H";
  Image2.dump(tmpFn,true);
  strCmd = pad2D_Timg_cmd(param.imagFileName, tmpFn); //do the paddings.
  VERIFY(system(strCmd.c_str())==0);
  
  PRV2(tmpFn,strCmd);
  strCmd = "Rm "; strCmd += tmpFn; VERIFY(system(strCmd.c_str())==0); //clean the tmp file.
  LOG("DD3");
  
  //Now ask each process to simulate the data
  const waveOpParams* wPar = this->waveOpPars().get();
  PRV2(param.csouFileName,param.crecFileName); PRV(param.logfilename);
  branchFileName bf_csou(param.csouFileName,wPar->suf_MPI_), bf_crec(param.crecFileName,wPar->suf_MPI_), bf_log(param.logfilename,wPar->suf_MPI_); 
  char szGeom[512];
  for (int ir=1; ir<world_.size(); ir++) {
    //find the Geom for this process
    this->setGeomPar(szGeom, &param, world_.size(), ir);
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgfwd memchk=n report=y %s %s >&%s", //redirect the output
            szWemig,bf_csou.fn_at(ir).c_str(),bf_crec.fn_at(ir).c_str(),param.bvelFileName.c_str(),
            param.imagFileName.c_str(),param.szMigPar,szGeom,bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd); //ask the children to do their work.
    PRV(strCmd);
  }
  world_.barrier(); //wait for other worknodes to finish.
  
  //now cat the crec files together
  strCmd = bf_crec.catCmd(world_.size(),4); //the 4th axis is the sx axis.
  PRV(strCmd);
  VERIFY(system(strCmd.c_str())==0);

  LOG("finish invoking the wem3d modeling program");
  //cleaned up.
  for (int ir=1; ir<world_.size(); ir++)  {
    sprintf(szCmd,"rm %s;", bf_log.fn_at(ir).c_str());
    //strCmd = szCmd; world_.send(ir,WRKTAG,strCmd); 
  }
  LOG("DD5");
  //Need to wait for all process to finish.
  
  //usleep(20e6);
  LOG("DD6");
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.bvelFileName.c_str(),param.imagFileName.c_str());
  //VERIFY(system(szCmd)==0);usleep(int(1e5));
  usleep(2e6);
}


//Input Image perturbation and output slowness perturbation
void waveOpOnewayMPI3d::TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  VERIFY(0); 
  FUNC_TRACE_BEG(waveOpOneway::TomAdj_op);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ          = acqPar->shotDepth();

  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground velocity, and pad.
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,vel,true);
  string tmpFn = param.bvelFileName+"-a.H";
  vmodCube2.dump(tmpFn,true); usleep(int(1e5));
  strCmd = pad2D_Tvmod_cmd(param.bvelFileName,tmpFn);
  VERIFY(system(strCmd.c_str())==0); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,*pDeltaIh,true);
  Image2.dump(tmpFn,true); usleep(int(1e5));
  strCmd = pad2D_Timg_cmd(param.dimgFileName,tmpFn);
  VERIFY(system(strCmd.c_str())==0); usleep(int(1e5));
  
  const velAnaParam* vaPar = this->velAnaPars();
  VERIFY(NULL != vaPar);
  char cIllumBal = 'n';
  PRV(vaPar->bIllumBal_);
  if (vaPar && vaPar->bIllumBal_) {cIllumBal='y';}
  
  int ir=0;
  string bf_suf = this->waveOpPars()->suf_MPI_;
  branchFileName bf_csou(param.csouFileName,bf_suf),bf_crec(param.crecFileName,bf_suf),bf_log(param.logfilename,bf_suf),bf_dvel(param.dslwFileName,bf_suf),
    bf_srIl(param.srIlFileName,bf_suf);
  for (ir=1; ir<world_.size(); ir++) {
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y %s %s mode=tomadj bIllumBal=%c srIl=%s memchk=n report=y >&%s",
            szWemva,bf_csou.fn_at(ir).c_str(),bf_crec.fn_at(ir).c_str(),param.bvelFileName.c_str(),
            bf_dvel.fn_at(ir).c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szOffImgPar,cIllumBal,bf_srIl.fn_at(ir).c_str(),bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd); //ask the children to do their work.
  }
  world_.barrier(); //sync
  PRV(strCmd);

  //cleaned up.
  for (ir=1; ir<world_.size(); ir++)  {
    sprintf(szCmd,"Rm %s;Rm %s", bf_log.fn_at(ir).c_str(), bf_srIl.fn_at(ir).c_str());
    //strCmd = szCmd; world_.send(ir,WRKTAG,strCmd); 
  }
  //might need to stack the srIl illumnation pattern.
  
  cubeData<float> velY;
  //Stack the output
  if (0) {
    strCmd = bf_dvel.stackCmd(world_.size());
    VERIFY(system(strCmd.c_str())==0); usleep(int(1e5));
    //Need read the result image back
    velY.load(param.dslwFileName,3); usleep(int(1e5));
  } else { //Use MPI reduce to stack the final slowness
    sepHeaderReader hdr(bf_dvel.fn_at(1),dm::IN,"vel perturbation",true);
    hypercube shp_dvel;  hdr.getShape(shp_dvel,3);
    velY.alloc(shp_dvel);
    mpiY::ReduceCube_MPI(velY,PosInt(3),bf_dvel); //stack thre results using MPI_Reduce
  }
  
  //Now convert shape from slowY to pDeltaS, the cropping is included
  this->convertVelModel(velY,*pDeltaS,false);
  //the output of yaxun's program is vel, convert dvel -> ds by dividing -v^2;
  pDeltaS->divide(vel); pDeltaS->divide(vel); (*pDeltaS).scale(-1.0);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bvelFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  VERIFY(system(szCmd)==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

//Input image and image perturb, then return the slowness perturbation, assume 2D case.
void waveOpOnewayMPI3d::TomoImit_op (const cubeData<float>* pRefl, const cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  FUNC_TRACE_BEG(waveOpOneway::TomAdj_op);

  OneWayParamTang param1;
  this->setOneWayParam(&param1);
  
  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground velocity
  cubeDataF vmodCube1,vmodCube2;
  this->convertVelModel(vmodCube1,vel,true);
  //Pad x, the first axis
  seplib::callSepTool(vmodCube2,vmodCube1,"Pad","beg1=50 end1=50 extend=1",true);
  vmodCube2.dump(param1.bvelFileName,true); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,const_cast<cubeDataF&>(*pRefl),true);
  Image2.dump(param1.imagFileName,true); usleep(int(1e5));
  
  string bf_suf = this->waveOpPars()->suf_MPI_;
  branchFileName bf_csou(param1.csouFileName,bf_suf), bf_tcrec(param1.tempRecFileName,bf_suf), bf_log(param1.logfilename,bf_suf);
  branchFileName bf_dvel(param1.dslwFileName,bf_suf),bf_srIl(param1.srIlFileName,bf_suf);

  char szGeom[512];
  for (int ir=1; ir<world_.size(); ir++) {
    //find the Geom for this process
    this->setGeomPar(szGeom, &param1, world_.size(), ir);
    //do not need the subsurf imaging parameter, bcoz it is contained in file bimg
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s %s %s mode=imgfwd memchk=n report=y >&%s",
          szWemig,bf_csou.fn_at(ir).c_str(),bf_tcrec.fn_at(ir).c_str(),param1.bvelFileName.c_str(),param1.imagFileName.c_str(),
          szGeom, param1.szMigPar, bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd);
  }
  PRV(strCmd);
  world_.barrier();

  const velAnaParam* vaPar = this->velAnaPars();  VERIFY(NULL!=vaPar);
  char cIllumBal = 'n';
  if (vaPar && vaPar->bIllumBal_) {cIllumBal='y';}
  
  //Now use the synthesized data to perform the tomo operator
  this->convertImage(Image2,const_cast<cubeDataF&>(*pDeltaIh),true);
  Image2.dump(param1.dimgFileName,true); usleep(int(1e5));
  
  for (int ir=1; ir<world_.size(); ir++) {
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y %s mode=tomadj bIllumBal=%c srIl=%s memchk=n report=y 2>> %s",
            szWemva,bf_csou.fn_at(ir).c_str(),bf_tcrec.fn_at(ir).c_str(),param1.bvelFileName.c_str(),
            bf_dvel.fn_at(ir).c_str(),param1.dimgFileName.c_str(),param1.szMigPar,cIllumBal,bf_srIl.fn_at(ir).c_str(),bf_log.fn_at(ir).c_str());
    //Send the workers the cmd
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd);usleep(int(1e5));
  }
  PRV(strCmd);
  world_.barrier(); //sync
  
  cubeData<float> velY;
  //Stack the dslw files.
  if (0) {
    strCmd = bf_dvel.stackCmd(world_.size());
    VERIFY(system(strCmd.c_str())==0);
    velY.load(param1.dslwFileName,3); usleep(int(1e5));
  } else { //Use MPI reduce to stack the final slowness
    sepHeaderReader hdr(bf_dvel.fn_at(1),dm::IN,"vel perturbation",true);
    hypercube shp_dvel;  hdr.getShape(shp_dvel,3);
    velY.alloc(shp_dvel);
    mpiY::ReduceCube_MPI(velY,PosInt(3),bf_dvel); //stack thre results using MPI_Reduce
  }
  const axis& ax_x = vel.domain().get_axis(2);
  //Now convert shape from slowY to pDeltaS, the cropping is included
  this->convertVelModel(velY,*pDeltaS,false);
  
  //read the illumination back, and divide
  if (vaPar && vaPar->bIllumBal_) {
    cubeData<float> illuTang,illuY;
    
    //stack the Illumination file
    strCmd = bf_srIl.stackCmd(world_.size());
    VERIFY(system(strCmd.c_str())==0);
    
    sprintf(szCmd,"<%s Window3d f1=50 n1=%d squeeze=0 >%s", 
            param1.srIlFileName.c_str(),ax_x.n,param1.tmpFileName.c_str());
    VERIFY(0 == system(szCmd)); usleep(int(1e5));
    illuTang.load(param1.tmpFileName,3); 
    vel.allocMemoryTo(illuY);
    this->convertVelModel(illuTang,illuY,false); //convert from Tang to Yang format
    
    int shotOffZ = this->acqPars()->shotDepth();
    //Smooth the illumination and divide
    smoothIllumBalancing(*pDeltaS,illuY,illuTang,shotOffZ);
  }
  
  //the output of yaxun's program is vel, convert dvel -> ds by dividing -v^2;
  pDeltaS->divide(vel); pDeltaS->divide(vel); (*pDeltaS).scale(-1.0);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param1.bvelFileName.c_str(),param1.dimgFileName.c_str(),param1.dslwFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

#endif

