#include "stdafx.h"
#include "param_entry.h"

#include "cubeData.hpp"
#include "param_func.h"
#include "waveOpParams.h"

using namespace Yang;

bool AcqPar::bSrcRecvLocCovered(const float* coordS, const float* coordR) {
  int ndim = vmodShape_->get_ndim();
  const vecAxis& axes = vmodShape_->get_axes();
  bool bCovered = true;
  int isz = axes[0].get_offset(coordS[0]), isx = axes[1].get_offset(coordS[1]);
  int irz = axes[0].get_offset(coordR[0]), irx = axes[1].get_offset(coordR[1]);
  int isy,iry; isy=iry=0;
  if (ndim>2) {
    isy = axes[2].get_offset(coordS[2]) ; iry = axes[2].get_offset(coordR[2]);
  }
  //PRV(srType_);
  int isxmin,isxmax; 
  int irxBeg,irxEnd; 
  if (srcType_ == STpoint) {
    isxmin = mShot1_ix; isxmax = mShot1_ix + (m_nShots-1)*mShotIncX;
    irxBeg = mTrace1ix; irxEnd = mTrace1ix + (m_nTraces-1)*mTraceIncX;
  } else if (srcType_ == STplane) {
    //if plane wave then assume src/recv are co-located, not always true.
    irxBeg = mTrace1ix; irxEnd = mTrace1ix + (m_nTraces-1)*mTraceIncX;
    isxmin = irxBeg; isxmax = irxEnd;
  } else {
    VERIFY(0);
  }
  //PRV4(isz,irz,isx,irx);  PRV2(mShot1_ix,isxmax); 
  //check if the src/recv are covered.
  if (isz != mShot1_iz) return false;
  if (irz != mTrace1iz) return false;
  if (isx < isxmin || isx > isxmax)  return false;
  else { //src x in range, need to see if receiver in range.
    irxBeg = mTrace1ix; irxEnd = mTrace1ix + (m_nTraces-1)*mTraceIncX;
    if (srType_ == splitSpread) {
      //do nothing
    } else if (srType_ == tagAlong) {
      irxBeg += isx; irxEnd+= isx;
    } else {VERIFY(0);}
    //PRV2(irxBeg,irxEnd); 
    if (irx<irxBeg || irx>irxEnd)  return false;
  }
  if (ndim>2) {
    VERIFY(0);
    if (isy < mShot1_iy || isy > mShot1_ix + (m_nShots-1)*mShotIncX)    return false;
    if (iry < mTrace1iy || iry > mTrace1iy + (m_nTraces-1)*mTraceIncY)  return false;
  }
//bSrcRecvLocCovered_exit:
  return bCovered;
}

//See if src and recv ray hits the acquisition
bool AcqPar::bSrcRecvLocCovered(const stdVecFloat& SrcRay, const stdVecFloat& RecvRay) {
  int ndim = vmodShape_->get_ndim();
  VERIFY(ndim<3);
  const vecAxis& axes = vmodShape_->get_axes();
  int nptsS = SrcRay.size()/ndim, nptsR = RecvRay.size()/ndim;
  int is,ir;
  //PRV2(mShot1_iz,mTrace1iz); axes[0].print_info("ax_z in bSrcRecvLoc");
  
  //Find the point that is at the shot depth
  for (is=0; is<nptsS; is++) {
    if (axes[0].get_offset(SrcRay[is*ndim]) == mShot1_iz) {
      break;
    }
  }
  if (is>=nptsS) {
    return false; //not found
  } else { //check to find the receiver at depth R.
    for (ir=0; ir<nptsR; ir++) {
      if (axes[0].get_offset(RecvRay[ir*ndim]) == mTrace1iz) {
        break;
      }
    }
    if (ir>=nptsR) {
      return false;
    } else{ //Now check the coords
      //PRVEC(&SrcRay[is*ndim],2);PRVEC(&RecvRay[ir*ndim],2);
      return this->bSrcRecvLocCovered(&SrcRay[is*ndim],&RecvRay[ir*ndim]);
    }
  }
}

int Yang::convert_img_shp_1way_2way(hypercube& shp_zxy, hypercube& shp_xyz_1way, bool bFrom2wayTo1way, bool b3D) {
  axis ax_dummy(1,0,1.0f);
  vecAxis axes_out;
  if (bFrom2wayTo1way) {
    const vecAxis& axes_in = shp_zxy.get_axes();
    if (!b3D) { //(z,x,hx)->(hx,hy,x,y,z)
      axes_out.push_back(axes_in[2]);axes_out.push_back(ax_dummy);axes_out.push_back(axes_in[1]);
      axes_out.push_back(ax_dummy);axes_out.push_back(axes_in[0]);
    } else { //(z,x,y,hx,hy)->(hx,hy,x,y,z)
      axes_out.push_back(axes_in[3]);axes_out.push_back(axes_in[4]);axes_out.push_back(axes_in[1]);
      axes_out.push_back(axes_in[2]);axes_out.push_back(axes_in[0]);
    }
    shp_xyz_1way.set_axes(axes_out);
  } else { //from oneway to two way
    const vecAxis& axes_in = shp_xyz_1way.get_axes();
    if (!b3D) { //(hx,hy,x,y,z)->(z,x,hx)
      axes_out.push_back(axes_in[4]);axes_out.push_back(axes_in[2]);axes_out.push_back(axes_in[0]);
    } else { //(hx,hy,x,y,z)->(z,x,y,hx,hy)
      axes_out.push_back(axes_in[4]);axes_out.push_back(axes_in[2]);axes_out.push_back(axes_in[3]);
      axes_out.push_back(axes_in[0]);axes_out.push_back(axes_in[1]);
    }
    shp_zxy.set_axes(axes_out);
  } //end if conver shp from 2way to 1way.
  return 0;
}

int Yang::convert_vel_shp_1way_2way(hypercube& shp_zxy, hypercube& shp_xyz_1way, bool bFrom2wayTo1way, bool b3D) {
  axis ax_dummy(1,0,1.0f);
  vecAxis axes_out;
  if (bFrom2wayTo1way) {
    const vecAxis& axes_in = shp_zxy.get_axes();
    if (!b3D) { //(z,x,y=1)->(x,y,z)
      axes_out.push_back(axes_in[1]);axes_out.push_back(ax_dummy);axes_out.push_back(axes_in[0]);
    } else { //(z,x,y)->(x,y,z)
      axes_out.push_back(axes_in[1]);axes_out.push_back(axes_in[2]);axes_out.push_back(axes_in[0]);
    }
    shp_xyz_1way.set_axes(axes_out);
  } else { //from oneway to two way
    const vecAxis& axes_in = shp_xyz_1way.get_axes();
    if (!b3D) { //(x,y,z)->(z,x)
      axes_out.push_back(axes_in[2]);axes_out.push_back(axes_in[0]);
    } else { //(x,y,z)->(z,x,y)
      axes_out.push_back(axes_in[2]);axes_out.push_back(axes_in[0]);axes_out.push_back(axes_in[1]);
    }
    shp_zxy.set_axes(axes_out);
  } //end if conver shp from 2way to 1way.
  return 0;
}

//update the 1way img shps from my convention of img shp axes arrangement, for oneway code from yaxun Tang, the axes are always are hx,hy,x,y,z
void ImgPar::update_img_shp_1way() {
  bool b3D = (finalImageShape_.get_ndim()==3);
  convert_img_shp_1way_2way(finalSubOffsetImageShape_, finalSubOffsetImageShape1way_, true, b3D);
  finalImageShape1way_ = finalSubOffsetImageShape1way_.return_compatible(3,5);
}

void  ImgPar::finalImageShpIs (const hypercube& finalImgShp) {
  finalImageShape_ = finalImgShp;
  get_prestack_img_cube_shape(finalSubOffsetImageShape_, offImgParams_, finalImageShape_); //update offset image shape
  update_img_shp_1way();
}

void  ImgPar::offImgParsIs    (const offsetImagingParams& offImageParam) {
  offImgParams_ = offImageParam;
  get_prestack_img_cube_shape(finalSubOffsetImageShape_, offImgParams_, finalImageShape_); //update offset image shape
  update_img_shp_1way();
}

int waveOpParams::readParam(param_func* pars) {
  bUseOnewayPropagator_ = pars->get_bool  ("bUseOneway");
  suf_MPI_              = pars->get_string("suf_MPI","");
  return 0;
}

int waveOpParams::calcTimeParamsFor2wayModeling(float maxFreq, float vmax, float vmin, const hypercube& propGridShape)
{ 
  FUNC_TRACE_BEG(waveOpParams::calcTimeParamsFor2wayModeling);
  //Because the variance of the aperture, dtComp should be recalculated for each shot, the 
	//modeling operation should be reinitialized at the begining of each shot.	
  float dx_max,dx_min;  //the min and max value in (dx,dy,dz) grid size.
  stdVecFloat dVec = propGridShape.get_ds();
  dx_max = *(std::max_element(dVec.begin(),dVec.end()));
  dx_min = *(std::min_element(dVec.begin(),dVec.end()));
  
  //assert(vmin>100.0);
  assert(vmax<1e5);
  //First check the spatial sampling
  if ((vmin / maxFreq) < 4.0 * dx_max) {
    prt("\n !WARNING, the spatial sampling is too coarse, lamda=%f, dx=%f \n", vmin/maxFreq, dx_max);
  }
  
  //Use the courant number to compute the possible dtComp;
  int ndim  =  propGridShape.get_ndim();  //Should be 2 or 3.
  assert(ndim==2 || ndim==3);
  float dtmax = 1.0/sqrt(float(ndim)) * dx_min / vmax;  // dt <= dx/v*(1/sqrt(2.0))
  //make some engineering margin
  float mdtmax = dtmax * 0.75;
  
  //compare dtmax with the given dtComp
  PosFloat& dtComp = timeParam_.dtComp;
  PosFloat& dtRec  = timeParam_.dtRecord;
  
  if (dtComp > epsilon && dtComp < mdtmax ) {mdtmax = dtComp;}
  VERIFY(!equalf(dtRec,0.0f));
  
  int n = int(dtRec/mdtmax+0.5f);
  dtComp = dtRec/n;
  
  timeParam_.Nt     =   lroundf(timeParam_.totalTime/dtComp)+1;
  timeParam_.jRec   =   lroundf(dtRec/dtComp);
  timeParam_.jSnap  =   lroundf(timeParam_.dtSnap/dtComp);
  timeParam_.jImaging = lroundf(timeParam_.dtImaging/dtComp);
  
  //Verify that there is no round-off in the dividing above
  ETYPE tdtRec = timeParam_.jRec*dtComp;
  ETYPE tdtSnap = timeParam_.jSnap*dtComp;
  ETYPE tdtImaging = timeParam_.jImaging*dtComp;
  VERIFY(equalf(tdtRec,dtRec));
  //PRV(tdtSnap); PRV(timePar->dtSnap);
  VERIFY(equalf(tdtSnap,timeParam_.dtSnap));
  //PRV(tdtImaging); PRV(timePar->dtImaging);
  VERIFY(equalf(tdtImaging,timeParam_.dtImaging));
  
  //the total number of Snapshots that will be recorded
  //fixed record at time step it=1,1+m_jSnap,1+2*m_jSnap,...
  timeParam_.nRec     = int((timeParam_.Nt-1) / timeParam_.jRec)     + 1;
  timeParam_.nSnap    = int((timeParam_.Nt-1) / timeParam_.jSnap)    + 1;
  timeParam_.nImaging = int((timeParam_.Nt-1) / timeParam_.jImaging) + 1;
  cerr  <<"Time param, T, dtRec, dtSnap, dtComp, Theory dtMax: " << timeParam_.totalTime <<','<< timeParam_.dtRecord<<','
    << timeParam_.dtSnap << ',' << timeParam_.dtComp <<',' << dtmax << endl;

  int nz = propGridShape.get_n(1); float dz = dVec[0];
  verbPrt("T should be at least %f for a round-trip from bottom reflection to be recorded\n", 
          (nz*dz*sqrt(2.0))/sqrt(vmin*vmax));
  FUNC_TRACE_END(0);
}

int waveOpParams::read_oneway_prop_param(param_func* pars) {
  OnewayOpPar& par = this->onewayPar_;
  //Read in the oneway file tags
  string sz_csou = "csou";
  if(!tagExist(sz_csou)) {
    prt("!!! Missing csou= in parameter file, make sure that src is supplied else\n");
  }
  par.crecFileName         = pars->get_string(csz_tag_dataObs,""); //!assume we use the same data every time
  par.csouFileName         = pars->get_string("csou","");
  par.sz_additional_mig_par = pars->get_string("add_migpar_1w");

  par.iw_beg_ = pars->get_int("iw_beg",0);
  par.nws_ = pars->get_int("nw_wnd",0);
  return 0;
}

int waveOpParams::read_twoway_prop_param(param_func* pars) {
  waveOpParams& params = *this;
  
  //also get the time param from the params
  params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
  params.stencilWidth = pars->get_int(csz_stenWidth);
  params.bUse4thOrder= false; //no 4th order in time.
  params.offx = params.stencilWidth;params.offz = params.stencilWidth;params.offy = params.stencilWidth;
  params.m_nPad = pars->get_int(csz_npad);  

  //Read in Absorption types
  int itype = pars->get_int(csz_bdType);
  switch (itype)
  {
   case 1:
      params.m_bdType = abs_ord1;
      break;
   case 2:
      params.m_bdType = abs_ord2;
      break;
   case 3:
      params.m_bdType = abs_Dave;
      break;
   default:
      seperr("!unknown boundary condition type!\n");
  }
  params.mImposeRecvType = AsForce;
  params.bTake2ndTimeDerivForRTM = pars->get_bool(csz_take2ndTimeDerivRTM,false);
  
  params.pSrcSnapshot = NULL;  params.pSrcSnapshot2 = NULL;
  params.pRecvSnapshot = NULL; params.pRecvSnapshot2 = NULL;
  
  //calc time parameters
  params.timeParam_.dtComp = pars->get_float(csz_dtComp);
  params.timeParam_.dtSnap = pars->get_float(csz_dtSnap);
  params.timeParam_.dtImaging = pars->get_float(csz_dtImaging,params.timeParam_.dtRecord);   
  //Always assume the recording starts from zero time. !! 
  params.timeParam_.dtRecord = pars->get_float(csz_dtRec);
  params.timeParam_.totalTime = pars->get_float(csz_totalTime);
  
  //Recording Snapshot params, only for 2D RTM
  params.f_shot_snap = pars->get_int(csz_snap_f,0);
  params.j_shot_snap = pars->get_int(csz_snap_j,1);
  params.n_shot_snap = pars->get_int(csz_snap_n,0);

  //checkSnapRecordParams(&params);
  return 0;
}

void ImgPar::readParam(param_func* pars, const hypercube& shpV_2way) {
  bool b3D = shpV_2way.get_ndim()>2;
  ImgPar& params = *this;
  //read the imaging parameters if possible, image can have a different size from velocity model
  const hypercube& shpV = shpV_2way;
  hypercube &finalImageShape = params.finalImageShape_;
  finalImageShape = shpV; //set the default size for final image.
  
  float imgxmin,imgxmax,imgzmin,imgzmax;
  float imgymin = 0.f,imgymax = 0.f;
  
  axis ax_x,ax_y,ax_z; int nx,nz,ny;
  ax_z = shpV.get_axis(1); nz = ax_z.n;
  ax_x = shpV.get_axis(2); nx = ax_x.n;
  
  imgxmin = pars->get_float(csz_img_xmin,ax_x.get_coord(0)); imgxmax = pars->get_float(csz_img_xmax,ax_x.get_coord(nx-1));
  imgzmin = pars->get_float(csz_img_zmin,ax_z.get_coord(0)); imgzmax = pars->get_float(csz_img_zmax,ax_z.get_coord(nz-1));
  float dx,dy,dz;
  dz = pars->get_float(csz_img_dz,ax_z.d);
  dx = pars->get_float(csz_img_dx,ax_x.d);
  //make sure that dx in velocity are a multiple of dx in images
  int rat_x,rat_y,rat_z;
  rat_x = lroundf(ax_x.d/dx); rat_z = lroundf(ax_z.d/dz);
  VERIFY((ax_x.d/dx-rat_x)<1e-6); VERIFY((ax_z.d/dz-rat_z)<1e-6);

  axis ax_xn, ax_yn, ax_zn;
  ax_xn = ax_x.return_truncated(imgxmin,imgxmax); ax_zn = ax_z.return_truncated(imgzmin,imgzmax);
  if (rat_z!=1) {ax_zn.d /= rat_z; ax_zn.n=(ax_zn.n-1)*rat_z+1;}
  if (rat_x!=1) {ax_xn.d /= rat_x; ax_xn.n=(ax_xn.n-1)*rat_x+1;}
  finalImageShape.set_axis(1,ax_zn); finalImageShape.set_axis(2, ax_xn);
  
  if (b3D) {
    ax_y = shpV.get_axis(3); ny = ax_y.n;
    dy = pars->get_float(csz_img_dy,ax_y.d);
    rat_y = lroundf(ax_y.d/dy);
    VERIFY((ax_y.d/dy-rat_y)<1e-6); 
    imgymin = pars->get_float(csz_img_ymin,ax_y.get_coord(0)); imgymax = pars->get_float(csz_img_ymax,ax_y.get_coord(ny-1));
    ax_yn = ax_y.return_truncated(imgymin,imgymax);
    if (rat_y!=1) {ax_yn.d/=rat_y; ax_yn.n=(ax_yn.n-1)*rat_y+1;}
    finalImageShape.set_axis(3,ax_yn);
  }
  
  //Load the offset Image parameters.
  bool bDoOffsetGather = pars->get_bool(csz_do_offset,true);
  params.offImgParams_.bDoOffsetGather = bDoOffsetGather;
  if (bDoOffsetGather) {
    params.offImgParams_.dhx = pars->get_int(csz_img_dhx);
    params.offImgParams_.nhx = pars->get_int(csz_img_nhx);
    params.offImgParams_.ohx = pars->get_int(csz_img_ohx);
    params.offImgParams_.AxesUsage = 0x1;

    params.offImgParams_.nhy = 1;
    if (b3D) { //need y now.
      params.offImgParams_.dhy = pars->get_int(csz_img_dhy);
      params.offImgParams_.nhy = pars->get_int(csz_img_nhy);
      params.offImgParams_.ohy = pars->get_int(csz_img_ohy);
      params.offImgParams_.AxesUsage |= 0x2; //bits signal. 1/2/4/8/16
    }
  }
  
  //Generate offset img shp
  get_prestack_img_cube_shape(finalSubOffsetImageShape_,offImgParams_,finalImageShape);
  update_img_shp_1way();
  //Do src&recv Intensity balancing in RTM imaging, not used so far
  params.bIllumBalancing_ = pars->get_bool("bIllumBalancing",false);
}

int AcqPar::readParam(param_func* pars, const shared_ptr<const hypercube>& vmodShape_2way) {
  this->vmodShape_ = vmodShape_2way;
  b3D_ = pars->get_bool(csz_b3D,false);
  
  //Geometry, these are based on the velocity models.
  m_nShots   = pars->get_int(csz_nShots);
  mShot1_ix  = pars->get_int(csz_shot1_ix);
  mShot1_iz  = pars->get_int(csz_shot1_iz);
  mShotIncX  = pars->get_int(csz_shotIncX);
  
  //PRV4(m_nShots,mShot1_ix,mShot1_iz,mShotIncX);

  mTrace1ix  = pars->get_int(csz_trace1_ix);
  mTrace1iz  = pars->get_int(csz_trace1_iz);
  
  mTraceIncX = pars->get_int(csz_traceIncX);
  m_nTraces  = pars->get_int(csz_nTraces);

  if (b3D_) {
    mTrace1iy  = pars->get_int(csz_trace1_iy);
    mShot1_iy  = pars->get_int(csz_shot1_iy);
    m_nShotsY  = pars->get_int(csz_nShotsY);
    mShotIncY  = pars->get_int(csz_shotIncY);
    mTraceIncY = pars->get_int(csz_traceIncY);
    m_nTracesY = pars->get_int(csz_nTracesY);
  }

  bool bMarine  = pars->get_bool(csz_bShot_offset,false); //If input CMP gather
  if (bMarine) srType_ = tagAlong;
  else         srType_ = splitSpread;
  
  std::string szSrcType;
  szSrcType =  pars->get_string("srcType","point");
  if (szSrcType.compare("point")==0) {srcType_ = STpoint;}
  else if (szSrcType.compare("plane")==0) {srcType_ = STplane;}
  else if (szSrcType.compare("random")==0) {srcType_ = STrandom;}
  else {
    VERIFY(0);
  }
  return 0;
}

