#include "stdafx.h"

#include "dmRsf.hpp"

//#undef NDEBUG
//#include "assert.h"
#include "cubeData.hpp"
#include "dmRsfDump.h"

#include "sep_params.h"
#include "stackTraceC.h"
#include "commonOps.h"
#include "param_entry.h"
#include "dampingFunc.h"
#include "misc.h"
#include <numeric>

/** 
 * Blend data sets with and without multiples 
 * Read two data sets, one is data without multiple and the other is with multiple. 
 */

//Reflection case
using namespace Yang;
using namespace std;

using boost::shared_ptr;

int main (int argc, char *argv[])
{
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  boost::shared_ptr<sep_params> prgPars(new sep_params(argc,argv));
  sep_params *pars = prgPars.get(); 
  
  //auto_ptr<cubeData<float> > imgObj(new cubeData<float>(sz_image,dm::IN,"image")); imgObj->init(NULL,3);
  hypercube shp;
  boost::shared_ptr<cubeData<float> > dataF(new cubeData<float>("in",dm::IN,"")); dataF->init(NULL,NUM_AXIS_DEFAULT);
  boost::shared_ptr<cubeData<float> > dataM(new cubeData<float>()); dataM->load("data2", NUM_AXIS_DEFAULT);
  shp = dataF->domain();
  shp.info("data shape."); //the data should be in order (t,offset,sx) or (t,rx,sx).

  VERIFY(dataM->domain().sameShape(dataF->domain()));
  boost::shared_ptr<cubeData<float> > dataO(new cubeData<float>()); dataO->alloc(shp);
  
  bool bOffset = pars->get_bool("boffset");
  SEEV(bOffset);

  LOG("c0");
  //need to set the blending wts for two datasets.
  //First the total weight,
  float Twt2 = pars->get_float("Tweight2",0.0f);
  VERIFY(Twt2>-epsilon && Twt2<1+epsilon);
  float Twt1 = 1.0f - Twt2;
  PRV2(Twt1,Twt2);
  //Then taper the data based on offset. Usually design a bigger weight for near offset and smaller weight for near offset.
  
  float refOffset = pars->get_float("refOffset",1e7); //refOffset set to very large means no attenuation
  
  shared_ptr<dampingFunc::params_df> df_par(new dampingFunc::params_df());
  df_par->readParam(pars, ""); //read parameters for the damping func.
  dampingFunc dampFunc(df_par);
  
  bool bCompensatePrimary = pars->get_bool("bCompensate1",false);

  //if not, will taper far offset.
  bool bTaperFarOffset = pars->get_bool("bTaperFarOffset");
  
  PRV(bCompensatePrimary);
  
  //dampingFunc::getCoef(stdVecFloat &coefVec, int length, float extent);
  int ndim = shp.get_ndim();
  VERIFY(ndim <= 5);
  
  const axis& ax_t = shp.get_axis(1);
  const axis &ax_2 = shp.get_axis(2);
  const axis &ax_sx = shp.get_axis(3); 
  
  int isx,i2,it; float maxOffset;
  int nt = ax_t.n;
  
  shp.info("shp");
  
  stdVecFloat s1,s2,s;  //the damping coefficients.
  int noff_neg,noff_pos; 
  if (bOffset) { //if marine geometry
    float f1,f2;
    ax_2.getRange(&f1,&f2);
    maxOffset = std::max(fabs(f1),fabs(f2));
    
    noff_neg = std::max(0, int(ax_2.get_offsetf(0.0f)-ax_2.get_offsetf(f1)+1.5f));
    noff_pos = std::max(0, int(ax_2.get_offsetf(f2) - ax_2.get_offsetf(0.0f)+1.5f));
    
    //build the damping function
    s1.resize(noff_neg); s2.resize(noff_pos);
    PRV2(noff_neg,noff_pos);
    dampFunc.getCoefEx(s1, noff_neg, float(noff_neg-1) * ax_2.d / refOffset); 
    dampFunc.getCoefEx(s2, noff_pos, float(noff_pos-1) * ax_2.d / refOffset); 
    
    //join the two vectors together
    if (noff_neg+noff_pos != ax_2.n+1) {
      PRV2(noff_neg, noff_pos); PRV(ax_2.n);
      errPrt("check the values here!\n");
    }
    s.assign(s1.rbegin(),s1.rend());
    s.insert(s.end(),s2.begin(),s2.end());
    
    dumpStdVec(s,"df.H",true,NULL);
    for (isx = 0; isx < ax_sx.n; isx++) {
      //instantiate the damping function. start from zero offset.
      if (!bCompensatePrimary) {
        for (i2 = 0; i2 < ax_2.n; i2++) {
          for (it = 0; it < nt; it++) {
            (*dataO)(isx, i2, it) = (*dataF)(isx, i2, it) * Twt1 + Twt2 * ((*dataM)(isx, i2, it)) * s[i2];
          }
        }
      } else {
        for (i2 = 0; i2 < ax_2.n; i2++) {
          for (it = 0; it < nt; it++) {
            (*dataO)(isx, i2, it) = (*dataF)(isx, i2, it) * (Twt1+1.0f-s[i2]) + Twt2 * ((*dataM)(isx, i2, it))*s[i2];
          }
        }
      }
    }
  } else { //if land geometry.
    float f1, f2, g1, g2;
    //figure out the maximum offset.
    ax_2.getRange(&f1, &f2);
    //combined with the src locations
    ax_sx.getRange(&g1,&g2);
    LOG("D2");
    float m1,m2;
    m1 = std::max(fabs(f1-g1), fabs(f1-g2)); 
    m2 = std::max(fabs(f2-g1), fabs(f2-g2)); 
    maxOffset = std::max(m1,m2);
    
    cubeDataF taperWts; 
    vecAxis axes2; //axes2.push_back(ax_t); 
    axes2.push_back(ax_2.n); axes2.push_back(ax_sx); 
    hypercube shpT(axes2); 
    shpT.info("jake");
    taperWts.alloc(axes2); 
    
    LOG("D3");
    for (isx = 0; isx < ax_sx.n; isx++) {
      axis ax_off(ax_2);
      ax_off.o -= ax_sx.get_coord(isx);
      
      //instantiate the damping function. start from zero offset.
      ax_off.getRange(&f1,&f2);
      noff_neg = std::max(0, int(ax_off.get_offsetf(0.0f) - ax_off.get_offsetf(f1) + 1.5f)); 
      noff_pos = std::max(0, int(ax_off.get_offsetf(f2)   - ax_off.get_offsetf(0.0f) + 1.5f)); 
      //build the damping function
      s1.resize(noff_neg); s2.resize(noff_pos);
      dampFunc.getCoefEx(s1, noff_neg, float(noff_neg-1) * ax_off.d / refOffset); 
      dampFunc.getCoefEx(s2, noff_pos, float(noff_pos-1) * ax_off.d / refOffset); 
      
      PRV(isx);
      //PRV2(noff_neg, noff_pos); ax_off.print_info("ax_off"); PRVEC(s1, noff_neg);  PRVEC(s2, noff_pos);
      
      //join the two vectors together
      if (noff_neg + noff_pos != ax_2.n + 1) {
        PRV2(noff_neg, noff_pos); PRV(ax_2.n);
        errPrt("check the values here!\n");
      }

      s.assign(s1.rbegin(), s1.rend() - 1); 
      s.insert(s.end(), s2.begin(), s2.end());
      // If taper near offset, then flip the Tapering function.
      if (!bTaperFarOffset) {
        for (int i=0; i<s.size(); i++) {
          s[i] = 1.0f - s[i];
        }
      } 
      
      for (i2 = 0; i2 < ax_off.n; i2++) {
        taperWts(isx,i2) = s[i2];
        if (!bCompensatePrimary) {
          for (it = 0; it < nt; it++) {
            (*dataO)(isx, i2, it) = (*dataF)(isx, i2, it) * Twt1 + Twt2 * ((*dataM)(isx, i2, it)) * s[i2];
          } 
        } else {
          for (it = 0; it < nt; it++) {
            //(*dataO)(isx, i2, it) = (*dataF)(isx, i2, it) * (Twt1+1.0f-s[i2]) + Twt2 * ((*dataM)(isx, i2, it)) * s[i2];
            //1.6f is the boost factor, since the multiple-free data (no ghost) is weaker than the ghost data
            (*dataO)(isx, i2, it) = 1.6f*(*dataF)(isx,i2,it) * (Twt1+Twt2*(1.0f-s[i2])) + Twt2*s[i2]*((*dataM)(isx, i2, it));
            //another extreme, amplify the far-offset
            //(*dataO)(isx, i2, it) = (*dataF)(isx,i2,it) + 2.0*(1.0f-s[i2])*((*dataM)(isx, i2, it));
          }
        }
      }
    } // end for isx
    //the dump the taper wts.
    taperWts.dump("df.H",true);
  } //end if bOffset=0;
  
  dataO->dump("out");
  return 0;
} //end main

