#include "stdafx.h"
#include <fstream>

#include <unistd.h>
#include "dmRsf.hpp"

#include "cubeData.hpp"
#include "srcWav.h"
#include "point3d.hpp"

#include "param_entry.h"

#include "stackTraceC.h"
#include "commonOps.h"
#include "util_seplib.h"
#include "misc.h"
#include "cubeAlgo2.h"

#include "dmHeaderReader.h"
#include "my_timer.h"

/**
 * Stacking the cubes from many small mpi jobs to one final result.
 * Could work for images/gathers/vmodel cubes. 
 */

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;

/// Usage
/// Combine.x <header.H oe1=o1,e1 oe2=... filelist=flist.txt ndim= output=out.H
/// It will combine all files in the flist.txt, one for each line. The files in flist.txt does not have to be aligned in coordinates. But needs to have same number of dimensions.The output grid will use the provided header.H file's information
//  ndim sets the number of axis we exepect in the data files.
/// oe1,oe2... will allow you to adjust the output grid on the beginning,and end pt coordinates for each axis.

int main (int argc, char *argv[])
{  
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  //initialize the temporary path
  //pars->get_tmp_path();

  boost::shared_ptr<std::ifstream> fstrm_desc; //the file pointer to the desc files
  sepHeaderReader header("in"); //output domain

  hypercube shape;
  int ndim = pars->get_int("ndim",-1);
  if (ndim==-1) {
    header.getShape(shape,true);
  } else {
    header.getShape(shape,ndim);
  }
  ndim = shape.get_ndim();
  VERIFY(ndim<=5);

  //get dimensions
  int idim=1; char sz_cmd[2048]; int nread;
  vecAxis axes = shape.get_axes(); float f_def2[2] = {-1,-1};
  for (idim=1;idim<=ndim;idim++) {
    sprintf(sz_cmd,"oe%d",idim);
    float* pOE = pars->get_floats(sz_cmd,2,f_def2,&nread); //beginning and ending of certain axis
    axis ax = axes[idim-1];
    if (nread==2) {
      ax = axes[idim-1].return_truncated_ext(pOE[0],pOE[1]);
      axes[idim-1] = ax;
    } //otherwise use default axis dimension, do nothing
  }

  LOG(A1);
  hypercube shpOut(axes), shpOutSlice;
  hypercube* pShpOutSlice = &shpOut;

  bool bSliceBySlice=false;
  axis ax5(1); int iLastAxis = 5; int nslice=10,dslice; axis ax5_slice;
  
  float f_memMB = Yang::getPhysMemorySize()*0.5f;
  float f_maxMemUsed = shpOut.get_n123()*sizeof(float)*2*0.001f;
  if (f_maxMemUsed > f_memMB) { //if the total size is bigger than 8GB, divide the domain to 10 pieces.
    ax5 = axes[ndim-1];
    VERIFY(ax5.n>1);
    bSliceBySlice = true; 
    nslice = int(f_maxMemUsed/f_memMB)+1;
    dslice = (ax5.n-1)/nslice+1;
    ax5_slice = ax5; ax5_slice.n = dslice;
    axes[ndim-1] = ax5_slice; shpOutSlice.set_axes(axes);
    pShpOutSlice = &shpOutSlice;
  } else {
    nslice = 1;
  }

  shape.info("total input shp info");  shpOut.info("total output shp info");
  bool verb = pars->get_bool("verb",true);

  cubeDataF cubeOut(*pShpOutSlice);
  string fn = pars->get_string("filelist"); string str_fpname;
  LOG(A2);
  fstrm_desc.reset(new std::ifstream(fn.c_str(),std::ifstream::in));
  string str_ln, str_tmp;
  LOG(A3);
  int iline = 0; stringstream ss; cubeDataF cube;
  VERIFY(tagExist("output"));
  dmRsf_oc<float> dm_out("output",dm::OUT,"dm_out");
  dm_out.init(&shpOut);

  int islice = 0;
  vector<bool> bSkipFileVec;
  for (islice=0; islice<nslice; islice++) {
    iline = 0; cubeOut.zero();
    //if last slice, need to change the size of cubeOut
    if (islice==nslice-1 && bSliceBySlice) {
      axis ax_last = ax5_slice;
      ax_last.n = ax5.n-islice*dslice;
      VERIFY(ax_last.n>1);
      pShpOutSlice->set_axis(ndim,ax_last);
      cubeOut.alloc(*pShpOutSlice);
      cubeOut.info("last cubeout info");
    }
    PRV(islice);
    
    fstrm_desc->clear(); fstrm_desc->seekg(0,std::ios_base::beg);
    while (std::getline(*fstrm_desc,str_ln)) { //if read not end
      //parse this line, should be single filename
      ss.clear(); //clear the status for stringstream, this is important, otherwise ss only can parse one single line
      ss.str(str_ln);
      if (str_ln[0]=='#' || str_ln.empty()) { //ignore this line, regarding it as comment
        continue;
      }
      iline++;
      str_fpname = str_ln;
      if (islice==0) {
        bSkipFileVec.push_back(false); //only allocate the space for vec bSkipFile in the first run
        //first check if the file is a valid sep file
        sprintf(sz_cmd,"In %s | grep '100\\./' ",str_fpname.c_str()); //Attr takes too much time
        string str_out = systemY(sz_cmd);
        int stat = system(sz_cmd);
        if (stat!=0 || str_out.size()<10) { //the binary part or the header part is not correct
          PRT_LOC; prt("!!!! NaN, sth is wrong with file %s, it might not be a valid seplib data file!",str_fpname.c_str());
          bSkipFileVec[iline-1] = true;
        }
        if (verb) {
          PRV2(sz_cmd,str_out);
        }
      }
      
      //PRV(iline); PRVEC(bSkipFileVec,bSkipFileVec.size()); PRV(bSliceBySlice);
      float rms=0.f, coord5 = 0.f;
      bool bCropAlign = !bSkipFileVec[iline-1];
      if (!bSkipFileVec[iline-1]) {
        if (!bSliceBySlice) {
          cube.load(str_fpname,ndim);
        } else { //only read a single slice of it, and updates it
          dmRsf_oc<float> dm_cube_in(str_fpname,dm::IN,"",true);
          dm_cube_in.init(NULL,ndim);
          const vecAxis& axes_in = dm_cube_in.domain().get_axes();
          vecAxis axes_inslice(axes_in); axes_inslice.pop_back();
          axes_inslice.push_back(ax5_slice);
          hypercube shapeInSlice(axes_inslice);
          coord5 = ax5.get_coord(islice*dslice);
          //find the slice corresponding to coord5
          float off_f = axes_in[ndim-1].get_offsetf(coord5);
          int   off_i = lround(off_f);
          if (verb) PRV2(coord5,off_f);
          if (fabs(off_f-off_i) > 0.1f) VERIFY(0);
          windParam wnd(ndim);
          int wf,wj,wn;
          if (off_i>=0 && off_i<=(axes_in[ndim-1].n-1)) { //if within range
            wnd[ndim-1].f = off_i;
            wnd[ndim-1].j = lround(ax5.d/axes_in[ndim-1].d);
            wf = wnd[ndim-1].f; wj = wnd[ndim-1].j;
            wnd[ndim-1].n = std::min(dslice,int((axes_in[ndim-1].n-wf-1)/wj)+1);
            wn = wnd[ndim-1].n;
            //the last slice might not have the same size as before
            if (wn < dslice) {
              ax5_slice.n = wn;
              shapeInSlice.set_axis(ndim,ax5_slice);
              cube.alloc(shapeInSlice);
              ax5_slice.n = dslice; //restore ax5_slice
            } else {
              cube.alloc(shapeInSlice);
            }
            if (verb) {
              dm_cube_in.domain().info("dm_cube_in.domain"); cube.domain().info("cubein.domain");
            }
            PRV4(ndim,wf,wj,wn);
            dm_cube_in.grabWindow(&cube,wnd);
          } else {
            bCropAlign = false;
          }
        }

        if (bCropAlign) {
          rms = cube.RMS();
          if (isnan(rms)) {
            prt("!! NaN exception on file %s\n", str_fpname.c_str());
            bSkipFileVec[iline-1] == true; //mark this file, so to skip it every time 
          }
          if (!bSkipFileVec[iline-1]) {
            cropAlign(cubeOut,cube,true,false);
            prt("rms val=%f on file %s at slice %d\n", rms,str_fpname.c_str(),islice);
          }
        }
      } //bSkipFileVec
    } //while read each fileshot file
    dm_out.seq_update(&cubeOut); //update the output file
  } //islice
  //cubeOut.dump("output",true);
  SEEV(iline);
  return 0;
}

  
