#include "stdafx.h"
#include <fstream>
#include <unistd.h>

#include "dmRsf.hpp"
#include "cubeData.hpp"

#include "param_entry.h"

#include "stackTraceC.h"
#include "commonOps.h"
#include "util_seplib.h"
#include "misc.h"
#include "cubeAlgo2.h"

#include "sep_params.h"
#include "dmHeaderReader.h"
#include "my_timer.h"

/**
 * Use it for Hessian Balancing. Can handle large file
 * Usage: Divide.x f1=numerator.H f2=denom.H eps=0.2 >output.H 
 * output = f1/(f2+eps) 
 */

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;

int main (int argc, char *argv[])
{  
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();

  sepHeaderReader header("f1"); //output domain
  hypercube shape,shp_f2;
  int ndim = pars->get_int("ndim",-1);
  if (ndim==-1) {
    header.getShape(shape,true);
  } else {
    header.getShape(shape,ndim);
  }
  ndim = shape.get_ndim();
  VERIFY(ndim<=5);

  dmRsf_oc<float> dm_cube_f1("f1",dm::IN,"",true); dm_cube_f1.init(NULL,ndim);
  dmRsf_oc<float> dm_cube_f2("f2",dm::IN,"",true); dm_cube_f2.init(NULL,ndim);
  shp_f2 = dm_cube_f2.domain();
  VERIFY(shape.sameShape(shp_f2));
  float feps = pars->get_float("eps");

  bool verb = pars->get_bool("verb",true);
  if (verb) {
    PRV(ndim);
    shape.info("f1 shp"); shp_f2.info("f2 shp");
  }

  vecAxis axes = shape.get_axes();
  hypercube shpOut(shape), shpOutSlice;
  hypercube* pShpOutSlice = &shpOut;

  bool bSliceBySlice=false;
  axis ax5(1); int nslice=10,dslice; axis ax5_slice; 
  if (shpOut.get_n123()>2e9) { //if the total size is bigger than 8GB, divide the domain to 10 pieces.
    ax5 = axes[ndim-1];
    VERIFY(ax5.n>1);
    bSliceBySlice = true; 
    dslice = (ax5.n-1)/nslice + 1;
    ax5_slice = ax5; ax5_slice.n = dslice;
    axes[ndim-1] = ax5_slice; shpOutSlice.set_axes(axes);
    pShpOutSlice = &shpOutSlice;
  } else {
    nslice = 1;
  }

  if (verb) {
    pShpOutSlice->info("ShpSlice info");
  }

  cubeDataF cubeOut(*pShpOutSlice),cube1,cube2;
  dmRsf_oc<float> dm_out("out",dm::OUT,"dm_out");
  dm_out.init(&shpOut);

  int islice = 0;

  for (islice=0; islice<nslice; islice++) {
    PRV(islice);
    cubeOut.zero();
    //if last slice, need to change the size of cubeOut
    if (islice==nslice-1 && bSliceBySlice) {
      axis ax_last = ax5_slice;
      ax_last.n = ax5.n-islice*dslice;
      VERIFY(ax_last.n>1);
      pShpOutSlice->set_axis(ndim,ax_last);
      cubeOut.alloc(*pShpOutSlice);
      if (verb) {
        cubeOut.info("last cubeout info");
      }
    }

    cube1.alloc(*pShpOutSlice); cube2.alloc(*pShpOutSlice);
    
    //load the input data
    if (!bSliceBySlice) {
      dm_cube_f1.seq_grab(&cube1);
      dm_cube_f2.seq_grab(&cube2);
    } else { //only read a single slice of it, and updates it
      if (verb) {
        dm_cube_f1.domain().info("dm_cube_f1.domain");
      }
      int wf = islice*dslice, wj=1, wn=pShpOutSlice->get_n(ndim);
      PRV4(ndim,wf,wj,wn);
      windParam wnd(ndim); wnd[ndim-1].n=wn;wnd[ndim-1].j=wj; wnd[ndim-1].f=wf;
      dm_cube_f1.grabWindow(&cube1,wnd);
      dm_cube_f2.grabWindow(&cube2,wnd);
    } //end if bSliceBySlice
    
    //Then do the division
    long long N = pShpOutSlice->get_n123();
    float fo;
    for (long long i=0;i<N;i++) {
      fo = cube1[i]/(cube2[i]+feps);
      if (isnan(fo)) {
        PRV2(i,cube2[i]);
        VERIFY(0);
      }
      cubeOut[i] = fo;
    }
    dm_out.seq_update(&cubeOut); //update the output file
  } //for islice
  return 0;
}


