#include "stdafx.h"
#include "axis.h"
#include "dmRsf.hpp"
#include "dmRsfDump.h"

#include "sep_params.h"

#include "stackTraceC.h"
#include "commonOps.h"

#include "algo.h"
#include "cubeAlgo2.h"

/*
 * Gen msk to mask out multiple on the parabolic radon-domain.
 */

using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;

int main (int argc, char *argv[]) {
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  shared_ptr<cubeDataF> vmod(new cubeDataF());
  
  vmod->load("in",NUM_AXIS_DEFAULT); //load in the radon domain size.
  //Then generate multiple msk
  axis ax_semb_inter, ax_depth_inter; //make a small msk in the range defined below.
  float zmin = pars->get_float("depthmin"), zmax = pars->get_float("depthmax");
  float semb_min = pars->get_float("semb_min"), semb_max = pars->get_float("semb_max");

  //Now create a small mask.
  const hypercube& shp = vmod->domain();
  cubeDataF msk(shp);

  //input axis are z,semb,x
  axis ax_z = shp.get_axis(1); axis ax_semb = shp.get_axis(2);
  int n3 = shp.get_n123()/shp.get_n23(1,2);
  int izmin = ax_z.get_offset(zmin), izmax = ax_z.get_offset(zmax);
  int isembmin = ax_semb.get_offset(semb_min), isembmax = ax_semb.get_offset(semb_max);

  if (!(izmin>=0 && izmax<ax_z.n && isembmin>=0 && isembmax<ax_semb.n )) {
    PRV4(izmin,isembmin,isembmin,isembmax);
    VERIFY(0);
  }

  msk = 1.0f;
  int n12 = ax_z.n*ax_semb.n;
  for (int i3=0; i3<n3; i3++) {
    for (int is=isembmin; is<=isembmax; is++) {
      for (int iz=izmin; iz<=izmax; iz++) {
        msk[i3*n12 + is*ax_z.n + iz] = 0.0f; 
      }
    }
  }
  msk.dump("out"); 
}

