#include "stdafx.h"

#include "dmRsf.hpp"

//#undef NDEBUG
//#include "assert.h"
#include "dmHeaderReader.h"
#include "cubeData.hpp"
#include "dmRsfDump.h"

#include "sep_params.h"
#include "stackTraceC.h"
#include "commonOps.h"
#include "param_entry.h"
#include "misc.h"
#include <numeric>

/**
 * For beam focusing, Wave Equation Tomography
 * This prog generates the random sampling mask for the subsurface common image gather.
 */

//Reflection case
using namespace Yang;
using namespace std;

   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  boost::shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  //auto_ptr<cubeData<float> > imgObj(new cubeData<float>(sz_image,dm::IN,"image")); imgObj->init(NULL,3);
  sepHeaderReader hdr("in",dm::IN,"");
  hypercube imageShape; //I just need the shape
  hdr.getShape(imageShape,NUM_AXIS_REMOVE_DUMMY_AXIS);
  
  imageShape.info("original image dimension"); //the image should be of axis order (z,x,hx)

  LOG("c0");
  //Here, need to gen the mask as the random sparse sampling of the migrated gather.
  cubeData<float>* pImageMask = new cubeData<float>();
  
  LOG("c1");
  pImageMask->alloc(imageShape); //Set the size of the image mask.
  LOG("c2");
  pImageMask->zero();
  LOG("c3");
  
  //Read the downsampling parameters.
  int ndim = imageShape.get_ndim(), idim;
  VERIFY(ndim <= 5);
  int nRead = 0; 
  float fReadDef[5]; for (int i=0; i<5; i++) {fReadDef[i]=1.0f;}
  float* ar_ds_rates = pars->get_floats("ds_rates",5,fReadDef,&nRead); 
  PRV(nRead); PRVEC(ar_ds_rates,5);
  for (int i=0; i<5; i++) {fReadDef[0]=0;}
  int nRead2=0;
  int iReadDef[5]; for (int i=0; i<5; i++) {iReadDef[i]=1;}
  int* ar_ds_origins = pars->get_ints("ds_os",5,iReadDef,&nRead2); 
  PRV(nRead2); PRVEC(ar_ds_origins,5);
  
  LOG("c4");
  stdVecInt ns_old = imageShape.get_ns();
  long n123 = imageShape.get_n123();
  
  stdVecInt ns_new_ext(5,1), ns_old_ext(5,1);
  for (int i=0; i<ns_old.size(); i++) {ns_old_ext[i] = ns_old[i];}  

  //If all zero offset needs to be kept for sure.  
  bool bKeepAllZeroOffset = pars->get_bool("bKeepAllZeroOffset",false);
  PRV(bKeepAllZeroOffset);
  int  iAxisOff;
  if (bKeepAllZeroOffset) {
    iAxisOff = pars->get_int("iAxisOff");
    VERIFY(iAxisOff>0);
  }
  long index;

  stdVecInt ns_new(ns_old); stdVecFloat ds_new(ns_old.size(),1.0f);
  //Determin the dimension
  for (int i=0; i<nRead; i++) {
    ns_new[i] = int((ns_old[i]-1-ar_ds_origins[i])/(ar_ds_rates[i]))+1;
    ds_new[i] = ar_ds_rates[i];
  }
  for (int i=0; i<ns_new.size(); i++) {ns_new_ext[i] = ns_new[i];}
  PRVEC(ns_new,ns_new.size()); PRVEC(ns_old,ns_old.size());
  //Generate mask
  string rand_method = pars->get_string(csz_tag_img_rand_samp_method);
  PRV(rand_method);
  
  unsigned int seed = time(NULL);
  srand(seed);
  int i4,i3,i2,i1,i0; int ind4,ind3,ind2,ind1,ind0; int off4,off3,off2,off1,off0;

  stdVecFloat ds_new_ext(5,1); 
  for (int i=0; i<ds_new.size(); i++) {ds_new_ext[i] = ds_new[i];}
  
  PRVEC(ns_new_ext,5); PRVEC(ds_new_ext,5);
  PRVEC(argv,argc);

  int iter=0;
  int nIter = pars->get_int("nIter",1);
  vecAxis axes = imageShape.get_axes();
  axis ax_iter(nIter,0,1.0f);
  axes.push_back(ax_iter);
  hypercube shpTl(axes);
  Yang::dmRsf_ocDump<float> mskTl("out","");
  mskTl.init(&imageShape);
  
  for (iter=0; iter<nIter; iter++) {
    pImageMask->zero();
    if (bKeepAllZeroOffset) {
      int i0off_axisOff = imageShape.get_axis(iAxisOff).get_offset(0.0f);
      //Mark all zero offset points to be 1
      int is[5];
      for (is[4]=0; is[4] < ns_old_ext[4]; is[4]++) {
      for (is[3]=0; is[3] < ns_old_ext[3]; is[3]++) {
      for (is[2]=0; is[2] < ns_old_ext[2]; is[2]++) {
      for (is[1]=0; is[1] < ns_old_ext[1]; is[1]++) {
      for (is[0]=0; is[0] < ns_old_ext[0]; is[0]++) {
        if (is[iAxisOff-1]==i0off_axisOff) {
            index = (((is[4]*ns_old_ext[3]+is[3])*ns_old_ext[2]+is[2])*ns_old_ext[1]+is[1])*ns_old_ext[0]+is[0];
            (*pImageMask)[index] = 1.0f;
        }
      }}}}}
    }

  if (rand_method.compare("nr") == 0) { //not random
    //Gen it on regular grid, no randomness
    vector< stdVecInt >  indexVec(ndim); 
    for(idim=0; idim < ndim; idim++) {
      indexVec[idim].resize(ns_new[idim]);
      for (int i=0; i<ns_new[idim]; i++) {
        indexVec[idim][i] = lround(ar_ds_origins[idim]+ds_new[idim]*i);
      }
      //PRVEC(indexVec[idim],ns_new[idim]);
    }
    vector< stdVecInt > indexVecExt(5,std::vector<int>(1,0));
    for (int i=0; i<indexVec.size(); i++) {indexVecExt[i] = indexVec[i];}
    PRVEC(ns_new_ext,5); PRVEC(ns_old_ext,5);

    //The set the masks.
    for (i4=0; i4 < ns_new_ext[4]; i4++) {
      ind4 = indexVecExt[4][i4];
      off4 = ind4;
    for (i3=0; i3 < ns_new_ext[3]; i3++) {
      ind3 = indexVecExt[3][i3];
      off3 = off4 * ns_old_ext[3] + ind3;
    for (i2=0; i2 < ns_new_ext[2]; i2++) {      
      ind2 = indexVecExt[2][i2];
      off2 = off3 * ns_old_ext[2] + ind2;
    for (i1=0; i1 < ns_new_ext[1]; i1++) {      
      ind1 = indexVecExt[1][i1];
      off1 = off2 * ns_old_ext[1] + ind1;
    for (i0=0; i0 < ns_new_ext[0]; i0++) {      
      ind0 = indexVecExt[0][i0];
      off0 = off1 * ns_old_ext[0] + ind0;
      if (off0 >= imageShape.get_n123()) {
        PRV4(ind4,ind3,ind2,ind1); PRV(ind0);
        VERIFY(0);
      }
      (*pImageMask)(off0) = 1.0f; 
    }}}}}
  }
  else if (rand_method.compare("br")==0) { //big random
    //Gen it using uniform sampling on the large grid.
    int N = accumulate (ns_old.begin(), ns_old.end(), 1, std::multiplies<int>()) / (ar_ds_rates[0]*ar_ds_rates[1]*ar_ds_rates[2]*ar_ds_rates[3]*ar_ds_rates[4]); //Find the total number of points.
    int tmpi; 
    for (int ip=0; ip<N; ip++) {
      index = 0;
      for(idim=ndim-1; idim >= 1; idim--) { //for each axis, find a value
        tmpi = lround(float(rand())/float(RAND_MAX) * ns_old[idim] - 0.5f);
        index += tmpi;
        index = index * ns_old[idim-1];
      }
      index += lround(float(rand())/float(RAND_MAX) * ns_old[0] - 0.5f);
      //Set the value for that coord
      if ((*pImageMask)(index)<0.5f) {
        (*pImageMask)(index) = 1.0f; 
      } else {
        ip--;
      }
    }
  } else if (rand_method.compare("rdx")==0) { //random only on specific axis
    //only be random on specif axis or axes
    int n_ax_read;
    int* iAxes = pars->get_ints("axes",3,&n_ax_read);
    VERIFY(n_ax_read>=1);
    SEEV(n_ax_read);
    //build a random mask for those dimensions
    vecAxis subAxes;
    for (idim=0; idim<n_ax_read; idim++) {
      VERIFY(iAxes[idim]>0 && iAxes[idim]<=ndim);
      subAxes.push_back(imageShape.get_axis(iAxes[idim]));
    }
    hypercube shpSub(subAxes);
    cubeDataF cubeSubt(shpSub),cubeSubSamp(shpSub);

    //Now invoke this program itself (!!Be extra careful of this recursive invoke!!) to generate a random msk on its subcube size.
    string szthisTool(argv[0]);
    SEEV(szthisTool);

    stringstream ss;
    string sub_rand_method = pars->get_string("sub_rand_method");
    VERIFY(sub_rand_method.compare("rdx")!=0);

    for (int i=1; i<argc; i++) {
      ss << " " << argv[i] << " ";
    }
    ss << csz_tag_img_rand_samp_method << "=" << sub_rand_method << " bKeepAllZeroOffset=n " << " nIter=1 ";
    
    PRV(ss.str());
    seplib::callSepTool(cubeSubSamp,cubeSubt,szthisTool,ss.str(),false);

    cubeSubSamp.dump("dump1");

    long indexsub=0;
    int offs_sub[n_ax_read];

    int isubdim;
    LOG("dd1"); PRVEC(ns_old_ext,5);

    for (i4=0; i4 < ns_old_ext[4]; i4++) {
      for (isubdim=0; isubdim<n_ax_read; isubdim++) { if (iAxes[isubdim]==5) offs_sub[isubdim]=i4;}
    for (i3=0; i3 < ns_old_ext[3]; i3++) {
      for (isubdim=0; isubdim<n_ax_read; isubdim++) { if (iAxes[isubdim]==4) offs_sub[isubdim]=i3;}
    for (i2=0; i2 < ns_old_ext[2]; i2++) {
      for (isubdim=0; isubdim<n_ax_read; isubdim++) { if (iAxes[isubdim]==3) offs_sub[isubdim]=i2;}
    for (i1=0; i1 < ns_old_ext[1]; i1++) {
      for (isubdim=0; isubdim<n_ax_read; isubdim++) { if (iAxes[isubdim]==2) offs_sub[isubdim]=i1;}
    for (i0=0; i0 < ns_old_ext[0]; i0++) {
      for (isubdim=0; isubdim<n_ax_read; isubdim++) { if (iAxes[isubdim]==1) offs_sub[isubdim]=i0;}
      //check if the point in the subcube is masked or not
      // PRVEC(offs_sub,n_ax_read); 
      indexsub = shpSub.get_index1d(offs_sub, n_ax_read);

      if (cubeSubSamp[indexsub]>0.5f) {
        index = (((i4*ns_old_ext[3]+i3)*ns_old_ext[2]+i2)*ns_old_ext[1]+i1)*ns_old_ext[0]+i0;
        (*pImageMask)[index] = 1.0f;
      }
    }}}}}
  } else if (rand_method.compare("sr")==0) { //subrandom
    //Gen it by randomly position a point in each coarse grid region
    index = 0;
    for (i4=0; i4 < ns_new_ext[4]; i4++) {
    for (i3=0; i3 < ns_new_ext[3]; i3++) {
    for (i2=0; i2 < ns_new_ext[2]; i2++) {      
    for (i1=0; i1 < ns_new_ext[1]; i1++) {      
    for (i0=0; i0 < ns_new_ext[0]; i0++) {      
      //determine the location of the sample points.
      if    (ns_new_ext[4]==1) off4=0;
      else  off4 = lround( (float(rand())/float(RAND_MAX) + float(i4))*ds_new[4] - 0.5f);
      if    (ns_new_ext[3]==1) off3=0;
      else  off3 = lround( (float(rand())/float(RAND_MAX) + float(i3))*ds_new[3] - 0.5f);
      if    (ns_new_ext[2]==1) off2=0;
      else  off2 = lround( (float(rand())/float(RAND_MAX) + float(i2))*ds_new[2] - 0.5f);
      if    (ns_new_ext[1]==1) off1=0;
      else  off1 = lround( (float(rand())/float(RAND_MAX) + float(i1))*ds_new[1] - 0.5f);
      if    (ns_new_ext[0]==1) off0=0;
      else  off0 = lround( (float(rand())/float(RAND_MAX) + float(i0))*ds_new[0] - 0.5f);
      index = (((off4*ns_old_ext[3]+off3)*ns_old_ext[2]+off2)*ns_old_ext[1]+off1)*ns_old_ext[0]+off0;
      (*pImageMask)(index) = 1.0f;
    }}}}}   
  } else if (rand_method.compare("pr")==0) { 
    //Gen by Poisson Disk sampling maybe...
    cerr<<"poisson Disk sampling not implemented yet !!!!!"<<endl; VERIFY(0);
  } else if (rand_method.compare("bb")==0) { //use bob
    
    int n1 = ns_old_ext[0],n2 = ns_old_ext[1];
    float pct = pars->get_float("pct",50.0f); //default sample only half of gathers.
    hypercube shpff(1000,n1); //n1 is depth dimension, the sampling is denser at deeper part.
    cubeDataF ff(shpff),func(shpff);
    hypercube shpff2(n2,n1); //The order of axis 1 and 2 is reversed
    cubeDataF ff2(shpff2); stdVecFloat vpow(n1);
    
    for (i1=0;i1<n1;i1++) vpow[i1] = 0.0f; //0.0f+float(n1-i1)/float(n1);
    
    int count=0; float sm;
    // SY2 
    for(i2=0;i2<n1;i2++) {
      for(i1=0;i1<n2;i1++) {
        ff(i2,i1)=powf(std::max(0.f,sinf(i1*PI/float(n2))),vpow[i2]);
      }
      sm=std::accumulate(&ff(i2,0),&ff(i2,n1-1)+1,0.f,std::plus<float>());
      for (i1=0; i1<n2; i1++) ff(i2,i1)=ff(i2,i1)*999.f/sm;
    }
    
    for(i2=0;i2<n1;i2++) {
      for(i1=1;i1<n2;i1++) {
        ff(i2,i1) += ff(i2,i1-1);
      }
    }
    

    // SY3
    for(i2=0;i2<n1;i2++) {
      for(i1=0;i1<n2;i1++) {
        for(int ii=max(0,int(lroundf(ff(i2,i1))));ii<1000;ii++) func(i2,ii) = i1;
      }
    }
  
    ///BobNew <
    float mymax,mysq,ww;
    float f_hfnh = lroundf(float(n2)/2); PRV(f_hfnh);

    float trRatio = pars->get_float("trRatio",0.4f); //the shape of the Trapazoid
    for(i2=0;i2<n1;i2++) {
      mymax=f_hfnh*trRatio+(f_hfnh*(1.0f-trRatio))/float(n1)*float(i2+1);
      mysq=float(500*500);
      for (i1=0; i1<500; i1++) {
        ww=mymax*powf(float(i1)/500.0f,1.5f);
        ff(i2,499-i1)=std::max(1,int(lroundf(f_hfnh-ww)));
        ff(i2,500+i1)=std::min(n2-1,int(lroundf(f_hfnh+1.0f+ww)));
      }
    }
    ///BobNew >

    ff.dump("ff");  func.dump("func");
    
    cubeDataF& mask = (*pImageMask);
    float rnd[3]; int n34 = n123/n1/n2;
    PRV(n34); PRV(n123); 
    count = 0;

    bool do_cone = pars->get_bool("do_cone",true);
    if (pct<=50) {
      mask = 0.0f;
      //This is doing the random-subsampling of the encoded offset-gather data
      while (count < pct/100.f*n123) {
        for (int ii=0;ii<3;ii++) {rnd[ii]=float(rand())/float(RAND_MAX);}
        //PRVEC(rnd,3);
        i3=std::max(0,std::min(n34-1, int(lroundf(rnd[0]*n34))));  //choose gather among (1,n3*n4)
        i1=std::max(0,std::min(n1-1,  int(lroundf(rnd[1]*n1))));    //among (1,n1)
        
        if (do_cone) {
          i2=std::max(0,std::min(999,int(lroundf(rnd[2]*999.f)))); //Draw a
          //i2=func(i1,i2); ///bobOld
          i2=int(ff(i1,i2));  ///bobNew
          if (i2==0) {
            PRV2(i1,i2);
            PRV(ff(i1,i2));
            VERIFY(0);
          }
        } else {
          i2=std::max(0,std::min(n2-1,int(lroundf(rnd[2]*n2))));
        }
        i4=i1+(i2)*n1+(i3)*n1*n2;
        if (i4<0 || i4 > n123) {
          PRV4(i1,i2,i3,i4); 
          VERIFY(0);
        }
        if(mask[i4]<0.5f) {
          mask[i4]=1.0f;
          count=count+1;
        }
        if((count%10000)==0) prt("found %d, %2.1f pct\n",count,pct/100.0f*n123);
      }
    } else {
      mask = 1.0f;
      while (count < long((100.f-pct)/100.f*n123)) {
        for (int ii=0;ii<3;ii++) {rnd[ii]=rand()/RAND_MAX;}
        i3=max(0,min(n34-1,(int)lroundf(rnd[0]*n34)));  //choose gather among (1,n3*n4)
        i1=max(0,min(n1-1,(int)lroundf(rnd[1]*n1)));    //among (1,n1)
        i2=func(i1,std::max(0,std::min(999,int(lroundf(rnd[2]*999.f)))));
        i4=i1+(i2)*n1+(i3)*n1*n2;
        if(mask[i4]>0.5f) {
          mask(i4)=0.0f;
          count=count+1;
        }
      }
      if((count%10000)==0) prt("found %d, %2.1f pct\n",count,pct/100.0f);
    } //end if pct<50
  } else {
    cerr<<"Unknown method param !!!!!"<<endl; VERIFY(0);
  }
  
  mskTl.dumpFrame(*pImageMask); 
  } //end for each iterations.
  pImageMask->unInit();
  return 0;
}

