#include "stdafx.h"
#include "dmRsf.hpp"

#include "cubeData.hpp"
#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "rayTracer.h"
#include "RayTomoOp.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"
#include "misc.h"

#include "my_mpi.h"
#include "mpi_util.h"
#include "dmHeaderReader.h"
#include "waveOpParams.h"
#include "waveOpMPI.h"

#include "my_timer.h"

#include <mpi.h>
/**
 * Oneway migration using Tang's code.
 * Real 3D, migrate small data file chunks.
 */

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;
//Doing beam wave-equation tomography for reflection case.

#include <unistd.h>
//Do born modeling
int main (int argc, char *argv[])
{
  mpiY::environment MPI_Init(&argc,&argv);
  mpiY::communicator world;
  
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  //initialize the temporary path
  pars->get_tmp_path();

  int rank,nw;
  rank  = world.rank();
  nw    = world.size();

  if (rank==0) {
    LOGC("in process 0");
    PRV(world.size());
    PRVEC(argv,argc);
    char hostname[1024];
    gethostname(hostname,1024);
    SEEV(hostname);
  }
  
  //important, fix the stdin/stdout if no is provided. Otherwise, MPI program will hang!!
  if (rank==0) mpi_sep_send_args(nw,10,0);
  else mpi_sep_receive_args();
  
  //if (rank==0) {
  if (0) {
    for (int ir=1; ir<nw; ir++) {
      string sz("sleep 10");
      world.send(ir,WRKTAG,sz);
      prt("Sending msg %s to node %d\n", sz.c_str(), ir);
    }
    
    char sz_cwd[512];
    getcwd(sz_cwd,511);
    PRV(sz_cwd);
    PRVEC(argv,argc);
    cubeDataF tool2Cube;
    strcpy(sz_cwd, "/net/server/homes/sep/yang/Research/beamWET2/bp/sepToolTemp7537.H");
    //strcpy(sz_cwd, "vel");
    tool2Cube.load(sz_cwd,3);
    //tool2Cube.load("s1",3);
    tool2Cube.info("t2c");
    //tool2Cube.dump("tt159.H",true);
  }
  
  //if (0) { //master node, execute all the main logic
  if (world.rank()==0) { //master node, execute all the main logic
    Yang::timer clk;
    LOGC("in root process");
    //PRVEC(argv,argc);
    //PRV(world.size());
    initCallStackTrace(); 
    init_omp(0);
    //Read params, params need contain geometry info that is exact the same as the input data. In
    //synthetic this is easy to do.
    //shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
    //sep_params* pars = autoPars.get();
    
    shared_ptr<vel>  velObj     (new vel(csz_tag_vel,dm::IN,"vel"));
    shared_ptr<srcWav> waveletObj (new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
    //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
    //waveletObj->loadWavelet(); //if oneway, no need to load wavelet
    //velObj->loadVel(3); velObj->info("Initial velocity model");
    velObj->peekVelShape(3);

    shared_ptr<ImgPar> imagingPar(new ImgPar());  ImgPar& imgPar = *imagingPar;
    //load acquisition parameters, here use shotsInfo not AcqPar
    shared_ptr<AcqPar> acqPar(new AcqPar()); 
    //acqPar->readParam(pars,velObj->domain());

    shared_ptr<waveOpParams>  waveOpPar(new waveOpParams()); waveOpParams& params = *waveOpPar;
    params.readParam(pars);
    bool bUseOneWay = params.bOneway();
    shared_ptr<hypercube> vmodShp_2way(new hypercube());
    if (bUseOneWay) {
      convert_vel_shp_1way_2way(*vmodShp_2way, const_cast<hypercube&>(velObj->vmodShp()), false, true);
    } else {
      *vmodShp_2way = velObj->vmodShp();
    }
    vmodShp_2way->info("vmodshp_2way");

    shared_ptr<shotFilesInfo> SshotsInfo(new shotFilesInfo());
    string fname_shotsInfo = pars->get_string("data_desc_file");
    LOG("A1.1");
    int ishot_beg = pars->get_int("ishot_beg",0);
    int ishot_end = pars->get_int("ishot_end",-1); //-1 means take all shots in the file
    PRV2(ishot_beg,ishot_end);
    SshotsInfo->parsefile(fname_shotsInfo,vmodShp_2way,ishot_beg,ishot_end);
    LOG("A1.2");
    hypercube imgshp0off_2way = *vmodShp_2way;
    float xmin,xmax,ymin,ymax,zmin,zmax;
    const axis& axx = vmodShp_2way->get_axis(2); axx.getRange(&xmin,&xmax);
    const axis& axz = vmodShp_2way->get_axis(1); axz.getRange(&zmin,&zmax);
    const axis& axy = vmodShp_2way->get_axis(3); axy.getRange(&ymin,&ymax);

    SshotsInfo->getGlobalImgDomain(xmin, xmax, ymin, ymax, zmin, zmax);
    axis axn_x(axx), axn_y(axy), axn_z(axz);
    axn_x.o = xmin; axn_x.n = int((xmax-axn_x.o)/axn_x.d+0.5f) + 1;
    axn_y.o = ymin; axn_y.n = int((ymax-axn_y.o)/axn_y.d+0.5f) + 1;
    axn_z.o = zmin; axn_z.n = int((zmax-axn_z.o)/axn_z.d+0.5f) + 1;
    imgshp0off_2way.set_axis(1,axn_z);
    imgshp0off_2way.set_axis(2,axn_x);
    imgshp0off_2way.set_axis(3,axn_y);

    //If the user specified the imaging domain, then use that instead
    float f_def2[2] = {-1,-1}; int nRead; axis axI;

    float* pOEx = pars->get_floats("xs",2,f_def2,&nRead);
    if (nRead==2) { //Due to the large size, we only save the overlap region of GlobalImgDomain and targe Image domain
      axI = axn_x.return_truncated(pOEx[0],pOEx[1]); imgshp0off_2way.set_axis(2,axI);
    } else {VERIFY(nRead==0);}

    float* pOEy = pars->get_floats("ys",2,f_def2,&nRead);
    if (nRead==2) {axI = axn_y.return_truncated(pOEy[0],pOEy[1]); imgshp0off_2way.set_axis(3,axI);}
    else {VERIFY(nRead==0);}

    float* pOEz = pars->get_floats("zs",2,f_def2,&nRead);
    if (nRead==2) {axI = axn_z.return_truncated(pOEz[0],pOEz[1]); imgshp0off_2way.set_axis(1,axI);}
    else {VERIFY(nRead==0);}

    imgshp0off_2way.info("global image_2way shape");
    imgPar.readParam(pars,imgshp0off_2way); //Load imaging parameters
    LOG("A1.3");
    const hypercube& imgOffShp = imgPar.finalSubOffImageShp1way();
    imgOffShp.info("imgOffShp_1way");

    LOG("A2");
    shared_ptr<waveOpOneway> modelingObj;
    srand(time(NULL)); //set a random generate intermediate filenames.
    //Read in the parameters for the modeling Operator
    LOG("A3.5");
    const string vmod_fname = pars->get_string("vel");
    //Here we just focus on use one-way
    if (bUseOneWay) {
      PRV(velObj->domain().isEmpty());
      LOG("B1");
      modelingObj.reset(new waveOpOnewayMPI3d(waveOpPar,acqPar,imagingPar,*velObj,pars,SshotsInfo));
      LOG("B2");
      params.read_oneway_prop_param(pars);
      LOG("B3");
      modelingObj->vmodFileNameIs(vmod_fname);
      LOG("B4");
    } else {
      VERIFY(0);
      params.read_twoway_prop_param(pars);
      //calc time parameters
      float maxFreq,vmax,vmin;
      //Read in the peak/center freq of the seismic wavelet
      maxFreq = waveletObj->get_float("peakFreq");
      waveletObj->setPeakFreq(maxFreq);
      vmax = velObj->maxval();  vmin = velObj->minval();
      params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
    }
    LOG("A4")
    hypercube shp_cdata;
    bool bBornModeling = pars->get_bool("bornFwd"); //true means modeling, otherwise imaging.
    if (!bBornModeling) { //migration
      LOG("B5");
      //Migration
      cubeDataF img(imgOffShp);
      LOG("B6");
      clk.reset("start migration"); clk.go();
      modelingObj->migration_op("",&img);
      clk.reset("finish migration");
      LOG("B7");
      PRT_LOC; img.info("img");
      img.dump("imgh");
    } else { //born modeling
      bool b3D = acqPar->b3D();
      cubeDataF img;
      if (!b3D) {
        img.load("refl",3);
      } else {
        VERIFY(0);
      }
      VERIFY(0);
      string fn_crec = params.onewayPar()->crecFileName;
      modelingObj->bornModeling_op(&img,fn_crec);
    }
    LOG("A6");
    //release the slave modes, send exit cmd.
    for (int ir=1; ir<world.size(); ir++) {
      world.send(ir,EXITAG,string("Exit"));
    }
    LOG("A7");
    clk.check("Migration finish");
  }
  
  if (world.rank() > 0) { //the slaver node, just accept cmds, execute it, and wait for another cmd. Keep in mind that the master process is also a slave node.
    mpiY::slaveModeMPI3d slaveMode;
    slaveMode.on();
    prt("exit slaveMode, irank=%d \n",world.rank());
  }
  ////finally synchronize, then exit.
  world.barrier();
  return 0;
}

