#include "stdafx.h"
#include "axis.h"
#include "dmRsf.hpp"
#include "dmRsfDump.h"
#include "sep_params.h"
#include "param_entry.h"
#include "stackTraceC.h"
#include "commonOps.h"
#include "misc.h"
///Read parameters
#include "waveletCompress_f.h"
#undef NDEBUG
#include <assert.h>
#include "algo.h"
#include "cubeAlgo2.h"
#include "ADCIG_Recon.h"
#include "cdstep.h"
#include "linSolver.h"
#include "istc_solver.h"


#define DEBUG
#include "myDebug.h"

/** Use parabolic radon transform to reconstruct angle gather
 *  and do the muting. Input migrated offset gather(might be
 *  random subsampled, then invert for the model space and mute
 *  the mutiples and recon it back.
 *   */

//Reflection case
using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;

int main(int argc, char *argv[]) {
  initCallStackTrace();
  init_omp(1);

  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc, argv));
  sep_params *pars = autoPars.get();

  int   n[4], i, j;
  int   nc, count, niter_stomp, niter_lin;

  shared_ptr<cubeDataF> datMsk(new cubeDataF()), modMsk(new cubeDataF());
  //float dot1[2],dot2[2],val;

  LOG("c1");
  cubeDataF dm, offd, angd;
  cubeDataF dat_in, waved, dd, m0, ma, wt2, dco;

  dat_in.load("in", NUM_AXIS_REMOVE_DUMMY_AXIS); const hypercube &shpIn = dat_in.domain(); //currently will handle 4 dimensions at most
                                                                                      //Notice that the cube is arranged in h,z,x, therefore do transpose first
  cubeDataF dat;
  int ndim = shpIn.get_ndim(); VERIFY(ndim == 3);

  int ar[3] = { 1, 2, 3 }, map[3] = { 2, 1, 3 };
  vector<int> reshape(ar, ar + 3), mapping(map, map + 3);
  shuffleAxis(dat, dat_in, reshape, mapping, false, true);
  //then data is arranged in the correct order

  const hypercube &shp = dat.domain();

  n[0] = n[1] = n[2] = n[3] = 1;
  for (i = 0; i < ndim; i++) {
    n[i] = shp.get_n(i + 1);
  }
  LOG("c2");

  //Init hyperbolic radon operator.
  shared_ptr<opRadon2D> radonOp(new opRadon2D());
  //Now read in the semblance axis, this should be read from the mod-msk
  cubeDataF modMskMute; modMskMute.load("modMskMute", ndim);

  const hypercube &modShp = modMskMute.domain();
  axis ax_semb = modShp.get_axis(2);
  ax_semb.print_info("axis semblance");
  radonOp->init(modShp, shp, Yang::Quadratic, 1.0f);

  LOG("c2.5"); PRVEC(n, 4);
  waved.alloc(modShp); dd.alloc(shp); dco.alloc(shp);

  const axis &ax_z    = shp.get_axis(1);
  const axis &ax_off  = shp.get_axis(2); //the second axis is offset.
  const axis &ax_x    = shp.get_axis(3);
  ax_off.print_info("ax_h");

  dmRsf_ocDump<float> datDump("datDump", "rsfDump"); datDump.init(&shp);
  dmRsf_ocDump<float> modDump("modDump", "rsfDump"); modDump.init(&modShp);
  float pct;

  LOG("c3");
  //Test is over, now do the reconstruction inversion.
  LOG("c4");
  modMsk->alloc(modShp); (*modMsk) = 1.0f;
  offd.alloc(shp); offd = dat;

  //susb_sample the input data
  dat.info("input offsetImg data");
  if (tagExist("subSampMask")) {
    datMsk->load("subSampMask", ndim); VERIFY(datMsk->domain().sameShape(shp));
    dat.multiply(*datMsk);
    pct = datMsk->sum() / float(datMsk->domain().get_n123()) * 100.0f; PRV(pct);  //how much percent of subsampling, can be figure out from the mask
  } else {
    datMsk->alloc(shp); (*datMsk) = 1.0f;
    prt("No subsample mask present.\n");
  }
  datDump.dumpFrame(dat); //dump the subsampled data.

  LOG("c5");

  niter_stomp = pars->get_int("niter_stomp", 5);
  niter_lin   = pars->get_int("niter_lin", 10);
  TRACE(("niters big/small_lin %d:%d. \n",niter_stomp,niter_lin));

  LOG("c5.2");
  long long n123 = shp.get_n123();
  LOG("c5.5");

  //Now join the two operator together. datMask, (RadonOp_fwd -> ModMask)
  shared_ptr<opMask>      mmaskOp(new opMask(modShp)); mmaskOp->maskIs(modMsk);
  shared_ptr<opMask>      dmaskOp(new opMask(shp)); dmaskOp->maskIs(datMsk);
  shared_ptr<opChain20>   Bop(new opChain20(radonOp, mmaskOp));
  shared_ptr<opChain20>   stomp_it_op1(new opChain20(dmaskOp, radonOp));
  shared_ptr<op>   stomp_it_op(stomp_it_op1);
  shared_ptr<op> pFop, pStomp_backPrj_op;

  bool bUseAngle = false;
  if (bUseAngle) {VERIFY(0);} else { //get the backprj op
    pFop = Bop;
    pStomp_backPrj_op = stomp_it_op1;
  }

  op &Fop = *pFop;
  op &stopm_backPrj_op = *pStomp_backPrj_op;

  int belowCnt;float cut;
  int nd = shp.get_n123(), nm = Fop.ModShape().get_n123();
  cdstepper cdstep(nm, nd);
  LOG("c6"); PRV2(nm, nd);
  bool bSparse = false;
  if (bSparse) {modMsk->zero();} else {*modMsk = 1.0f;}

  int imethod = 2;
  if (imethod == 1) {
    float addRatio;
    dd  = dat; //residual rd = dat - 0;
    m0  = 0.0f; //set clear the model to be inverted.
    for (i = 1; i <= niter_stomp; i++) {

      if (bSparse) {
        //First map the data residual $dd$ into model space $waved$; call set_direction(.true.)
        LOG("d2");
        stopm_backPrj_op.oper(waved, dd, true, false);
        //compute the energy envelope of waved, to avoid gaps in the cube
        ma.info("ma"); waved.info("waved"); waved.dump("Dump.H", true);
        LOG("d2.3");
        seplib::callSepTool(ma, waved, "Envelope ", "tc1=3 tc2=2 tc3=3");
        modDump.dumpFrame(ma);
        //dd.info("dd"); waved.info("waved");
        LOG("d2.5");
        //now find hard clip-off
        ma.multiply(wt2); cubeAbs(ma, ma);
        addRatio  = 0.4f * pct / 100.f / (1. * niter_stomp); //according to the phase transition diagram.
        belowCnt  = int(nm * (1.0f - addRatio) + 0.5f); //0.1f is a controling coef, can be changed.
        cut       = quantile(belowCnt, ma.dataPtr(), nm); PRV(cut);
        LOG("d4");
        //begin to do inversion.
        //update modMsk value, decide where to invert in the model space.
        count = 0;
        for (j = 0; j < nm; j++) {
          if (fabsf(ma(j) * wt2(j)) >= cut) {
            if ((*modMsk)(j) < 0.5f) {
              (*modMsk)(j) = 1.0f;
              count++;
            }
          }
        }
        prt("cuttoff:%f,%d,%d,fitting LS err=%f, count pct =%f,should be %f\n",
            cut, belowCnt, count, dotProd(dd, dd), count / float(nm), addRatio);
      }
      //dd = dat;Fop.oper(m0,dd,false,true);dd.info("datAfterFop");
      //waved.dump("dump1"); waved.info("Fop adj Cube output");

      //opA2O->oper(m0,dat,false,true);dat.info("datAfterA2O");
      //The model mask is in modMsk array, which is embeded in Fop
      //dat.info("data"); datMsk->info("datmsk");
      solver(Fop, cdstep, nm, nd, waved.dataPtr(), dat.dataPtr(), niter_lin, datMsk->dataPtr(), NULL, m0.dataPtr()); //begin from m0=0

      LOG("d5");
      //after inverted model is solved, final model is in waved.
      m0 = waved;
      //double check
      count = 0; float fmax = waved.maxvalAbs();
      for (j = 0; j < nm; j++) {
        if (fabsf(waved[j]) > fmax * 1e-6f) {
          count++;
          if ((*modMsk)(j) < 0.5f) {
            prt("!!! Illogical value for msk array, ind=%d", j);
          }
        }
      }
      prt("COUNT TOTAL:%d, over addRation %f \n", count, count / float(nm) / addRatio);

      //LOG("c6"); //recalculate the data residual, dd = dat-dco*dMsk;
      stopm_backPrj_op.oper(waved, dco, false, false);
      dd = dat; dco.multiply(*datMsk); dd.subtract(dco);
      //datDump.dumpFrame(dd); //dump out the data fitting residual.
    } //end for niter_stomp

    LOG("b1");
    modDump.dumpFrame(*modMsk); modDump.dumpFrame(waved); //this is inverted wavelet domain model representation.
  } else if (imethod == 2) {

    waved = 0.f;
    dd = dat; offd = dat;

    PRVEC(n, 4);
    //std::cerr << "CHECK PCT" << pct <<sep<< n123 <<sep<< waved.domain().get_n123() << endl;
    m0.zero();

    //LOG("Istc_solver STOMP 1");
    long long nx;
    nx = modShp.get_n123(); nd = shp.get_n123();
    int nlIter = niter_stomp; float fEigenvalue = 5.0f;
    PRV2(nlIter, niter_lin);

    bool bUseModelMsk = pars->get_bool("bUseModMsk", false); PRV(bUseModelMsk);
    if (bUseModelMsk) {modMsk->load("modMsk_in", modShp.get_ndim());}

    Yang::solver(*stomp_it_op, cdstep, nm, nd, waved.dataPtr(), dat.dataPtr(), niter_lin * nlIter,
                 NULL, NULL, NULL, NULL, NULL, 1, modMsk->dataPtr()); //begin from m0=0

    //Yang::istcSolver(*stomp_it_op,waved.dataPtr(),nx,dat.dataPtr(),nd,nlIter,niter_lin,va_ptr(level),level.size(),va_ptr(start),start.size(),NULL,NULL,bUseModelMsk,msk_model.dataPtr(), true);
    //if (!bUseModelMsk) {modMsk->dump("modMsk_out");}
    //After the inversion.
    modDump.dumpFrame(waved);
    //waveCompress->oper(waved,dat,false,false);
  } // end if imethod

  LOG("b4"); //Reconstuct the offset gather
  if (bUseAngle) {} else { //from off gather construct angle gather
    LOG("d1");
    //multiple the muteMsk in model domain to remove the model
    waved.multiply(modMskMute);
    modDump.dumpFrame(waved);

    //go back to data domain.
    radonOp->oper(waved, offd, false, false); //This is offset gather now
    LOG("d2");
    offd.multiply(*datMsk);

    //now do the transpose, because guowemva's program use different axis order
    cubeDataF offdTr;
    shuffleAxis(offdTr, offd, reshape, mapping, false, true);
    offdTr.dump("out");
    //datDump.dumpFrame(offd);
    LOG("d3");
    //waved.info("waved");  offd.info("offd");  angd.info("angd");
    //angd.dump("out");
  }
} //end main

