#include "stdafx.h"
#include "axis.h"
#include "dmRsf.hpp"
#include "dmRsfDump.h"
#include "sep_params.h"
#include "param_entry.h"
#include "stackTraceC.h"
#include "commonOps.h"
#include "misc.h"
///Read parameters
#include "waveletCompress_f.h"
#undef NDEBUG
#include <assert.h>
#include "algo.h"
#include "cubeAlgo2.h"
#include "ADCIG_Recon.h"
#include "cdstep.h"
#include "linSolver.h"
#include "istc_solver.h"


#define DEBUG
#include "myDebug.h"

/*
 * Input migrated offset gather(might be random subsampled,
 then compute the objective function and calculate the back projected data.
 */
//Reflection case
using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;


/// Use sparse sampled ODCIG to reconstructed ADCIG
int main (int argc, char *argv[]) {
  initCallStackTrace();
  init_omp(1);
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  int   n[4],i,j;
  fwk::degree oAng,dAng;
  int   nc,count,niter_stomp,niter_lin;
  //float zero,rnd[3];err,
  float cut,amax,eps,pct;
  shared_ptr<cubeDataF> datMsk(new cubeDataF()),modMsk(new cubeDataF());
  //float dot1[2],dot2[2],val;
  
  LOG("c1");
  
  cubeDataF dm,offd,angd;
  cubeDataF dat,waved,dd,m0,ma,wt2,dco;
  //double  sqrt_3 = sqrt(3.0); double  denom  = 4.0 * sqrt( 2.0 );
  //float   tol,delta,gamma;
  dat.load("in",NUM_AXIS_REMOVE_DUMMY_AXIS); const hypercube& shp = dat.domain(); //currently will handle 4 dimensions at most

  n[0] = n[1] = n[2] = n[3] = 1;
  int ndim = shp.get_ndim();
  for (i=0;i<ndim;i++) {
    n[i] = shp.get_n(i+1);
  }
  
  LOG("c2");
  
  float rgh0[9] ={.037828,-.023849,-.110624,.377402,.852699,.377402,-.110624,-.023849,.037828};
  float sm0[7]  ={-.064539,-.040689,.418092,.788486,.418092,-.040689,-.064539};
  int lev1,lev2,lev3,lev4;
  //int levels[4] = {3,2,4,0}; //for offest domain
  int levels[4] = {3,2,4,0};   //for angle domain

  lev1=levels[0],lev2=levels[1],lev3=levels[2],lev4=levels[3]; //the transform on axis 4 should not be performed.

  //Init wavelet operator
  shared_ptr<opWaveletComp> waveCompress(new opWaveletComp());
  waveCompress->init(sm0,rgh0,7,9,n[0],n[1],n[2],n[3],lev1,lev2,lev3,lev4);
  waveCompress->setShape(shp,shp);
  
  LOG("c2.5"); PRVEC(n,4);
  //call dot_test(wavelet_transform_op,n1*n2*n3*n4,n1*n2*n3*n4,dot1,dot2)
  shp.info("imageShape");
  waved.alloc(shp);m0.alloc(shp);ma.alloc(shp);
  dd.alloc(shp);dco.alloc(shp);
  
  TRACE(("CHECK HERE, dat.RMS_NZ = %f \n",dat.RMS_NZ()));
  
  const axis& ax_z    = shp.get_axis(1);
  const axis& ax_off  = shp.get_axis(2); //the second axis is offset.
  const axis& ax_x    = shp.get_axis(3);
  ax_off.print_info("ax_h");
  
  bool bUseAngle = pars->get_bool("bAngleDomain"); PRV(bUseAngle);

  string suf("ang");
  axis ax; read_axis_from_par(ax,pars,suf,&n[1]); //the angle is in degree
  
  //Initialize off2ang operators
  axis_deg ax_gamma(ax); //make sure it is in degreee.
  //ax_gamma.print_info("ax_gamma"); PRV2(oAng,nAng); PRV(dAng);
  
  bool b3D = false, bOneway = true;
  shared_ptr<Param_o2a> par_o2a(new Param_o2a(ax_gamma, ax_off,slantstackYangAA,b3D,bOneway));
  shared_ptr<hypercube> mShp;
  vecAxis axes(ndim); axes[0] = ax_z; axes[1] = par_o2a->ax_gamma(); axes[2] = ax_x;

  int idim;
  for (idim=4; idim<=ndim; idim++) {
    axes[idim-1] = shp.get_axis(idim);
  }
  hypercube angShp(axes);
  if (bUseAngle) {
    mShp.reset(new hypercube(axes)); //angle gather shape  
  } else {
    mShp.reset(new hypercube(shp)); //off gather shp
  }
  
  const hypercube& modShp = *mShp;
  dmRsf_ocDump<float> datDump("datDump","rsfDump"); datDump.init(&shp);
  dmRsf_ocDump<float> modDump("modDump","rsfDump"); modDump.init(&modShp);
  dmRsf_ocDump<float> oDump("oDump","outputDump");  oDump.init(&shp);
  LOG("c3");

  
//if (par_o2a->bOnewayCompensate_) { //divide the angle gather by cos^2(gamma) to balance the amplitude.
//  axis_deg ax_p(img_ax_t.domain().get_axis(2));
//  float fang;
//  cubeData<float> cosr(ax_p);
//  for (int ia=0; ia<ax_p.n; ia++) {
//    fang = deg2rad(ax_p.get_coord(ia));
//    cosr(ia) = (SQ(1.0/cosf(fang)));
//  }
//  //cosr.dump("dump10"); cosr.info("cosr");
//  multiplySubCube(img_ax_t,img_ax_t,cosr,2,2);
//}
  //Test the wavelet transform reversability.
  //int ierr;
  bool bTestTF = false; //test wavelet transform
  if (bTestTF) {
    LOG("c3.1");
    //Then compress to wavelet domain.
    waveCompress->set_wavelet_direction(true);
    waveCompress->oper(waved,dat,true,false);
    waved.dump("wavelet");
    LOG("c3.5");
    //Convert back from wavelet
    waveCompress->set_wavelet_direction(false);
    waveCompress->oper(waved,dco,false,false);
    datDump.dumpFrame(dco); //reconstructed angle gather.
    exit(0);
  }
  
  
  //Test is over, now do the reconstruction inversion.
  LOG("c4");
  waved.alloc(modShp);m0.alloc(modShp);ma.alloc(modShp);angd.alloc(angShp);modMsk->alloc(modShp);
  offd.alloc(shp); offd = dat;
  
  //sub_sample the input data
  dat.info("input offsetImg data");
  if (tagExist("subSampMask")) {
	  datMsk->load("subSampMask",ndim); VERIFY(datMsk->domain().sameShape(shp));
    dat.multiply(*datMsk);
  	pct = datMsk->sum()/float(datMsk->domain().get_n123())*100.0f; PRV(pct);  //how much percent of subsampling, can be figure out from the mask
  } else {
	  prt("No subsample mask present.\n");
    VERIFY(0);
  }
  datDump.dumpFrame(dat); //dump the subsampled data.
  eps   = pars->get_float("eps",1e-10); amax=pars->get_float("amax",60.0f);
  
  LOG("c4.5");
  //Reinitialize the wavelet_compress_operator
  for (i=0;i<ndim;i++) n[i] = modShp.get_n(i+1); PRVEC(n,4);
  waveCompress->init(sm0,rgh0,7,9,n[0],n[1],n[2],n[3],lev1,lev2,lev3,lev4);
  LOG("c4.6");
  waveCompress->setShape(modShp,modShp);waveCompress->set_wavelet_direction(true);
  
  LOG("c5");
  //Init to Off2ang operator
  shared_ptr<OpOff2Ang2d> opO2A(new OpOff2Ang2d());  
  const hypercube& datShp = shp;   //off gather
  opO2A->setShape(datShp,angShp); //set Shape, here the dat and mod definition is reversed, because fwd is Off to Ang.
  shared_ptr<opAdj> opA2O(new opAdj(opO2A)); //The adj is angle to offset.
  
  bool bTestTF2 = true; //test wavelet transform, this time with angle gather domain
  if (bTestTF2) {
    //Convert from off to ang
    opO2A->oper(offd,angd,false,false);
    //Then compress to wavelet domain.
    angd.dump("dump3");
    //waveCompress->set_wavelet_direction(true);
    //waveCompress->setShape(modShp,modShp);
    //waved.alloc(angShp);  waveCompress->oper(waved,angd,true,false); waved.dump("wavelet");
  }

  niter_stomp = pars->get_int("niter_stomp",5);
  niter_lin   = pars->get_int("niter_lin",10);
  TRACE(("niters big/small_lin %d:%d. \n",niter_stomp,niter_lin));

  LOG("c5.1");  
  ///////////////////////////////////////////////////////////////////////
  waveCompress->set_wavelet_direction(true);
  m0 = dat;
  LOG("c5.2");
  //waved.info("waved"); m0.info("m0");
  ///waveCompress->oper(waved,m0,true,false); oDump.dumpFrame(waved); 
  LOG("c5.3");
  //*
  long long n123 = shp.get_n123();
  valarray<float> lvls(n123),stt(10000);
  valarray<int> level(n123);

  PRV(level.max());
  //call perc_it(8.,(/n1,n2,n3,n4/),(/4,2,1,1/),lvls,level,stt);
  perc_it_f03(8.0f,n,levels,va_ptr(lvls),va_ptr(level),va_ptr(stt),stt.size());
  
  LOG("c5.4");
  PRV(level.size()); PRV(va_ptr(level));PRV(level.max());
  
  valarray<float> start(level.max());

  PRV(start.size()); stt.size();
  for (i=0;i<start.size();i++) { start[i] = 0.3*powf(stt[i],0.33f); }
  PRV(start.size()); PRVEC(start,start.size());

  bool do_level = pars->get_bool("do_level",true);

  LOG("c5.5");
  //Now join the two operator together. A2O -> waveComp -> ModMask
  shared_ptr<opMask>      mmaskOp(new opMask(modShp)); mmaskOp->maskIs(modMsk);
  shared_ptr<opMask>      dmaskOp(new opMask(datShp)); dmaskOp->maskIs(datMsk);
  shared_ptr<opChain20>   Bop(new opChain20(waveCompress,mmaskOp));
  shared_ptr<opChain20>   stomp_it_op1(new opChain20(dmaskOp,waveCompress)); 
  shared_ptr<op>   stomp_it_op;

  shared_ptr<op> pFop,pStomp_backPrj_op;
  if (bUseAngle) {
    pFop.reset(new opChain20(opA2O,Bop));
    pStomp_backPrj_op.reset(new opChain20(opA2O,waveCompress)); //this operator does not include model mask.
    stomp_it_op.reset(new opChain20(dmaskOp,pStomp_backPrj_op));
  } else {
    pFop = Bop;
    pStomp_backPrj_op = waveCompress; //this operator does not include model mask.
    stomp_it_op.reset(new opChain20(dmaskOp,waveCompress)); 
  }

  op& Fop = *pFop;
  op& stopm_backPrj_op = *pStomp_backPrj_op;
  
  wt2.alloc(modShp); wt2=1.0f; m0=0.0f; //Not sure how to calculate the wt2

  int belowCnt; int nd=shp.get_n123(), nm=Fop.ModShape().get_n123();
  cdstepper cdstep(nm,nd); 
  LOG("c6"); PRV2(nm,nd);

  int imethod=2;
  if (imethod==1) {
    LOG("c5.7");
    //call dot_test(stomp_oper,size(waved),size(waved),dot1,dot2)
    //call dot_test(stomp_it,size(waved),size(waved),dot1,dot2)
    //call dot_test(wavelet_transform_op,size(waved),size(waved),dot1,dot2)
    //call dot_test(offset_op,size(waved),size(waved),dot1,dot2)
    /// Something is not fully understand here. If set direction to false, the reconstruction err is big (remains at 25% of the initial error) and cannot be reduced, however, if not use set_direction(.false.), then the fitting gets much better, almost converge to zero
    
    float addRatio; 
    dd  = dat; //residual rd = dat - 0;
    m0  = 0.0f; //set clear the model to be inverted.
    
    for (i=1; i<=niter_stomp; i++) {
      //First map the data residual $dd$ into model space $waved$; call set_direction(.true.)
      LOG("d2");
      stopm_backPrj_op.oper(waved,dd,true,false);
      
      //compute the energy envelope of waved, to avoid gaps in the cube
      ma.info("ma"); waved.info("waved"); waved.dump("Dump.H",true);

      LOG("d2.3");
      seplib::callSepTool(ma,waved, "Envelope ","tc1=3 tc2=2 tc3=3");
      modDump.dumpFrame(ma);
      //dd.info("dd"); waved.info("waved");
      
      //opA2O->oper(waved,dd,true,false); dd.info("dd2"); waved.info("waved2");
      //modDump.dumpFrame(waved);
      
      LOG("d2.5");
      //now find hard clip-off
      ma.multiply(wt2); cubeAbs(ma,ma);
      addRatio  = 0.4f*pct/100.f/(1.*niter_stomp); //according to the phase transition diagram.
      belowCnt  = nm*(1.0f-addRatio); //0.1f is a controling coef, can be changed.
      cut       = quantile(belowCnt,ma.dataPtr(),nm); PRV(cut);
      LOG("d4");
      //begin to do inversion.
      //update modMsk value, decide where to invert in the model space.
      count = 0;
      for (j=0;j<nm;j++) {
        if(fabsf(ma(j)*wt2(j)) >= cut) {
          if ((*modMsk)(j)<0.5f) {
            (*modMsk)(j) = 1.0f;
            count++;
          }
        }
      }
      prt("cuttoff:%f,%d,%d,fitting LS err=%f, count pct =%f,should be %f\n",
          cut,belowCnt,count, dotProd(dd,dd), count/float(nm),addRatio);
      //dd = dat;Fop.oper(m0,dd,false,true);dd.info("datAfterFop");
      //waved.dump("dump1"); waved.info("Fop adj Cube output");
      
      //opA2O->oper(m0,dat,false,true);dat.info("datAfterA2O");
      //The model mask is in modMsk array, which is embeded in Fop
      //dat.info("data"); datMsk->info("datmsk");
      solver(Fop,cdstep,nm,nd,waved.dataPtr(),dat.dataPtr(),niter_lin,datMsk->dataPtr(),NULL,m0.dataPtr()); //begin from m0=0
      
      LOG("d5");
      //after inverted model is solved, final model is in waved.
      m0 = waved;
      //double check
      count=0; float fmax = waved.maxvalAbs();
      for(j=0;j<nm;j++) {
        if (fabsf(waved[j])>fmax*1e-6f) {
          count++;
          if((*modMsk)(j)<0.5f) {
            prt("!!! Illogical value for msk array, ind=%d", j);
          }
        }
      }
      prt("COUNT TOTAL:%d, over addRation %f \n",count,count/float(nm)/addRatio);
      
      //LOG("c6"); //recalculate the data residual, dd = dat-dco*dMsk;
      stopm_backPrj_op.oper(waved,dco,false,false);
      dd=dat; dco.multiply(*datMsk); dd.subtract(dco); 
      //datDump.dumpFrame(dd); //dump out the data fitting residual.
    } //end for niter_stomp
    
    LOG("b1");
    modDump.dumpFrame(*modMsk); modDump.dumpFrame(waved); //this is inverted wavelet domain model representation.
  } else if (imethod==2) {
    LOG("I DIEE22");
    ////call mask1_init(msk)
    
    waved=0.f;
    dd=dat; offd=dat;
    
    ///call weight_init(wet)
    waveCompress->set_wavelet_direction(false);

    LOG("I DIEE62");
    PRVEC(n,4);
    std::cerr << "CHECK PCT" << pct <<sep<< n123 <<sep<< waved.domain().get_n123() << endl;
    m0.zero();
    stdVecInt myc(10);
    for (i=0;i<2;i++)   myc[i]=4;
    for (i=2;i<5;i++)   myc[i]=3;
    for (i=5;i<7;i++)   myc[i]=2;
    for (i=7;i<8;i++)   myc[i]=1;
    for (i=8;i<10;i++)  myc[i]=0;
    
    ///call init_levels((/n1,n2,n3,n4/),(/4,2,1,1/),myc)
    init_levels_f03(n,levels,4,va_ptr(myc),myc.size());

    LOG("Istc_solver STOMP 1");
    if (!do_level) start=-1;
    long long nx;
    nx = modShp.get_n123(); nd = shp.get_n123();
    int nlIter = niter_stomp; float fEigenvalue = 5.0f;
    PRV2(nlIter,niter_lin);
    
    bool bUseModelMsk = pars->get_bool("bUseModMsk",false); PRV(bUseModelMsk);
    cubeDataF msk_model(modShp);
    if (bUseModelMsk) {
      msk_model.load("modMsk_in",modShp.get_ndim());
    }

    Yang::solver(*stomp_it_op,cdstep,nm,nd,waved.dataPtr(),dat.dataPtr(),niter_lin*nlIter,
                 NULL,NULL,NULL,NULL,NULL,1,msk_model.dataPtr()); //begin from m0=0

    //Yang::istcSolver(*stomp_it_op,waved.dataPtr(),nx,dat.dataPtr(),nd,nlIter,niter_lin,va_ptr(level),level.size(),va_ptr(start),start.size(),NULL,NULL,bUseModelMsk,msk_model.dataPtr(), true);
    if (!bUseModelMsk) {msk_model.dump("modMsk_out");}

    //After the inversion.
    oDump.dumpFrame(waved);
    //waveCompress->oper(waved,dat,false,false);
    
  } // end if imethod
  //*/

  LOG("b4"); //Reconstuct the offset gather

  if (bUseAngle) {
    waveCompress->oper(waved,angd,false,false); //This is angle gather now
    opA2O->oper(angd,offd,false,false);
    angd.dump("out"); datDump.dumpFrame(offd);
  } else { //from off gather construct angle gather
    LOG(d1);
    waveCompress->oper(waved,offd,false,false); //This is offset gather now
    LOG(d2);
    opA2O->oper(angd,offd,true,false);
    datDump.dumpFrame(offd);
    LOG(d3);
    //waved.info("waved");  offd.info("offd");  angd.info("angd");
    angd.dump("out");
    
  }
} //end main

