#include "stdafx.h"
#include <fstream>

#include <unistd.h>
#include "dmRsf.hpp"

#include "cubeData.hpp"
#include "srcWav.h"
#include "point3d.hpp"

#include "param_entry.h"

#include "stackTraceC.h"
#include "commonOps.h"
#include "util_seplib.h"
#include "misc.h"
#include "cubeAlgo2.h"

#include "dmHeaderReader.h"
#include "my_timer.h"

/**
 * Symmetrize a cube of I(z,h,...) as I(z,h)+I(z,-h)
 * h can be any axis except the last one
 */

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;

/// Usage
/// Symmetrize.x <in.H >out.H ndim=? axis=? verb=y/n

int main (int argc, char *argv[])
{  
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();

  sepHeaderReader header("in"); //output domain
  int iAxis = pars->get_int("axis");

  hypercube shape;
  int ndim = pars->get_int("ndim",-1);
  if (ndim==-1) {
    header.getShape(shape,true);
  } else {
    header.getShape(shape,ndim);
  }
  ndim = shape.get_ndim();
  VERIFY(ndim<=5);

  bool verb = pars->get_bool("verb",false);
  if (verb) {
    PRV2(ndim,iAxis);
  }

  vecAxis axes = shape.get_axes();
  hypercube shpOut(shape), shpOutSlice;
  hypercube* pShpOutSlice = &shpOut;

  VERIFY(iAxis>0 && iAxis<=ndim);

  bool bSliceBySlice=false;
  axis ax5(1); int iLastAxis = 5; int nslice=10,dslice; axis ax5_slice; 
  if (shpOut.get_n123()>2e9) { //if the total size is bigger than 8GB, divide the domain to 10 pieces.
    ax5 = axes[ndim-1];
    VERIFY(ax5.n>1);
    VERIFY(iAxis<ndim); //in this case, the symmetric axis cannot be the last one.
    bSliceBySlice = true; 
    dslice = (ax5.n-1)/nslice+1;
    ax5_slice = ax5; ax5_slice.n = dslice;
    axes[ndim-1] = ax5_slice; shpOutSlice.set_axes(axes);
    pShpOutSlice = &shpOutSlice;
  } else {
    nslice = 1;
  }

  if (verb) {
    shape.info("total input shp info");
    pShpOutSlice->info("ShpSlice info");
  }

  cubeDataF cubeOut(*pShpOutSlice),cube;
  dmRsf_oc<float> dm_out("out",dm::OUT,"dm_out");
  dm_out.init(&shpOut);

  int islice = 0;
  dmRsf_oc<float> dm_cube_in("in",dm::IN,"",true);
  dm_cube_in.init(NULL,ndim);

  for (islice=0; islice<nslice; islice++) {
    cubeOut.zero();
    //if last slice, need to change the size of cubeOut
    if (islice==nslice-1 && bSliceBySlice) {
      axis ax_last = ax5_slice;
      ax_last.n = ax5.n-islice*dslice;
      VERIFY(ax_last.n>1);
      pShpOutSlice->set_axis(ndim,ax_last);
      cubeOut.alloc(*pShpOutSlice);
      if (verb) {
        cubeOut.info("last cubeout info");
      }
    }

    PRV(islice);
    cube.alloc(*pShpOutSlice);

    if (!bSliceBySlice) {
      dm_cube_in.seq_grab(&cube);
    } else { //only read a single slice of it, and updates it
    
      if (verb) {
        dm_cube_in.domain().info("dm_cube_in.domain"); cube.domain().info("cubein.domain");
      }
      int wf = islice*dslice, wj=1, wn=pShpOutSlice->get_n(ndim);
      PRV4(ndim,wf,wj,wn);
      windParam wnd(ndim); wnd[ndim-1].n=wn;wnd[ndim-1].j=wj; wnd[ndim-1].f=wf;
      dm_cube_in.grabWindow(&cube,wnd);
    } //end if bSliceBySlice
    
    //now the input is in "cube", and the output should be in "cubeOut"
    const float* pin; float *pout; 
    pin = cube.dataPtr(); pout = cubeOut.dataPtr();
    int hf = 0;
    if (iAxis==1) {
      long long ntrc = pShpOutSlice->get_n23();
      int nt = pShpOutSlice->get_n(1);
      VERIFY(nt%2==1);
      for (long long i=0; i<ntrc; i++) {
        for (int it=0; it<nt; it++) {
          pout[it] = pin[it]+pin[nt-1-it];
        }
        pin += nt; pout += nt;
      }
    } else if (iAxis>1 && iAxis<ndim){
      long long n1        = pShpOutSlice->get_n23(1,iAxis-1);
      long long nplane    = pShpOutSlice->get_n23(iAxis+1,ndim);
      int nt = pShpOutSlice->get_n(iAxis);
      VERIFY(nt%2==1);
      for (long long ip=0; ip<nplane; ip++) {
        for (int it=0; it<nt; it++) {
          for (long long i1=0; i1<n1; i1++) {
            pout[it*n1+i1] = pin[it*n1+i1]+pin[(nt-1-it)*n1+i1];
          }
        }
        pin += n1*nt; pout += n1*nt;
      }
    } else if (iAxis==ndim) { //last axis
      long long n1        = pShpOutSlice->get_n23(1,iAxis-1);
      int nt = pShpOutSlice->get_n(iAxis);
      VERIFY(nt%2==1);
      for (int it=0; it<nt; it++) {
        for (long long i1=0; i1<n1; i1++) {
          pout[it*n1+i1] = pin[it*n1+i1]+pin[(nt-1-it)*n1+i1];
        }
      }
    } else {
      VERIFY(0);
    }
    dm_out.seq_update(&cubeOut); //update the output file
  } //for islice
  return 0;
}


