#include "stdafx.h"
#include <fstream>

#include <unistd.h>
#include "dmRsf.hpp"

#include "cubeData.hpp"
#include "srcWav.h"
#include "point3d.hpp"

#include "param_entry.h"

#include "stackTraceC.h"
#include "commonOps.h"
#include "util_seplib.h"
#include "misc.h"
#include "cubeAlgo2.h"

#include "dmHeaderReader.h"
#include "my_timer.h"

/**
 * Symmetrize a cube of I(z,h,...) as I(z,h)+I(z,-h)
 * h can be any axis except the last one, 2D case, means (hx,hy)<->(-hx,-hy)
 */

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;

/// Usage, the two axis has to bee axis1 and axis1+1
/// Symmetrize2D.x <in.H >out.H ndim=? axis1=? verb=y/n

int main (int argc, char *argv[])
{  
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();

  sepHeaderReader header("in"); //output domain
  int iAxis = pars->get_int("axis1");

  hypercube shape;
  int ndim = pars->get_int("ndim",-1);
  if (ndim==-1) {
    header.getShape(shape,true);
  } else {
    header.getShape(shape,ndim);
  }
  ndim = shape.get_ndim();
  VERIFY(ndim<=5);

  bool verb = pars->get_bool("verb",false);
  if (verb) {
    PRV2(ndim,iAxis);
  }

  vecAxis axes = shape.get_axes();
  hypercube shpOut(shape), shpOutSlice;
  hypercube* pShpOutSlice = &shpOut;

  VERIFY(iAxis>0 && iAxis<=ndim-1);

  bool bSliceBySlice=false;
  axis ax5(1); int nslice=10,dslice; axis ax5_slice; 
  if (shpOut.get_n123()>2e9) { //if the total size is bigger than 8GB, divide the domain to 10 pieces.
    ax5 = axes[ndim-1];
    VERIFY(ax5.n>1);
    VERIFY(iAxis<=ndim-2); //in this case, the symmetric axes cannot be the last two
    bSliceBySlice = true; 
    dslice = (ax5.n-1)/nslice+1;
    ax5_slice = ax5; ax5_slice.n = dslice;
    axes[ndim-1] = ax5_slice; shpOutSlice.set_axes(axes);
    pShpOutSlice = &shpOutSlice;
  } else {
    nslice = 1;
  }

  if (verb) {
    shape.info("total input shp info");
    pShpOutSlice->info("ShpSlice info");
  }

  cubeDataF cubeOut(*pShpOutSlice),cube;
  dmRsf_oc<float> dm_out("out",dm::OUT,"dm_out");
  dm_out.init(&shpOut);

  int islice = 0;
  dmRsf_oc<float> dm_cube_in("in",dm::IN,"",true); dm_cube_in.init(NULL,ndim);

  for (islice=0; islice<nslice; islice++) {
    cubeOut.zero();
    //if last slice, need to change the size of cubeOut
    if (islice==nslice-1 && bSliceBySlice) {
      axis ax_last = ax5_slice;
      ax_last.n = ax5.n-islice*dslice;
      VERIFY(ax_last.n>1);
      pShpOutSlice->set_axis(ndim,ax_last);
      cubeOut.alloc(*pShpOutSlice);
      if (verb) {
        cubeOut.info("last cubeout info");
      }
    }

    PRV(islice);
    cube.alloc(*pShpOutSlice);
    
    //now the input is in "cube", and the output should be in "cubeOut".
    //reshape the cubes so it is easy to do indexing. into 4 axis (ax1,ax_hx,ax_hy,ax_plane)
    hypercube shpRe;
    vecAxis axesRe = pShpOutSlice->get_axes();
    axis ax_dummy(1,0,1.f);
    axis ax1,ax_hx,ax_hy,ax_plane;
    ax1 = ax_dummy; ax_plane = ax_dummy;
    ax_hx = axesRe[iAxis-1]; ax_hy = axesRe[iAxis];
    //Figure out the first axis of the reshaped grid
    if (iAxis==1) { //Pad one dummy axis to the beginning if hx is the first axis
      ax1 = ax_dummy;
    } else {
      ax1 = lumpAxes(axesRe, 1,iAxis-1);
      for (int i=1; i<iAxis; i++) {
        axesRe.erase(axesRe.begin());
      }
    }
    //Figureout the last axis ax_plane
    if (iAxis==ndim-1) {
      ax_plane = ax_dummy;
    } else {
      ax_plane=lumpAxes(pShpOutSlice->get_axes(), iAxis+2, ndim);
      for (int i=iAxis+2; i<=ndim; i++) {
        axesRe.pop_back();
      }
    }
    axesRe.insert(axesRe.begin(),ax1);
    axesRe.push_back(ax_plane);
    shpRe.set_axes(axesRe);
    shpRe.info("shp Reformed");
    VERIFY(axesRe.size()==4);

    //load the input data
    if (!bSliceBySlice) {
      dm_cube_in.seq_grab(&cube);
    } else { //only read a single slice of it, and updates it
      if (verb) {
        dm_cube_in.domain().info("dm_cube_in.domain"); cube.domain().info("cubein.domain");
      }
      int wf = islice*dslice, wj=1, wn=pShpOutSlice->get_n(ndim);
      PRV4(ndim,wf,wj,wn);
      windParam wnd(ndim); wnd[ndim-1].n=wn;wnd[ndim-1].j=wj; wnd[ndim-1].f=wf;
      dm_cube_in.grabWindow(&cube,wnd);
    } //end if bSliceBySlice

    cube.reshape(shpRe); cubeOut.reshape(shpRe);
    for (int ip=0; ip<ax_plane.n; ip++) {
      for (int ix=0; ix<ax_hx.n; ix++) {
      for (int iy=0; iy<ax_hy.n; iy++) {
        for (int i1=0; i1<ax1.n; i1++) {
          cubeOut(ip,iy,ix,i1) = cube(ip,iy,ix,i1)+cube(ip,ax_hy.n-1-iy,ax_hx.n-1-ix,i1);
        }
      }}
    }
    //reshape it back
    cube.reshape(*pShpOutSlice); cubeOut.reshape(*pShpOutSlice);
    dm_out.seq_update(&cubeOut); //update the output file
  } //for islice
  return 0;
}


